/******************************************************************************
*******************************************************************************
**
**    Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: gfs_linux.h
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_LINUX_DOT_H__
#define __GFS_LINUX_DOT_H__


/*  Include Files
    These should only be the ones necessary to compile the common code.  */

#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>



/*  Memory allocation abstraction  */

static __inline__ void *gmalloc(int size)
{
  if (size <= 131072)
    return(kmalloc(size, GFP_KERNEL));
  else
    return(vmalloc(size));
}

static __inline__ void gfree(void *data, int size)
{
  if (size <= 131072)
    kfree(data);
  else
    vfree(data);
}



/*  Memory copy abstraction

    memset() and memcopy() are already defined on Linux.
    */

#define GFS_IO_READ(b, c, f, l) copy_to_user((char *)(b) + (c), (f), (l))
#define GFS_IO_READ_HOLE(b, c, l) clear_user((char *)(b) + (c), (l))
#define GFS_IO_WRITE(b, c, t, l) copy_from_user((t), (char *)(b) + (c), (l))



/*  printk() abstraction

    None.  The linux printk() is already the interface GFS uses.
    KERN_INFO, KERN_DEBUG, KERN_ERR, and KERN_EMERG are already
    defined.
    */



/*  VFS to GFS abstraction  */

/*  Convert a Linux superblock in to a GFS superblock.  */
#define vsb2sbd(vsbp) ((gfs_sbd_t *)((vsbp)->u.generic_sbp))

/*  Convert a Linux inode into a GFS inode.  */
#define vi2ino(vip) ((gfs_inode_t *)((vip)->u.generic_ip))

/*  Convert a Linux Dcache entry into GFS Dcache data  */
#define vdc2dcd(dentry) ((gfs_dcached_t *)((dentry)->d_fsdata))

/*  Convert a linux Buffer head into GFS buffer head data */
#define bh2pd(bh) ((gfs_bhdata_t *)((bh)->b_pdata))



/*  GFS to VFS abstraction  */

typedef struct super_block gfs_vsb_t;
typedef struct inode gfs_vino_t;



/*  Buffer head abstraction 
    
    All macros take a pointer to a buffer_head_t structure.

    BH_DATA() returns a pointer to the data held in the buffer.
    BH_SIZE() returns the size in bytes of the data in the buffer.
    BH_COUNT() is the usage count of the buffer.
    */

typedef struct buffer_head buffer_head_t;
#define BH_DATA(x) ((x)->b_data)
#define BH_BLKNO(x) ((x)->b_blocknr)
#define BH_SIZE(x) ((x)->b_size)



/*  Endianness conversion abstraction

    The default is the Linux be64_to_cpu() type interface  */



/*  Directory translation abstraction

    The default is the Linux filldir_t routines  */



/*  struct qstr definition 
 
    "quick string" -- eases parameter passing, but more importantly
    saves "metadata" about the string (ie length and the hash).

    Already defined in Linux
    */



/*  Semaphore abstraction

    All macros take a pointer to a semaphore_t structure.
    GFS_SEMA_DOWN_TRY() returns FALSE if the down() succeeds 
    */

typedef struct semaphore semaphore_t;
#define GFS_SEMA_DOWN(x) {IFDEBUG(DEBUG_SEMAPHORE, \
				  printk(DEBUG_HOW "(%u) Down: <%.16lX> %d\t%s\t\"%s\"\n", \
					 kpid(), \
					 (unsigned long)(x), __LINE__, __FILE__, #x);); \
			  down((x)); \
			  IFDEBUG(DEBUG_SEMAPHORE, \
				  printk(DEBUG_HOW "(%u) done\n", kpid());); \
			 }
#define GFS_SEMA_DOWN_INTERRUPTIBLE(x) {IFDEBUG(DEBUG_SEMAPHORE, \
						printk(DEBUG_HOW "(%u) Down: <%.16lX> %d\t%s\t\"%s\"\n", \
						       kpid(), \
						       (unsigned long)(x), __LINE__, __FILE__, #x);); \
				        down_interruptible((x)); \
				        IFDEBUG(DEBUG_SEMAPHORE, \
						printk(DEBUG_HOW "(%u) done\n", kpid());); \
				       }
#define GFS_SEMA_DOWN_TRY(x) (down_trylock((x)))
#define GFS_SEMA_UP(x) {IFDEBUG(DEBUG_SEMAPHORE, \
				printk(DEBUG_HOW "(%u) Up:   <%.16lX> %d\t%s\t\"%s\"\n", \
				       kpid(), \
				       (unsigned long)(x), __LINE__, __FILE__, #x);); \
			up((x)); \
			IFDEBUG(DEBUG_SEMAPHORE, \
				printk(DEBUG_HOW "(%u) done\n", kpid());); \
		       }
#define GFS_SEMA_INIT(x) {*(x) = MUTEX;}
#define GFS_SEMA_INIT_LKD(x) {*(x) = MUTEX_LOCKED;}



/*  PID abstraction

    Returns the PID of the current process.
    */

#define kpid() (current->pid)



/*  CURRENT_TIME macro 
    Returns the standard UNIX time integer 

    CURRENT_TIME is already defined in Linux.
    */



/*  FSBTOBB() definition  */

#define FSBTOBB(b)  (b)



/*  Other Stuff  */



#endif  /*  __GFS_LINUX_DOT_H__  */

