/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name:  bmap.c
**
**    Description:
**      These routines translate from extents in a file to block numbers.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "inode.h"



typedef struct metapath_s
{
  uint64                 mp_offset;     /* Offset that his path follows */

  int                    mp_height;     /* Height of the metadata tree 
					   i.e. the number of elements in mp_list */

  uint64                *mp_list;       /* Array of indexes in the metadata to get
					   to mp_offset */
} metapath_t;



#define MT_LEFT        (13)
#define MT_RIGHT       (14)
#define MT_ONLY        (15)
#define MT_MIDDLE      (16)

typedef struct metatree_s
{
  int                    mt_height;     /* The height of this pointer block */
  int                    mt_flag;       /* What part of the tree is this pointer block
					   (MT_LEFT, MT_RIGHT, etc...) */

  int                    mt_start;      /* The first index used in this pointer block */
  int                    mt_len;        /* The number of indexes in this pointer block */

  int                    mt_dirty;      /* Is this block dirty? */
  buffer_head_t         *mt_bh;         /* The pointer block on disk */

  uint64                *mt_leaf_bn;    /* Array of block numbers in the pointer block */
  struct metatree_s    **mt_leaf;       /* Children of this pointer block */
  struct metatree_s     *mt_parent;     /* Parent of this pointer block */

  struct metatree_s     *mt_next;       /* Next pointer in the list */
} metatree_t;



/******************************************************************************
*******************************************************************************
**
** gfs_unstuff_dinode()
**
** Description:
**   This routine unstuffs a dinode and returns it to a "normal" state such 
**   that the height can be grown in the traditional way.
**
** Input(s):
**   ip - The inode to be unstuffed
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_unstuff_dinode(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_blklist_t *blk;
  buffer_head_t *bh;

  int error;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering gfs_unstuff_dinode()\n"););


  ASSERT(ip->i_di.di_height == 0);


  /*  Get a free block, fill it with the stuffed data, and write it out to disk  */

  if ((error = gfs_blkalloc(ip, 1, &blk, 0)))
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_unstuff_dinode()\n"););
    return(error);
  }

  bh = gfs_dgetblk(sdp, blk->blk_no, ip->i_di.di_glock);
  if (!bh)
  {
    gfs_blkfree(sdp, &blk);
    if (blk)
    {
      GARBAGE_BLOCKS();
      gfs_blklist_free(blk);
    }
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_unstuff_dinode()\n"););
    return(-EIO);
  }

  gfs_mark_uptodate(bh, TRUE); 
  memset(BH_DATA(bh), 0, sdp->sd_sb.sb_bsize);
  memcpy(BH_DATA(bh), BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t), ip->i_di.di_bsize);
  gfs_dwrite(bh, DW_NOKEEP | ((ip->i_di.di_type == FILE_REG) ? DW_NOTMETA : 0));

  /*  Set up the pointer to the new block  */

  memset(BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t), 0, sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t));
  ((uint64 *)(BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t)))[0] = cpu_to_be64(blk->blk_no);

  ip->i_di.di_height = 1;
  ip->i_di.di_blocks++; 

  /*  Write the new dinode  */

  gfs_inode_attr_in(ip);
  gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
  gfs_dwrite(ip->i_didata, 0);

  gfs_blklist_free(blk);

  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting gfs_unstuff_dinode()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** build_height()
**
** Description:
**   This routine makes sure that the metadata tree is tall enough to hold
**   "offset + size" bytes of data.
**
** Input(s):
**   ip - the dinode to add height to
**   offset - the offset of the current write operation
**   size - the lenght of the current write operation
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int build_height(gfs_inode_t *ip, uint64 offset, uint64 size)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_blklist_t *blk;
  buffer_head_t *bh;

  uint64 sz;
  int height;
  uint64 space;
  int error;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering build_height()\n"););


  sz = (offset + size > ip->i_di.di_bsize) ? offset + size : ip->i_di.di_bsize;

  if (sz <= sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t))
    return(0);
  else
    for (height = 1, space = sdp->sd_diptrs * sdp->sd_sb.sb_bsize;
	 sz > space;
	 height++, space *= sdp->sd_inptrs)
    {
      /*  Do nothing  */
    }


  while (ip->i_di.di_height < height)
  {
    /*  Get a new block, fill it with the old direct pointers and write it out  */

    if ((error = gfs_blkalloc(ip, 1, &blk, 1)))
    {
      IFDEBUG(DEBUG_BMAP1,
	      printk(DEBUG_HOW "Exiting build_height(error 1)\n"););
      return(error);
    }

    bh = gfs_dgetblk(sdp, blk->blk_no, ip->i_di.di_glock);
    if (!bh)
    {
      gfs_blkfree(sdp, &blk);
      if (blk)
      {
	GARBAGE_BLOCKS();
	gfs_blklist_free(blk);
      }
      IFDEBUG(DEBUG_BMAP1,
	      printk(DEBUG_HOW "Exiting build_height(error 2)\n"););
      return(-EIO);
    }

    gfs_mark_uptodate(bh, TRUE);
    memset(BH_DATA(bh), 0, sdp->sd_sb.sb_bsize);
    memcpy(BH_DATA(bh), BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t), sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t));
    gfs_dwrite(bh, DW_NOKEEP);

    /*  Set up the new direct pointer and write it out to disk  */

    memset(BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t), 0, sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t));
    ((uint64 *)(BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t)))[0] = cpu_to_be64(blk->blk_no);

    ip->i_di.di_height++;
    ip->i_di.di_blocks++;

    gfs_inode_attr_in(ip);
    gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
    gfs_dwrite(ip->i_didata, 0);

    gfs_blklist_free(blk);
  }


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting build_height(ok)\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** find_metapath()
**
** Description:
**   This routine returns a metapath_t structure that defines a path through
**   the metadata of inode "ip" to get to offset "offset".
**
**   For Example:
**     Given:  "ip" is a height 3 file, "offset" is 101342453, and this is a
**       filesystem with a blocksize of 4096.
**
**     find_metapath() would return a metapath_t structure set to:
**       mp_offset = 101342453, mp_height = 3, mp_list[0] = 0, mp_list[1] = 48,
**       and mp_list[2] = 165.
**
**     That means that in order to get to the block containing the byte at
**     offset 101342453, we would load the indirect block pointed to by pointer
**     0 in the dinode.  We would then load the indirect block pointed to by
**     pointer 48 in that indirect block.  We would then load the data block
**     pointed to by pointer 165 in that indirect block.
**
**             ----------------------------------------
**             | Dinode |                             |
**             |        |                            4|
**             |        |0 1 2 3 4 5                 9|
**             |        |                            6|
**             ----------------------------------------
**                       |
**                       |
**                       V
**             ----------------------------------------
**             | Indirect Block                       |
**             |                                     5|
**             |            4 4 4 4 4 5 5            1|
**             |0           5 6 7 8 9 0 1            2|
**             ----------------------------------------
**                                |
**                                |
**                                V
**             ----------------------------------------
**             | Indirect Block                       |
**             |                         1 1 1 1 1   5|
**             |                         6 6 6 6 6   1|
**             |0                        3 4 5 6 7   2|
**             ----------------------------------------
**                                           |
**                                           |
**                                           V
**             ----------------------------------------
**             | Data block containing offset         |
**             |            101342453                 |
**             |                                      |
**             |                                      |
**             ----------------------------------------
**
** Input(s):
**   ip - the inode
**   offset - offset in the inode
**
** Returns:
**   A metapath_t structure on success, NULL on failure
**
*******************************************************************************
******************************************************************************/

static metapath_t *find_metapath(gfs_inode_t *ip, uint64 offset)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  metapath_t *mp;

  int i;
  int height = ip->i_di.di_height;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering find_metapath()\n"););


  mp = (metapath_t *)gmalloc(sizeof(metapath_t));
  if (!mp)
    return(NULL);
  mp->mp_height = height;
  mp->mp_offset = offset;

  mp->mp_list = (uint64 *)gmalloc(height * sizeof(uint64));
  if (!mp->mp_list)
  {
    gfree(mp, sizeof(metapath_t));
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting find_metapath()\n"););
    return(NULL);
  }


  mp->mp_list[height - 1] = offset / sdp->sd_sb.sb_bsize;
  
  for (i = height - 1; i--; )
    mp->mp_list[i] = mp->mp_list[i + 1] / sdp->sd_inptrs;

  for (i = 0; i < height; i++)
    mp->mp_list[i] %= sdp->sd_inptrs;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting find_metapath()\n"););

  return(mp);
}




/******************************************************************************
*******************************************************************************
**
** recurse_metatree()
**
** Description:
**   Recurse through the metadata tree and create metatree_t structures for
**   the indirect blocks that are used by this transfer.
**
** Input(s):
**   ip - the inode
**   mp1 - the path to the first byte of the extent
**   mp2 - the path to the last byte of the extent
**   mt - We fill in the children of this metatree_t structure
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int recurse_metatree(gfs_inode_t *ip, metapath_t *mp1, metapath_t *mp2, metatree_t *mt)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  metatree_t *mtmp, *mhead = NULL, *mlast = NULL;

  int i;
  uint64 bn;
  int error;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering recurse_metatree()\n"););


  /*  Figure out which of this block's leaves are going to be used  */

  switch(mt->mt_flag)
  {
  case MT_ONLY:
    mt->mt_start = mp1->mp_list[mt->mt_height];
    mt->mt_len = mp2->mp_list[mt->mt_height] - mp1->mp_list[mt->mt_height] + 1;
    break;

  case MT_LEFT:
    mt->mt_start = mp1->mp_list[mt->mt_height];
    mt->mt_len = sdp->sd_inptrs - mt->mt_start;
    break;

  case MT_RIGHT:
    mt->mt_start = 0;
    mt->mt_len = mp2->mp_list[mt->mt_height] + 1;
    break;

  case MT_MIDDLE:
    mt->mt_start = 0;
    mt->mt_len = sdp->sd_inptrs;
    break;
  }


  /*  Set up the pointers so we can get to the block numbers of those leaves  */

  if (mt->mt_bh)
    mt->mt_leaf_bn = ((uint64 *)(BH_DATA(mt->mt_bh) + ((mt->mt_height == 0) ? sizeof(gfs_dinode_t) : 0))) +
      mt->mt_start;


  /*  If there is a indirect layer below this one recurse into it.  */

  if (mt->mt_height + 1 < ip->i_di.di_height)
  {
    mt->mt_leaf = (metatree_t **)gmalloc(mt->mt_len * sizeof(metatree_t *));
    ASSERT(mt->mt_leaf != NULL);  /*  FixMe!!!  */


    for (i = 0; i < mt->mt_len; i++)
    {
      mtmp = (metatree_t *)gmalloc(sizeof(metatree_t));
      ASSERT(mtmp != NULL);  /*  FixMe!!!  */
      memset(mtmp, 0, sizeof(metatree_t));
      
      mt->mt_leaf[i] = mtmp;
      
      mtmp->mt_height = mt->mt_height + 1;


      switch(mt->mt_flag)
      {
      case MT_ONLY:
	if (mt->mt_len == 1)
	  mtmp->mt_flag = MT_ONLY;
	else if (i == 0)
	  mtmp->mt_flag = MT_LEFT;
	else if (i == mt->mt_len - 1)
	  mtmp->mt_flag = MT_RIGHT;
	else
	  mtmp->mt_flag = MT_MIDDLE;
	break;
	
      case MT_LEFT:
	if (i == 0)
	  mtmp->mt_flag = MT_LEFT;
	else
	  mtmp->mt_flag = MT_MIDDLE;
	break;
	
      case MT_RIGHT:
	if (i == mt->mt_len - 1)
	  mtmp->mt_flag = MT_RIGHT;
	else
	  mtmp->mt_flag = MT_MIDDLE;
	break;
	
      case MT_MIDDLE:
	mtmp->mt_flag = MT_MIDDLE;
	break;
      }
      
      if (mt->mt_bh)
      {
	bn = be64_to_cpu(mt->mt_leaf_bn[i]);
      
	if (bn)
	{
	  mtmp->mt_bh = gfs_dread(sdp, bn, ip->i_di.di_glock);
	  ASSERT(mtmp->mt_bh != NULL);  /*  FixMe!!!  */
	}
      }

      mtmp->mt_parent = mt;      
      
      error = recurse_metatree(ip, mp1, mp2, mtmp);
      ASSERT(error == 0);  /*  FixMe!!!  */

      if (mhead)
	mlast->mt_next = mtmp;
      else
	mhead = mtmp;

      for ( ; mtmp; mlast = mtmp, mtmp = mtmp->mt_next) ;
    }
  }


  mt->mt_next = mhead;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting recurse_metatree()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** build_metatree()
**
** Description:
**   Given the "offset" and "size" of a transfer, this routine creates a
**   in-core representation of the metadata structure out on disk.  For each
**   indirect block that is involved in this transfer, there is one metatree_t
**   structure that represents it.
**
**   This routine just sets up the metatree_t structure that corresponds
**   to the dinode and calls recurse_metatree() to fill in the rest.
**
**   One important thing this routine (and recurse_metatree()) does is convert
**   the tree structure of indirect blocks into a linear linked list.
**
** Input(s):
**   ip - The inode
**   offset - Starting byte of the transfer
**   size - length of the transfer
**
** Output(s):
**   metatree - the metatree to be returned
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int build_metatree(gfs_inode_t *ip, uint64 offset, uint64 size, metatree_t **metatree)
{
  metatree_t *mt;
  metapath_t *mp1, *mp2;

  int error;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering build_metatree()\n"););


  /*  Find paths to the beginning and end of the extent.  */

  mp1 = find_metapath(ip, offset);
  if (!mp1)
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting build_metatree()\n"););
    return(-ENOMEM);
  }

  mp2 = find_metapath(ip, offset + size - 1);
  if (!mp2)
  {
    gfree(mp1->mp_list, mp1->mp_height * sizeof(uint64));
    gfree(mp1, sizeof(metapath_t));
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting build_metatree()\n"););
    return(-ENOMEM);
  }


  /*  Allocate and setup the metatree_t for the dinode  */

  mt = (metatree_t *)gmalloc(sizeof(metatree_t));
  if (!mt)
  {
    gfree(mp1->mp_list, mp1->mp_height * sizeof(uint64));
    gfree(mp1, sizeof(metapath_t));
    gfree(mp2->mp_list, mp2->mp_height * sizeof(uint64));
    gfree(mp2, sizeof(metapath_t));
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting build_metatree()\n"););
    return(-ENOMEM);
  }
  memset(mt, 0, sizeof(metatree_t));

  mt->mt_height = 0;
  mt->mt_flag = MT_ONLY;
  
  mt->mt_bh = ip->i_didata;


  /*  Fill in the rest of the tree and return  */

  error = recurse_metatree(ip, mp1, mp2, mt);
  if (error)
  {
    gfree(mt, sizeof(metatree_t));
    mt = NULL;
  }


  gfree(mp1->mp_list, mp1->mp_height * sizeof(uint64));
  gfree(mp1, sizeof(metapath_t));
  gfree(mp2->mp_list, mp2->mp_height * sizeof(uint64));
  gfree(mp2, sizeof(metapath_t));


  *metatree = mt;

  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting build_metatree()\n"););

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** free_metatree()
**
** Description:
**   Free a metatree.  Write back the dirty blocks.
**
** Input(s):
**   mt - the metatree to free
**
*******************************************************************************
******************************************************************************/

static void free_metatree(metatree_t *mt)
{
  metatree_t *mtmp;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering free_metatree()\n"););


  while (mt)
  {
    mtmp = mt;
    mt = mt->mt_next;

    if (mtmp->mt_height != 0 && mtmp->mt_bh)
    {
      if (mtmp->mt_dirty)
	gfs_dwrite(mtmp->mt_bh, DW_NOKEEP);
      else
	gfs_dforget(mtmp->mt_bh);
    }

    if (mtmp->mt_leaf)
      gfree(mtmp->mt_leaf, mtmp->mt_len * sizeof(metatree_t *));

    gfree(mtmp, sizeof(metatree_t));
  }


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting free_metatree()\n"););
}


/******************************************************************************
*******************************************************************************
**
** metadata()
**
** Description:
**   This routine takes a file of height 1 or more and an extent to be
**   transfered.  It then returns a gfs_direct_t array of blocks that point
**   to the data blocks.  
**
** Input(s):
**   ip - the inode
**   offset - the start of the extent
**   size - the lenght of the extent
**   write - Is this a read or a write?
**
** Output(s):
**   An array of gfs_direct_t structures that describe the data blocks to
**   be written
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int metadata(gfs_inode_t *ip, uint64 offset, uint64 size, gfs_direct_t **direct, int write)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  metatree_t *mt, *mtmp;
  gfs_blklist_t *blks, *btmp;
  gfs_direct_t *dir;

  uint64 bn;
  int error;
  int blocks, ndir;
  int i, j = 0;
  uint64 aoff, aend, asize;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering metadata()\n"););


  /*  Figure out what the data blocks and the metadata supporting them look like.  */

  if ((error = build_metatree(ip, offset, size, &mt)))
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting metadata()\n"););
    return(error);
  }


  IFDEBUG(DEBUG_BMAP3,
	  int x;
	  for (mtmp = mt; mtmp; mtmp = mtmp->mt_next, x++)
	  {
	    printk(DEBUG_HOW "  metatree element <0x%.16lX>\n", (unsigned long)mtmp);
	    printk(DEBUG_HOW "    mt_height = %d\n", mtmp->mt_height);
	    printk(DEBUG_HOW "    mt_flag = %d\n", mtmp->mt_flag);
	    printk(DEBUG_HOW "    mt_start = %d\n", mtmp->mt_start);
	    printk(DEBUG_HOW "    mt_len = %d\n", mtmp->mt_len);
	    printk(DEBUG_HOW "    mt_dirty = %d\n", mtmp->mt_dirty);
	    printk(DEBUG_HOW "    mt_bh = <0x%.16lX>\n", (unsigned long)mtmp->mt_bh);
	    printk(DEBUG_HOW "    mt_leaf_bn = %s\n", (mtmp->mt_leaf_bn) ? "not NULL" : "NULL");
	    if (mtmp->mt_leaf_bn)
	      for (x = 0; x < mtmp->mt_len; x++)
		printk(DEBUG_HOW "    mt_leaf_bn[%d] = %"FMT_64"u\n", x, be64_to_cpu(mtmp->mt_leaf_bn[x]));
	    printk(DEBUG_HOW "    mt_leaf = %s\n", (mtmp->mt_leaf) ? "not NULL" : "NULL");
	    if (mtmp->mt_leaf)
	      for (x = 0; x < mtmp->mt_len; x++)
		printk(DEBUG_HOW "    mt_leaf[%d] = <0x%.16lX>\n", x, (unsigned long)mtmp->mt_leaf[x]);
	    printk(DEBUG_HOW "    mt_parent = <0x%.16lX>\n", (unsigned long)mtmp->mt_parent);
	    printk(DEBUG_HOW "    mt_next = <0x%.16lX>\n", (unsigned long)mtmp->mt_next);
	  });


  /*  Get the gfs_direct_t structures  */

  aoff = offset - (offset % sdp->sd_sb.sb_bsize);
  aend = offset + size;
  if (aend % sdp->sd_sb.sb_bsize)
    aend = aend - (aend % sdp->sd_sb.sb_bsize) + sdp->sd_sb.sb_bsize;
  asize = aend - aoff;

  ndir = asize / sdp->sd_sb.sb_bsize;

  dir = (gfs_direct_t *)gmalloc(ndir * sizeof(gfs_direct_t));
  if (!dir)
  {
    free_metatree(mt);
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting metadata()\n"););
    return(-ENOMEM);
  }

  memset(dir, 0, ndir * sizeof(gfs_direct_t));
  dir->dr_num = ndir;


  if (write)
  {
    /*  How many METADATA blocks do we need to alloc?  */

    for (blocks = 0, mtmp = mt; mtmp; mtmp = mtmp->mt_next)
      if (!mtmp->mt_bh)
	blocks++;

    IFDEBUG(DEBUG_BMAP3,
	    printk(DEBUG_HOW "  Need %d metadata blocks\n", blocks););

    if (blocks)
    {
      /*  Get them  */

      if ((error = gfs_blkalloc(ip, blocks, &blks, 1)))
      {
	free_metatree(mt);
	gfree(dir, ndir * sizeof(gfs_direct_t));
	IFDEBUG(DEBUG_BMAP1,
		printk(DEBUG_HOW "Exiting metadata()\n"););
	return(error);
      }

      /*  Increment the block count  */

      ip->i_di.di_blocks += blocks;

      /*  Place them in the tree  */

      for (mtmp = mt; mtmp; mtmp = mtmp->mt_next)
	if (!mtmp->mt_bh)
	{
	  bn = blks->blk_no;

	  blks->blk_no++;
	  blks->blk_len--;
	  if (blks->blk_len == 0)
	  {
	    btmp = blks;
	    blks = blks->blk_next;
	    gfree(btmp, sizeof(gfs_blklist_t));
	  }

	  mtmp->mt_bh = gfs_dgetblk(sdp, bn, ip->i_di.di_glock);
	  ASSERT(mtmp->mt_bh != NULL);  /*  FixMe!!!  */
	  gfs_mark_uptodate(mtmp->mt_bh, TRUE);
	  memset(BH_DATA(mtmp->mt_bh), 0, sdp->sd_sb.sb_bsize);

	  mtmp->mt_leaf_bn = ((uint64 *)(BH_DATA(mtmp->mt_bh))) + mtmp->mt_start;

	  for (i = 0; i < mtmp->mt_parent->mt_len; i++)
	    if (mtmp->mt_parent->mt_leaf[i] == mtmp)
	      break;

	  ASSERT(i < mtmp->mt_parent->mt_len);

	  mtmp->mt_parent->mt_leaf_bn[i] = cpu_to_be64(bn);

	  mtmp->mt_parent->mt_dirty = TRUE;
	  mtmp->mt_dirty = TRUE;
	}
    }

    /*  How many DATA blocks do we need to alloc?  */

    for (blocks = 0, mtmp = mt; mtmp; mtmp = mtmp->mt_next)
      if (mtmp->mt_height == ip->i_di.di_height - 1)
	for (i = 0; i < mtmp->mt_len; i++)
	  if (mtmp->mt_leaf_bn[i] == 0)  /*  Zero is zero, regardless of endianess  */
	    blocks++;
	     
    IFDEBUG(DEBUG_BMAP3,
	    printk(DEBUG_HOW "  Need %d data blocks\n", blocks););

    if (blocks)
    {
      /*  Get them  */

      if ((error = gfs_blkalloc(ip, blocks, &blks, 0)))
      {
	free_metatree(mt);
	gfree(dir, ndir * sizeof(gfs_direct_t));
	IFDEBUG(DEBUG_BMAP1,
		printk(DEBUG_HOW "Exiting metadata()\n"););
	return(error);
      }

      /*  Increment the block count  */

      ip->i_di.di_blocks += blocks;
    }

    /*  Place them in the tree (and fill in the gfs_direct_t at the same time.)  */

    for (mtmp = mt; mtmp; mtmp = mtmp->mt_next)
      if (mtmp->mt_height == ip->i_di.di_height - 1)
	for (i = 0; i < mtmp->mt_len; i++)
	{
	  if (mtmp->mt_leaf_bn[i] == 0)  /*  Zero is zero, regardless of endianess  */
	  {
	    bn = blks->blk_no;
	    
	    blks->blk_no++;
	    blks->blk_len--;
	    if (blks->blk_len == 0)
	    {
	      btmp = blks;
	      blks = blks->blk_next;
	      gfree(btmp, sizeof(gfs_blklist_t));
	    }

	    mtmp->mt_leaf_bn[i] = cpu_to_be64(bn);

	    mtmp->mt_dirty = TRUE;

	    dir[j].dr_flags = GFS_BMAP_NEW;
	    dir[j].dr_no = bn;
	  }
	  else
	    dir[j].dr_no = be64_to_cpu(mtmp->mt_leaf_bn[i]);

	  if (mtmp->mt_flag == MT_ONLY && mtmp->mt_len == 1)
	  {
	    dir[j].dr_start = offset % sdp->sd_sb.sb_bsize;
	    dir[j].dr_len = size;
	  }
	  else if (j == 0)
	  {
	    dir[j].dr_start = offset % sdp->sd_sb.sb_bsize;
	    dir[j].dr_len = sdp->sd_sb.sb_bsize - dir[j].dr_start;
	  }
	  else if (j == ndir - 1)
	  {
	    dir[j].dr_start = 0;
	    dir[j].dr_len = (offset + size) % sdp->sd_sb.sb_bsize;
	    if (dir[j].dr_len == 0)
	      dir[j].dr_len = sdp->sd_sb.sb_bsize;
	  }
	  else
	  {
	    dir[j].dr_start = 0;
	    dir[j].dr_len = sdp->sd_sb.sb_bsize;
	  }

	  j++;
	}

    free_metatree(mt);
    *direct = dir;

    if (offset + size > ip->i_di.di_bsize)
      ip->i_di.di_bsize = offset + size;

    gfs_inode_attr_in(ip);
    gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
    gfs_dwrite(ip->i_didata, 0);
  
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting metadata()\n"););

    return(0);
  }


  /*  Handle a read  */

  for (mtmp = mt; mtmp; mtmp = mtmp->mt_next)
    if (mtmp->mt_height == ip->i_di.di_height - 1)
      for (i = 0; i < mtmp->mt_len; i++)
      {
	if (mtmp->mt_bh)
	  bn = be64_to_cpu(mtmp->mt_leaf_bn[i]);
	else
	  bn = 0;

	if (bn)
	  dir[j].dr_no = bn;
	else
	  dir[j].dr_flags = GFS_BMAP_HOLE;

	if (mtmp->mt_flag == MT_ONLY && mtmp->mt_len == 1)
	{
	  dir[j].dr_start = offset % sdp->sd_sb.sb_bsize;
	  dir[j].dr_len = size;
	}
	else if (j == 0)
	{
	  dir[j].dr_start = offset % sdp->sd_sb.sb_bsize;
	  dir[j].dr_len = sdp->sd_sb.sb_bsize - dir[j].dr_start;
	}
	else if (j == ndir - 1)
	{
	  dir[j].dr_start = 0;
	  dir[j].dr_len = (offset + size) % sdp->sd_sb.sb_bsize;
	  if (dir[j].dr_len == 0)
	    dir[j].dr_len = sdp->sd_sb.sb_bsize;
	}
	else
	{
	  dir[j].dr_start = 0;
	  dir[j].dr_len = sdp->sd_sb.sb_bsize;
	}

	j++;
      }


  free_metatree(mt);
  *direct = dir;

  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Exiting metadata()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_bmapi_read()
**
** Description:
**   Return a gfs_direct_t array that describes which data blocks are to
**   be read for this transfer.
**
** Input(s):
**   ip - the inode to be read
**   offset - the start of the extent to be read
**   size - the length of the extent to be read
**
** Output(s):
**   direct - the description of the blocks to be read
**
** Returns:
**   0 on success, -EXXXX on error
**
*******************************************************************************
******************************************************************************/

int gfs_bmapi_read(gfs_inode_t *ip, uint64 offset, uint64 size, gfs_direct_t **direct)
{
  gfs_direct_t *dtmp;

  int error;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering gfs_bmapi_read()\n"););


  /*  Check for special cases  */

  *direct = NULL;

  if (offset < 0)
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_read()\n"););
    return(-EINVAL);
  }

  if (offset > ip->i_di.di_bsize)
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_read()\n"););
    return(0);
  }

  if (offset + size > ip->i_di.di_bsize)
    size = ip->i_di.di_bsize - offset;

  if (!size)
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_read()\n"););
    return(0);
  }


  if (ip->i_di.di_height == 0)  /*  Stuffed Dinode case  */
  {
    dtmp = (gfs_direct_t *)gmalloc(sizeof(gfs_direct_t));
    if (!dtmp)
    {
      IFDEBUG(DEBUG_BMAP1,
	      printk(DEBUG_HOW "Exiting gfs_bmapi_read()\n"););
      return(-ENOMEM);
    }

    dtmp->dr_num = 1;
    dtmp->dr_flags = GFS_BMAP_STUFFED_DINODE;
    dtmp->dr_no = 0;
    dtmp->dr_start = sizeof(gfs_dinode_t) + offset;
    dtmp->dr_len = size;

    *direct = dtmp;

    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_read()\n"););

    return(0);
  }
  else  /*  Regular file  */
  {
    error = metadata(ip, offset, size, direct, FALSE);

    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_read()\n"););

    return(error);
  }
}


/******************************************************************************
*******************************************************************************
**
** gfs_bmapi_write()
**
** Description:
**   Return a gfs_direct_t array that describes which data blocks are to
**   be written for this transfer.
**
** Input(s):
**   ip - the inode to be written
**   offset - the start of the extent to be written
**   size - the length of the extent to be written
**
** Output(s):
**   direct - the description of the blocks to be written
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_bmapi_write(gfs_inode_t *ip, uint64 offset, uint64 size, gfs_direct_t **direct)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_direct_t *dtmp = NULL;

  int error;


  IFDEBUG(DEBUG_BMAP1,
	  printk(DEBUG_HOW "Entering gfs_bmapi_write()\n"););


  /*  Check for special cases  */

  *direct = NULL;

  if (offset < 0)
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););
    return(-EINVAL);
  }

  if (!size)
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););
    return(0);
  }


  /*  Do we need to unstuff this dinode?  */

  if (ip->i_di.di_height == 0 &&
      offset + size > sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t))
  {
    if (ip->i_di.di_bsize)
    {
      if ((error = gfs_unstuff_dinode(ip)))
      {
	IFDEBUG(DEBUG_BMAP1,
		printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););
	return(error);
      }
    }
    else
      ip->i_di.di_height = 1;
  }


  /*  Make sure our tree is tall enough  */

  if ((error = build_height(ip, offset, size)))
  {
    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););
    return(error);
  }


  if (ip->i_di.di_height == 0)  /*  Stuffed Dinode case  */
  {
    dtmp = (gfs_direct_t *)gmalloc(sizeof(gfs_direct_t));
    if (!dtmp)
    {
      IFDEBUG(DEBUG_BMAP1,
	      printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););
      return(-ENOMEM);
    }

    dtmp->dr_num = 1;
    dtmp->dr_flags = GFS_BMAP_STUFFED_DINODE;
    dtmp->dr_no = 0;
    dtmp->dr_start = sizeof(gfs_dinode_t) + offset;
    dtmp->dr_len = size;

    *direct = dtmp;

    if (offset + size > ip->i_di.di_bsize)
      ip->i_di.di_bsize = offset + size;

    gfs_inode_attr_in(ip);
    gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
    gfs_dwrite(ip->i_didata, 0);

    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););

    return(0);
  }
  else  /*  Regular file  */
  {
    error = metadata(ip, offset, size, direct, TRUE);

    IFDEBUG(DEBUG_BMAP1,
	    printk(DEBUG_HOW "Exiting gfs_bmapi_write()\n"););

    return(error);
  }
}



