/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: dir.c
**
**    Description:
**	This module defines all GFS routines that have to do with directories.
**
**      Implements Extendible Hashing as described in:
**        "Extendible Hashing" by Fagin, et al in
**           __ACM Trans. on Database Systems__, Sept 1979.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfs_debug.h"
#include "blklist.h"
#include "dir.h"
#include "dio.h"
#include "file.h"
#include "inode.h"
#include "gfs_profile.h"
#include "fs.h"

#define DIR_EFFICIENCY_STATS 0

#define IS_LEAF     1
#define IS_DINODE   2

gfs_diroff_t gfs_diroff_head;



/******************************************************************************
*******************************************************************************
**
** int gfs_filecmp()
**
** Description:
**   This routine compares two filenames and returns TRUE if they are equal.
**
** Input(s):
**   file1 - a "struct qstr" from the filesystem
**   file2 - a string assumed to be NULL terminated.
**
** Returns:
**   TRUE (!=0) if the files are the same, otherwise FALSE (0).
**
*******************************************************************************
******************************************************************************/

int
gfs_filecmp (struct qstr *file1, char *file2, int len_of_file2)
{
   IFDEBUG (DEBUG_DIR1,
	    printk (DEBUG_HOW "Entering gfs_filecmp()\n");
      );

   if (file1->len != len_of_file2) {
      IFDEBUG (DEBUG_DIR1,
	       printk (DEBUG_HOW "Exiting gfs_filecmp(ERORR1)\n");
	 );
      return (FALSE);
   }
   if (memcmp (file1->name, file2, file1->len)) {
      IFDEBUG (DEBUG_DIR1,
	       printk (DEBUG_HOW "Exiting gfs_filecmp(ERROR2)\n");
	 );
      return (FALSE);
   }
   IFDEBUG (DEBUG_DIR1,
	    printk (DEBUG_HOW "Exiting gfs_filecmp(OK)\n");
      );

   return (TRUE);
}


static const uint32 crc_32_tab[] =
{
   0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
   0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
   0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
   0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
   0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
   0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
   0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
   0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
   0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
   0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
   0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
   0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
   0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
   0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
   0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
   0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
   0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
   0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
   0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
   0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
   0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
   0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
   0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
   0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
   0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
   0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
   0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
   0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
   0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
   0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
   0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
   0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};


/******************************************************************************
*******************************************************************************
**
** uint32 dir_hash()
**
** Description:
**   Take a filename and convert it to a 32-bit hash.
**
**   The hash function is a 32-bit CRC of the filename.  The algorithm uses
**   the crc_32_tab table above.
**
**   This may not be the fastest hash function, but it does a fair bit better
**   at providing uniform results than the others I've looked at.  That's
**   really important for efficient directories.
**
** Input(s):
**   filename - a "struct qstr" to be hashed
**
** Returns:
**   A 32 bit hash of the function
**
*******************************************************************************
******************************************************************************/

static            uint32
dir_hash (struct qstr *filename)
{
   const char       *name = filename->name;
   int               len = filename->len;

   uint32            hash = 0xFFFFFFFF;

   IFDEBUG (DEBUG_DIR1,
	    printk (DEBUG_HOW "Entering dir_hash()\n");
      );

   for (; len-- && *name; name++)
      hash = crc_32_tab[(hash ^ *name) & 0xFF] ^ (hash >> 8);

   hash = ~hash;

   IFDEBUG (DEBUG_DIR1,
	    printk (DEBUG_HOW "Exiting dir_hash()\n");
      );

   return (hash);
}


/******************************************************************************
*******************************************************************************
**
** gfs_e_first_dirent()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void
gfs_e_first_dirent (buffer_head_t *bh, gfs_dirent_t ** dent)
{
   *dent = (gfs_dirent_t *) (BH_DATA(bh) + sizeof(gfs_leaf_t));
}


/******************************************************************************
*******************************************************************************
**
** gfs_l_first_dirent()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void
gfs_l_first_dirent (buffer_head_t *bh, gfs_dirent_t ** dent)
{
   *dent = (gfs_dirent_t *) (BH_DATA(bh) + sizeof(gfs_dinode_t));
}

/******************************************************************************
*******************************************************************************
**
** gfs_dirent_first()
**
** Description:
**   return first dirent whether bh points to leaf or stuffed dinode
**
** Input(s):
**
** Output(s):
**
** Returns:
**   returns IS_LEAF or IS_DINODE
**   no errors returned; on errors you don't come out of here alive
**
*******************************************************************************
******************************************************************************/

static int
gfs_dirent_first (buffer_head_t *bh, gfs_dirent_t ** dent)
{
   gfs_leaf_t       *leaf;
   gfs_dinode_t     *dinode;


   ENTER(GFS_DIRENT_FIRST, DEBUG_DIR1);

   leaf = (gfs_leaf_t *) BH_DATA(bh);

   if (be32_to_cpu (leaf->lf_magic) == GFS_MAGIC_LF) {
      gfs_e_first_dirent (bh, dent);
      
      EXIT(GFS_DIRENT_FIRST, DEBUG_DIR1);
      return IS_LEAF;
   } else {
      dinode = (gfs_dinode_t *) BH_DATA(bh);
      ASSERT(be32_to_cpu (dinode->di_magic) == GFS_MAGIC_DI);
      gfs_l_first_dirent (bh, dent);

      EXIT(GFS_DIRENT_FIRST, DEBUG_DIR1);
      return IS_DINODE;
   }
}

/******************************************************************************
*******************************************************************************
**
** gfs_dirent_next()
**
** Description: The method for packing dirents is taken straight from ext2
**
** Input(s):
**   bh = leaf data
**   *dent = current dirent * 
**
** Output(s):
**   set *dent
**
** Returns:
**   return 0 on success
**
*******************************************************************************
******************************************************************************/

static int
gfs_dirent_next (buffer_head_t *bh, gfs_dirent_t ** dent)
{
   gfs_dirent_t     *tmp, *cur;
   char             *bh_end;

   ENTER(GFS_DIRENT_NEXT, DEBUG_DIR1);

   cur = *dent;
   bh_end = BH_DATA(bh) + BH_SIZE(bh);

   be32_to_cpus (&cur->de_rec_len);

   if (!((char *) cur + cur->de_rec_len < bh_end)) {
      ASSERT ((char *) cur + cur->de_rec_len == bh_end);
      cpu_to_be32s (&cur->de_rec_len);

      EXIT(GFS_DIRENT_NEXT, DEBUG_DIR1);
      return(-ENOENT);
   }

   tmp = (gfs_dirent_t *) ((char *) cur + cur->de_rec_len);

   ASSERT ((char *) tmp + be32_to_cpu (tmp->de_rec_len) <= bh_end);

   cpu_to_be32s (&cur->de_rec_len);

   ASSERT (tmp->de_ino);	/* only the first dent should ever have de_ino == 0 */

   *dent = tmp;

   EXIT(GFS_DIRENT_NEXT, DEBUG_DIR1);
   return 0;
}


/******************************************************************************
*******************************************************************************
**
** gfs_dirent_del()
**
** Description:
**   deletes a dirent given a pointer to the previous and the dirent
**   that you want to delete
**
** Input(s):
**
** Output(s):
**
** Returns:
**   0 on success
**
*******************************************************************************
******************************************************************************/

static int
gfs_dirent_del (buffer_head_t *bh, gfs_dirent_t * prev, gfs_dirent_t * cur)
{
   char             *bh_end;
   uint32           cur_rec_len;
   uint32           bytes_used;
   gfs_leaf_t       *leaf;

   ENTER(GFS_DIRENT_DEL, DEBUG_DIR1);

   leaf = (gfs_leaf_t *) BH_DATA(bh);
   
   ASSERT(cur->de_ino);

   if (be32_to_cpu(leaf->lf_magic) == GFS_MAGIC_LF) {
     bytes_used = GFS_DIR_REC_LEN(be32_to_cpu(cur->de_name_len));
     ASSERT(be32_to_cpu(leaf->lf_bytes_used) >= bytes_used);
     leaf->lf_bytes_used = cpu_to_be32(be32_to_cpu(leaf->lf_bytes_used) - bytes_used);
   }

   if (prev == NULL) {
      cur->de_ino = 0;		/* hopefully no cpu comes along where 0 isn't 0 */
      EXIT(GFS_DIRENT_DEL, DEBUG_DIR1);
      return 0;
   }

   bh_end = BH_DATA(bh) + BH_SIZE(bh);


   /* there's good reason to leave cur->de_rec_len alone (dir_split_leaf)*/
   /* (in fact, leave it alone) */

   be32_to_cpus (&prev->de_rec_len);
   cur_rec_len = be32_to_cpu(cur->de_rec_len);
   
   ASSERT ((char *) prev + prev->de_rec_len == (char *) cur);
   ASSERT ((char *) cur + cur_rec_len <= bh_end);
   cur->de_ino = 0;
   prev->de_rec_len += cur_rec_len;
   cpu_to_be32s (&prev->de_rec_len);
  
   EXIT(GFS_DIRENT_DEL, DEBUG_DIR1);
   return 0;
}



/******************************************************************************
*******************************************************************************
**
** gfs_dirent_alloc()
**
** Description:
**   allocates a dent for both leaves and stuffed dinodes
**
** Input(s):
**
** Output(s):
**
** Returns:
**   0 on success
**
*******************************************************************************
******************************************************************************/

int
gfs_dirent_alloc (buffer_head_t *bh, int name_len, gfs_dirent_t ** dent_out)
{
   gfs_dirent_t     *dent, *new;
   gfs_leaf_t       *leaf = NULL;
   gfs_dinode_t     *dinode;
   int               rec_len, type;
   int               max_entries = 0, offset = 0, x = 0;

   ENTER(GFS_DIRENT_ALLOC, DEBUG_DIR1);

   if ((type = gfs_dirent_first (bh, &dent)) == IS_LEAF) {
      leaf = (gfs_leaf_t *) BH_DATA(bh);
      max_entries = be16_to_cpu (leaf->lf_entries);
      offset = sizeof(gfs_leaf_t);
   } else if (type == IS_DINODE) {
      dinode = (gfs_dinode_t *) BH_DATA(bh);
      max_entries = be32_to_cpu (dinode->di_entries);
      offset = sizeof(gfs_dinode_t);
   } else {
      ASSERT (0);
   }

   rec_len = GFS_DIR_REC_LEN (name_len);

   if (!max_entries) {
      *dent_out = dent;
      (*dent_out)->de_rec_len = cpu_to_be32 (BH_SIZE(bh) - offset);
      (*dent_out)->de_name_len = cpu_to_be32 (name_len);
      goto gfs_dirent_alloc_success;
   }

   do {
      uint32            cur_rec_len, cur_name_len;

      cur_rec_len = be32_to_cpu (dent->de_rec_len);
      cur_name_len = be32_to_cpu (dent->de_name_len);

      if ((dent->de_ino == 0 && cur_rec_len >= rec_len) ||
	  (cur_rec_len >= GFS_DIR_REC_LEN (cur_name_len) + rec_len)) {

	if (dent->de_ino) {
	  new = (gfs_dirent_t *) ((char *) dent + GFS_DIR_REC_LEN (cur_name_len));
	  new->de_rec_len = cpu_to_be32 (cur_rec_len - GFS_DIR_REC_LEN (cur_name_len));
	  new->de_name_len = cpu_to_be32 (name_len);
	  dent->de_rec_len = cpu_to_be32 (cur_rec_len - be32_to_cpu (new->de_rec_len));
	  *dent_out = new;

	  goto gfs_dirent_alloc_success;
	}

	 dent->de_name_len = cpu_to_be32(name_len);
	 *dent_out = dent;

	 goto gfs_dirent_alloc_success;
      }

      ASSERT (x < max_entries);

      if (dent->de_ino)
	 x++;

   } while (gfs_dirent_next (bh, &dent) == 0);

   EXIT(GFS_DIRENT_ALLOC, DEBUG_DIR1);
   return 1;

 gfs_dirent_alloc_success:

   if (type == IS_LEAF) {
     leaf->lf_bytes_used = cpu_to_be32(be32_to_cpu(leaf->lf_bytes_used) + rec_len);
     ASSERT(be32_to_cpu(leaf->lf_bytes_used) <= BH_SIZE(bh) - sizeof(gfs_leaf_t));
   }

   EXIT(GFS_DIRENT_ALLOC, DEBUG_DIR1);
   return 0;
}


/******************************************************************************
*******************************************************************************
**
** leaf_search()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**   0 on success
**
*******************************************************************************
******************************************************************************/

static int
leaf_search (buffer_head_t *bh, struct qstr *filename,
	     gfs_dirent_t ** dent_out, gfs_dirent_t ** dent_prev)
{
   uint32            hash;
   gfs_dirent_t     *dent, *prev = NULL;
   int               type, max_entries = 0, x = 0;

   ENTER(LEAF_SEARCH, DEBUG_DIR1);

   hash = dir_hash (filename);

   if ((type = gfs_dirent_first (bh, &dent)) == IS_LEAF) {
      gfs_leaf_t       *leaf = (gfs_leaf_t *) BH_DATA(bh);
      max_entries = be16_to_cpu (leaf->lf_entries);
   } else if (type == IS_DINODE) {
      gfs_dinode_t     *dinode = (gfs_dinode_t *) (BH_DATA(bh));
      max_entries = be32_to_cpu (dinode->di_entries);
   } else { ASSERT (0); }

   do {

     if (!dent->de_ino) {
         prev = dent;
	 continue;
     }

     if (be32_to_cpu (dent->de_hash) == hash && gfs_filecmp (filename, dent->de_name, be32_to_cpu(dent->de_name_len))) {
       *dent_out = dent;
       if (dent_prev)
	 *dent_prev = prev;

       EXIT(LEAF_SEARCH, DEBUG_DIR1);
       return 0;
     }

     ASSERT (x < max_entries);
     x++;
     prev = dent;

   } while (gfs_dirent_next (bh, &dent) == 0);

   EXIT(LEAF_SEARCH, DEBUG_DIR1);
   return (-ENOENT);
}


/******************************************************************************
*******************************************************************************
**
** get_leaf()
**
** Description:
**
** Input(s):
**
** Output(s):
**   you must free the buffer_head
**
** Returns:
**   return 0 on success
**
*******************************************************************************
******************************************************************************/

static int
get_leaf (gfs_inode_t * dip, uint64 leaf_no, buffer_head_t **bh_out)
{
   gfs_sbd_t        *sdp = dip->i_sbd;
   buffer_head_t    *bh;

   ENTER(GET_LEAF, DEBUG_DIR1);

   bh = gfs_dread (sdp, leaf_no, dip->i_di.di_glock);
   if (!bh) {
      EXIT(GET_LEAF, DEBUG_DIR1);
      return (-EIO);
   }
   *bh_out = bh;

   EXIT(GET_LEAF, DEBUG_DIR1);
   return 0;
}


/******************************************************************************
*******************************************************************************
**
** get_leaf_nr()
**
** Description: get the leaf # associated with the index
**
** Input(s):
**
** Output(s):
**
** Returns:
**   0 on success
**
*******************************************************************************
******************************************************************************/

static int
get_leaf_nr (gfs_inode_t * dip, uint32 index, uint64 * leaf_out)
{
  uint64            leaf_no;
  int64             error;

  ENTER(GET_LEAF_NR, DEBUG_DIR1);
  ASSERT(dip->i_di.di_regime == GFS_DIR_EXHASH);

  if (dip->i_di.di_height == 0)
    leaf_no = ((uint64 *) (BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t)))[index];
  else {
    error = gfs_readi (dip, (char *) &leaf_no, index * sizeof (uint64), sizeof (uint64), TRUE);
    if (error != sizeof (uint64)) {
      EXIT(GET_LEAF_NR, DEBUG_DIR1);
      return (error);
    }
  }

  *leaf_out = be64_to_cpu (leaf_no);
  EXIT(GET_LEAF_NR, DEBUG_DIR1);
  return 0;
}


/******************************************************************************
*******************************************************************************
**
** get_first_leaf()
**
** Description:
**
** Input(s):
**   index = the hash index
**
** Output(s):
**   you must free the buffer_head
**
** Returns:
**   return 0 on success
**
*******************************************************************************
******************************************************************************/

static int
get_first_leaf (gfs_inode_t * dip, uint32 index, buffer_head_t **bh_out)
{
   uint64            leaf_no;
   int64             error;

   ENTER(GET_FIRST_LEAF, DEBUG_DIR1);
   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   if ( !(error = get_leaf_nr (dip, index, &leaf_no)) )
      error = get_leaf (dip, leaf_no, bh_out);

   EXIT(GET_FIRST_LEAF, DEBUG_DIR1);
   return error;
}


/******************************************************************************
*******************************************************************************
**
** get_next_leaf()
**
** Description:
**
** Input(s):
**   bh_in = the buffer of the current leaf
**
** Output(s):
**   you must free the buffer_head
**
** Returns:
**   return 0 on success
**
*******************************************************************************
******************************************************************************/

static int
get_next_leaf (gfs_inode_t * dip, buffer_head_t *bh_in, buffer_head_t **bh_out)
{
   gfs_leaf_t       *leaf;
   int error;

   ENTER(GET_NEXT_LEAF, DEBUG_DIR1);
   leaf = (gfs_leaf_t *) BH_DATA(bh_in);

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   if (!leaf->lf_next)
     error = -ENOENT;
   else
     error =  get_leaf (dip, be64_to_cpu (leaf->lf_next), bh_out);

   EXIT(GET_NEXT_LEAF, DEBUG_DIR1);
   return error;
}


/******************************************************************************
*******************************************************************************
**
** linked_leaf_search()
**
** Description:
**
** Input(s):
**
** Output(s):
**   return dent and buffer_head (which you must free)
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static int
linked_leaf_search (gfs_inode_t * dip, struct qstr *filename,
		    gfs_dirent_t ** dent_out, gfs_dirent_t ** dent_prev,
		    buffer_head_t **bh_out)
{

   buffer_head_t *bh = NULL, *bh_next;

   uint32            hsize, index;
   uint32            hash;
   int64             error;

   ENTER(LINKED_LEAF_SEARCH, DEBUG_DIR1);
   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   hsize = 1 << dip->i_di.di_depth;
   ASSERT (hsize * sizeof (uint64) == dip->i_di.di_bsize);

   /*  Figure out the address of the leaf node.  */

   hash = dir_hash (filename);
   index = hash >> (32 - dip->i_di.di_depth);
   ASSERT (index < hsize);

   if ((error = get_first_leaf (dip, index, &bh_next)) != 0) {
     EXIT(LINKED_LEAF_SEARCH, DEBUG_DIR1);
     return error;
   }

   /*  Find the entry  */

   do {
      if (bh)
	 gfs_dforget (bh);

      bh = bh_next;

      if ( (error = leaf_search (bh, filename, dent_out, dent_prev)) ) {
	 if (error != (-ENOENT))
	 {
	   gfs_dforget(bh);
	   EXIT(LINKED_LEAF_SEARCH, DEBUG_DIR1);
	   return error;
	 }

      } else {
	 *bh_out = bh;
	 EXIT(LINKED_LEAF_SEARCH, DEBUG_DIR1);
	 return 0;
      }

   } while ((error = get_next_leaf (dip, bh, &bh_next)) == 0);

   gfs_dforget(bh);
   EXIT(LINKED_LEAF_SEARCH, DEBUG_DIR1);
   return error;
}

#if 0

/******************************************************************************
*******************************************************************************
**
** leaf_pack
**
** Description:
**   Make it so only the last dirent has space at the end
**   removes dirent with (de_ino == 0) if there is one
**   Does NOT write buffer_heads
**
** Input(s):
**   bh - buffer_head with dirents
**
** Output(s):
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

static void
leaf_pack(buffer_head_t *bh, gfs_sbd_t  *sdp) {
  gfs_dirent_t *dent, *dent_next, *dent_next2;
  gfs_leaf_t *leaf;
  uint32 bytes_packed = 0;
  /*uint32 actual = 0, used = 0;*/
  
  ENTER(LEAF_PACK, DEBUG_DIR1);

  leaf = (gfs_leaf_t *) BH_DATA(bh);
  ASSERT(leaf->lf_entries);

  /* needs at least one dent to work */

  gfs_dirent_first(bh, &dent);
  dent_next = dent;
  gfs_dirent_next(bh, &dent_next);
  dent_next2 = dent_next;
  gfs_dirent_next(bh, &dent_next2);
 
  while (1) {
    uint32 len_used;

    if (dent->de_ino)
      len_used = GFS_DIR_REC_LEN(be32_to_cpu(dent->de_name_len));
    else 
      len_used = 0;
     

    if ((char *)dent_next > (char *)dent + len_used) {
      uint32 next_len_used = GFS_DIR_REC_LEN(be32_to_cpu(dent_next->de_name_len));
 
      /* do these two statements before the dirent possibly gets written over */

      bytes_packed += be32_to_cpu(dent->de_rec_len) - len_used;
  
      if (dent->de_ino)
	dent->de_rec_len = cpu_to_be32(len_used);
     
      memcpy((char *)dent + len_used, dent_next, next_len_used);
    }

    dent = (gfs_dirent_t *)((char *)dent + len_used);

    if (dent_next == dent_next2) {
      char *bh_end = BH_DATA(bh) + sdp->sd_sb.sb_bsize;

      ASSERT((char *)dent <= bh_end);

      if ((char *)dent == bh_end)
	dent = (gfs_dirent_t *)((char *)dent - len_used);

      dent->de_rec_len = cpu_to_be32(bytes_packed + be32_to_cpu(dent->de_rec_len));

      ASSERT((char *)dent + be32_to_cpu(dent->de_rec_len) == bh_end);

      break;
    }

    dent_next = dent_next2;
    gfs_dirent_next(bh, &dent_next2);
  } 

  EXIT(LEAF_PACK, DEBUG_DIR1);
}

/******************************************************************************
*******************************************************************************
**
** leaf_merge
**
** Description:
**   merge two leaves together, puts result in bh1
**   Does NOT write buffer_heads
**
** Input(s):
**   bh1, bh2 - buffer_heads containing the leaves
**
** Output(s):
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

static void
leaf_merge(buffer_head_t *bh1, buffer_head_t *bh2, gfs_sbd_t  *sdp) {
  gfs_dirent_t *dent, *dent_src;
  gfs_leaf_t * leaf1, *leaf2;
  uint32 len_used, src_len_used;
  uint32 de_rec_len, lf1_bytes_used, lf2_bytes_used;

  ENTER(LEAF_MERGE, DEBUG_DIR1);

  leaf1 = (gfs_leaf_t *) BH_DATA(bh1);
  leaf2 = (gfs_leaf_t *) BH_DATA(bh2);
  
  if (! leaf2->lf_entries) {
    EXIT(LEAF_MERGE, DEBUG_DIR1);
    return;
  }

  else
    leaf_pack(bh2, sdp);

  if (leaf1->lf_entries)
    leaf_pack(bh1, sdp);
  
  gfs_dirent_first(bh2, &dent_src);
  gfs_dirent_first(bh1, &dent);

  while (!gfs_dirent_next(bh1, &dent));
  
  if (dent->de_ino)
    len_used = GFS_DIR_REC_LEN(be32_to_cpu(dent->de_name_len));
  else 
    len_used = 0;

  dent->de_rec_len = cpu_to_be32(len_used);
  dent = (gfs_dirent_t *)((char *)dent + len_used);

  do {
    src_len_used = GFS_DIR_REC_LEN(be32_to_cpu(dent_src->de_name_len));
    memcpy(dent, dent_src,  src_len_used);
    dent = (gfs_dirent_t *)((char *)dent + src_len_used);
  } while (gfs_dirent_next(bh2, &dent_src) == 0);

  dent = (gfs_dirent_t *)((char *)dent - src_len_used);
  de_rec_len = be32_to_cpu(dent->de_rec_len);
  lf1_bytes_used = be32_to_cpu(leaf1->lf_bytes_used);

  ASSERT(de_rec_len - src_len_used >= lf1_bytes_used);

  dent->de_rec_len = cpu_to_be32(de_rec_len - lf1_bytes_used);
  lf2_bytes_used = be32_to_cpu(leaf2->lf_bytes_used);
  leaf1->lf_bytes_used = cpu_to_be32(lf1_bytes_used + lf2_bytes_used);
  leaf1->lf_entries = cpu_to_be16(be16_to_cpu(leaf1->lf_entries) + be16_to_cpu(leaf2->lf_entries));

  ASSERT(leaf1->lf_depth);
  leaf1->lf_depth = cpu_to_be16(be16_to_cpu(leaf1->lf_depth) - 1);

  /* this is an assumption based on how I choose to merge leaves right now */
  /* but it might change */

  ASSERT(be32_to_cpu(leaf1->lf_hash) > be32_to_cpu(leaf2->lf_hash))
  leaf1->lf_hash = leaf2->lf_hash;

  EXIT(LEAF_MERGE, DEBUG_DIR1);
}

#define HASH_PREV 1
#define HASH_NEXT 2

/******************************************************************************
*******************************************************************************
**
** get_adj_leafsize
**
** Description:
**   get adjacent leaf size
**
** Input(s):
**   dip - directory
**   bh  - buffer_head containing leaf
**   side - HASH_PREV or HASH_NEXT
**   size - returns adjacent leaf size defined by the number of hash
**          pointers that point to it
**   leaf_no - returns leaf_no of adjacent block if non-null
**
** Output(s):
**
** Returns:
**   0 on success
**
*******************************************************************************
******************************************************************************/

static int 
get_adj_leafsize(gfs_inode_t *dip, struct buffer_head *bh, int side, uint32 *size, uint64 *leaf_no) {
  gfs_leaf_t *leaf;
  gfs_sbd_t  *sdp = dip->i_sbd;
  int error, same_block = 1, read_another_block = 1;
  uint32 index, low_index;
  uint32 len, adj_len = 1, tmp_index, blocks_read = 0;
  char *buf, *buf_end;
  uint64 *lp, adj_blkno;

  ENTER(GET_ADJ_LEAFSIZE, DEBUG_DIR1);
  ASSERT((side == HASH_PREV) || (side == HASH_NEXT));

  leaf = (gfs_leaf_t *) BH_DATA(bh);
  index = be32_to_cpu(leaf->lf_hash) >> (32 - dip->i_di.di_depth);
  len = 1 << (dip->i_di.di_depth - be16_to_cpu(leaf->lf_depth));

  if (dip->i_di.di_height > 0) {
    if ((buf = gmalloc(sdp->sd_sb.sb_bsize)) == NULL) {
      EXIT(GET_ADJ_LEAFSIZE, DEBUG_DIR1);
      return 1;
    }

    low_index = index & ~(sdp->sd_inptrs - 1);
  
    if (side == HASH_NEXT) {
      if (index - low_index + len >= sdp->sd_inptrs) {
	ASSERT( (1 << dip->i_di.di_depth) - index > len);
	low_index += len;
	same_block = 0;
      }
    }
    else {
      if (low_index == index) {
	ASSERT(low_index >= sdp->sd_inptrs);
	low_index -= sdp->sd_inptrs;
	same_block = 0;
      }
    }

    /* should probably use gfs_bmapi_read to pick up a buffer_head pointer */
    /* since this hash block is almost certainly in memory and a kmalloc   */
    /* and copy seem wasteful */

    error = gfs_readi (dip, buf, low_index * sizeof (uint64), sdp->sd_sb.sb_bsize, TRUE);
   
    if (error != sdp->sd_sb.sb_bsize) {
      gfree(buf, sdp->sd_sb.sb_bsize);
      EXIT(GET_ADJ_LEAFSIZE, DEBUG_DIR1);
      return 1;
    }

    buf_end = buf + sdp->sd_sb.sb_bsize;
  }
  else {
    buf = BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t);
    buf_end = BH_DATA(dip->i_didata) + sdp->sd_sb.sb_bsize;
    low_index = 0;
  }

  lp = (uint64 *) buf;

  /* tmp_index is offset of closest pointer to the adjacent leaf pointers */
  /* that still points to the current leaf */

  if (same_block) {
    if (side == HASH_PREV)
      tmp_index = index - low_index;
    else
      tmp_index = index - low_index + len - 1;
  }
  else {
    if (side == HASH_PREV)
      tmp_index = sdp->sd_inptrs;
    else
      tmp_index = -1;
  }

  if (side == HASH_PREV) {
    adj_blkno = lp[tmp_index - 1]; /* endianness doesn't matter */

    while ((char *)(lp + tmp_index - adj_len) >= buf) {
      if (lp[tmp_index - adj_len] != adj_blkno) {
	read_another_block = 0;
	break;
      }

      adj_len <<= 1;
    }
  }

  else {
    adj_blkno = lp[tmp_index + 1];

    while ((char *)(lp + tmp_index + adj_len) < buf_end) {
      if (lp[tmp_index + adj_len] != adj_blkno) {
	read_another_block = 0;
	break;
      }

      adj_len <<= 1;
    }
  }

  /* can't have leaf pointers start in the middle of a block */
  /* and extend outside of it                                */

  if (read_another_block && same_block)
    read_another_block = 0;

  if (!read_another_block)
    goto get_adj_leafsize_success;
  
  ASSERT(dip->i_di.di_height > 0);

  /* if this code ever gets used, there's probably something wrong */

  while (read_another_block) {
    if (side == HASH_PREV) {
      if (!low_index)
	break;

      low_index -= (1 << blocks_read) * sdp->sd_inptrs;
    }
    else {
      low_index += (1 << blocks_read) * sdp->sd_inptrs;

      if (low_index == (1 << dip->i_di.di_depth))
	break;
    }
    
    error = gfs_readi (dip, buf, low_index * sizeof (uint64), sizeof(uint64), TRUE);

    if (error != sizeof(uint64)) {
      gfree(buf, sdp->sd_sb.sb_bsize);
      EXIT(GET_ADJ_LEAFSIZE, DEBUG_DIR1);
      return 1;
    }

    if (*(uint64 *)buf != adj_blkno)
      break;

    blocks_read++;
    ASSERT ( !(adj_len & 0x80000000));
    adj_len <<= 1;
  }

get_adj_leafsize_success:

  *size = adj_len >> 1;

  if (leaf_no)
    *leaf_no = be64_to_cpu(adj_blkno);

  if (dip->i_di.di_height > 0)
    gfree(buf, sdp->sd_sb.sb_bsize);

  EXIT(GET_ADJ_LEAFSIZE, DEBUG_DIR1);
  return 0;
}

/******************************************************************************
*******************************************************************************
**
** dir_halve_exhash()
**
** Description:
**   Take a wild guess
**
** Input(s):
**   dip - the directory to be converted.
**
** Output(s):
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static void
dir_halve_exhash (gfs_inode_t *dip) {
  gfs_sbd_t        *sdp = dip->i_sbd;
  uint64 *to, *from;
  uint32 block_size, same_depth = 0;
  uint32 nr_blocks, x;
  int64 error;
  uint64 cur_leafno = 0;
  uint32 cur_leaf_ct = 0;
  gfs_blklist_t *blks = NULL;

  char *to_buf, *from_buf; 

  ENTER(DIR_HALVE_EXHASH, DEBUG_DIR1);
  ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

  block_size = sdp->sd_sb.sb_bsize;

  if (1 << dip->i_di.di_depth * sizeof(uint64) <= block_size)
    goto da_end;

  to_buf = gmalloc(block_size);

  if (!to_buf)
    goto da_end;

  from_buf = gmalloc(block_size << 1);

  if (!from_buf)
    goto free_to_buf;

  nr_blocks = (1 << dip->i_di.di_depth * sizeof(uint64) / block_size);

  for (x=0; x < nr_blocks >> 1; x++) {
    char *to_buf_end = to_buf + block_size;
    
    error = gfs_readi(dip, from_buf, (x << 1) * block_size, block_size << 1, TRUE);

    if (error != block_size << 1)
      goto free_from_buf;

    from = (uint64 *) from_buf;
    to = (uint64 *) to_buf;
    /*cur_leafno = *from;*/

    if (!x) {
      cur_leaf_ct = 0;
      cur_leafno = *from;
    }

    while ((char *)to < to_buf_end) {
      *to++ = *from;
      from += 2;
      
      cur_leaf_ct++;

      if (*from != cur_leafno) {
	if (cur_leaf_ct == 1)
	  same_depth++;
	cur_leaf_ct = 0;
	cur_leafno = *from;
      }
	
    }

    error = gfs_writei(dip, to_buf, x * block_size, block_size, TRUE);

    if (error != block_size)
      goto free_from_buf;
  }

  dip->i_di.di_same_depth = same_depth;
  dip->i_di.di_depth--;

  error = gfs_truncate(dip, (1 << dip->i_di.di_depth) * sizeof(uint64), &blks);

  if (!error) {
    if (blks)
      {
	gfs_blkfree(sdp, &blks);
	if (blks)
	  {
	    GARBAGE_BLOCKS();
	    gfs_blklist_free(blks);
	  }
      }
  }
    
 free_from_buf:
  gfree(from_buf, block_size << 1);

 free_to_buf:
  gfree(to_buf, block_size);

 da_end:
  EXIT(DIR_HALVE_EXHASH, DEBUG_DIR1);
}

/******************************************************************************
*******************************************************************************
**
** check_leaf_merge
**
** Description:
**   checks if we can (and want to) merge two leaves and if so, does it.
**   presumably, we have just deleted something. The buffer_head does 
**   not get written out so the caller must write it. Watch out cause the
**   buffer_head you pass in might not be the one you get out.
**
** Input(s):
**   dip - the directory
**   bh  - the leaf we want to check (automatically figures out the adjacent leaf)
**   
**
** Output(s):
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

static void 
check_leaf_merge(gfs_inode_t *dip, struct buffer_head **bhpp) {
  gfs_leaf_t *leaf, *adj_leaf;
  gfs_sbd_t  *sdp = dip->i_sbd;
  uint32 len, index, buf_len, adj_len;
  uint64 adj_leaf_no, error;
  int prev, x;
  char *write_buf;
  buffer_head_t *bh_adj, *dst, *del, *bh;
  gfs_blklist_t *blk;

  ENTER(CHECK_LEAF_MERGE, DEBUG_DIR1);

  bh = *bhpp;
  leaf = (gfs_leaf_t *) BH_DATA(bh);

  if (leaf->lf_next)
    goto clf_end;

  if (! leaf->lf_depth)
    goto clf_end;

  if (be32_to_cpu(leaf->lf_bytes_used) > sdp->tune->b_try_leaf_mrg)
    goto clf_end;

  index = be32_to_cpu(leaf->lf_hash) >> (32 - dip->i_di.di_depth);
 
  len = 1 << (dip->i_di.di_depth - be16_to_cpu(leaf->lf_depth));

  /* when we split a leaf, the new block gets the smaller half of the hash */
  /* so I'm keeping the old block on the assumption that they are closer together */
  /* otherwise expanding and contracting could fragment like hell                 */

  if ((index / len) & 1) {
    error = get_adj_leafsize(dip, bh, HASH_PREV, &adj_len, &adj_leaf_no);
    prev = 1;
  }
  else {
    error = get_adj_leafsize(dip, bh, HASH_NEXT, &adj_len, &adj_leaf_no);
    prev = 0;
  }

  if (error)
    goto clf_end;

  if (adj_len != len)
    goto clf_end;

  if ( (error = get_leaf(dip, adj_leaf_no, &bh_adj)) )
    goto clf_end;

  if (prev == 1) {
    dst = bh;
    del = bh_adj;
  } 
  else {
    dst = bh_adj;
    del = bh;
  }

  adj_leaf = (gfs_leaf_t *) BH_DATA(bh_adj);
 
  if (adj_leaf->lf_next)
    goto free_bh;

  if (be32_to_cpu(adj_leaf->lf_bytes_used) + be32_to_cpu(leaf->lf_bytes_used) > sdp->tune->b_do_leaf_mrg)
    goto free_bh;

  buf_len = 2 * len * sizeof(uint64);
  write_buf = gmalloc(buf_len);
  
  if (! write_buf)
    goto free_bh;

  if ( !(blk = gmalloc(sizeof(gfs_blklist_t) )) )
    goto free_buf;

  for (x=0; x < len << 1; x++)
    ((uint64 *)write_buf)[x] = ((gfs_leaf_t *)BH_DATA(dst))->lf_leafno; /* endianness doesn't matter */

  error = gfs_writei (dip, write_buf, (index &  ~((len << 1) - 1)) * sizeof(uint64), buf_len, TRUE);

  if (error != buf_len)
    goto free_buf;

  leaf_merge(dst, del, sdp);

  blk->blk_no = be64_to_cpu(((gfs_leaf_t *)(BH_DATA(del)))->lf_leafno);
  blk->blk_len = 1;
  blk->blk_next = NULL;

  gfs_blkfree(sdp, &blk);

  if (blk)
    GARBAGE_BLOCKS();
         
  if (be16_to_cpu(leaf->lf_depth) == dip->i_di.di_depth - 1) {
    ASSERT(dip->i_di.di_same_depth >= 2);
    dip->i_di.di_same_depth -= 2;

    if (dip->i_di.di_same_depth == 0)
      dir_halve_exhash(dip);
  }

  dip->i_di.di_blocks--;

  gfs_inode_attr_in (dip);
  gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
  gfs_dwrite (dip->i_didata, 0);

  gfree(write_buf, buf_len);
  gfs_dforget(del);
  *bhpp = dst;

  EXIT(CHECK_LEAF_MERGE, DEBUG_DIR1);
  return;

free_buf:
  gfree(write_buf, buf_len);

free_bh:
  gfs_dforget(bh_adj);

 clf_end:
  EXIT(CHECK_LEAF_MERGE, DEBUG_DIR1);
}
  
#endif

/******************************************************************************
*******************************************************************************
**
** dir_make_exhash()
**
** Description:
**   Convert a stuffed directory into a ExHash directory
**
** Input(s):
**   dip - the directory to be converted.
**
** Output(s):
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int
dir_make_exhash (gfs_inode_t * dip)
{
   gfs_sbd_t        *sdp = dip->i_sbd;
   gfs_blklist_t    *blk;
   gfs_dirent_t     *dent;

   buffer_head_t *bh;
   gfs_leaf_t        *leaf;


   int               y;
   uint32            x;
   uint64           *lp, bn;
   int               error;

   ENTER(DIR_MAKE_EXHASH, DEBUG_DIR1);

   /*  Sanity checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_LINEAR);
   ASSERT (dip->i_di.di_height == 0);

   /*  Allocate a new block for the first leaf node  */

   if ((error = gfs_blkalloc (dip, 1, &blk, 1))) {
      EXIT(DIR_MAKE_EXHASH, DEBUG_DIR1);
      return (error);
   }
   bn = blk->blk_no;

   /*  Turn over a new leaf  */

   bh = gfs_dgetblk (sdp, bn, dip->i_di.di_glock);
   if (!bh) {
      gfs_blkfree (sdp, &blk);
      if (blk) {
	 GARBAGE_BLOCKS ();
	 gfs_blklist_free (blk);
      }
      
      EXIT(DIR_MAKE_EXHASH, DEBUG_DIR1);
      return (-EIO);
   }
   gfs_mark_uptodate (bh, TRUE);
   gfs_blklist_free (blk);
   memset(BH_DATA(bh), 0, sdp->sd_sb.sb_bsize);

   leaf = (gfs_leaf_t *) BH_DATA(bh);
   leaf->lf_magic = cpu_to_be32(GFS_MAGIC_LF);

   ASSERT(dip->i_di.di_entries < (1 << 16));

   leaf->lf_entries = cpu_to_be16(dip->i_di.di_entries);
   leaf->lf_depth = 0;
   leaf->lf_next = 0;
   leaf->lf_hash = 0;
   leaf->lf_leafno = cpu_to_be64(bn);

   x = 0;
   gfs_dirent_first(dip->i_didata, &dent);

   do {
     if (!dent->de_ino)
       continue;

    
     x += GFS_DIR_REC_LEN(be32_to_cpu(dent->de_name_len));
   }  while (gfs_dirent_next(dip->i_didata, &dent) == 0);

   leaf->lf_bytes_used = cpu_to_be32(x);

   ASSERT (sizeof(gfs_leaf_t) <= sizeof(gfs_dinode_t));

   /* copy dirents */

   memcpy (BH_DATA(bh) + sizeof(gfs_leaf_t), BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t), dip->i_di.di_bsize);

   gfs_dirent_first (bh, &dent);

   /* adjust the last dirent's record length */
   /* I suppose this assumes at least two dirents before splitting */

   while (gfs_dirent_next (bh, &dent) == 0) {
      uint32 cur_rec_len, offset;
      char *bh_end;

      bh_end = BH_DATA(bh) + BH_SIZE(bh);
      cur_rec_len = be32_to_cpu(dent->de_rec_len);
      offset = sizeof(gfs_dinode_t) - sizeof(gfs_leaf_t);

      if ((char *)dent + cur_rec_len == bh_end - offset) {
	dent->de_rec_len = cpu_to_be32(cur_rec_len + offset);
	break;
      }
   }

   gfs_dwrite (bh, DW_NOKEEP);

   /*  Set up the hash table  */

   lp = (uint64 *) (BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t));
   memset (lp, 0, sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t));

   for (x = sdp->sd_inptrs / 2; x--; lp++)
      *lp = cpu_to_be64 (bn);

   dip->i_di.di_bsize = sdp->sd_sb.sb_bsize / 2;
   dip->i_di.di_regime = GFS_DIR_EXHASH;
   dip->i_di.di_blocks++;
   dip->i_di.di_same_depth = 0; /* probably zeroed out already */

   for (x = sdp->sd_inptrs / 2, y = -1; x; x >>= 1, y++);
   dip->i_di.di_depth = y;

   gfs_inode_attr_in (dip);
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   EXIT(DIR_MAKE_EXHASH, DEBUG_DIR1);
   return (0);
}


/******************************************************************************
*******************************************************************************
**
** dir_split_leaf()
**
** Description:
**   Split a leaf block into two leaf blocks
**
** Input(s):
**   dip - the directory holding the leaf blocks
**   index - the index of the pointer to the leaf block
**   leaf - the block number of the leaf block
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int
dir_split_leaf (gfs_inode_t * dip, uint32 index, uint64 leaf_no)
{
   gfs_sbd_t        *sdp = dip->i_sbd;
   gfs_blklist_t    *blk;
   buffer_head_t    *nbh, *obh;
   gfs_leaf_t       *nleaf, *oleaf;
   gfs_dirent_t     *dent, *prev;

   uint32            hsize, len, start, divider, moved_bytes_used = 0;
   uint64            bn;
   uint64           *lp;
   
   int               x, old_length, new_length = 0;
   char             *from, *to, *last_dirent = NULL;
   int64             error;

   ENTER(DIR_SPLIT_LEAF, DEBUG_DIR1);

   /*  Sanity Checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   hsize = 1 << dip->i_di.di_depth;

   ASSERT (hsize * sizeof (uint64) == dip->i_di.di_bsize);
   ASSERT (index < hsize);

   /*  Get the new leaf block  */

   if ((error = gfs_blkalloc (dip, 1, &blk, 1))) {
      EXIT(DIR_SPLIT_LEAF, DEBUG_DIR1);
      return (error);
   }
   bn = blk->blk_no;

   nbh = gfs_dgetblk (sdp, bn, dip->i_di.di_glock);
   if (!nbh) {
      error = -EIO;
      goto fail_blkfree;
   }
   gfs_mark_uptodate (nbh, TRUE);
   memset (BH_DATA(nbh), 0, sdp->sd_sb.sb_bsize);

   nleaf = (gfs_leaf_t *) BH_DATA(nbh);
   nleaf->lf_magic = cpu_to_be32(GFS_MAGIC_LF);

   /*  Get the old leaf block  */

   if ((error = get_leaf (dip, leaf_no, &obh)))
      goto fail_nforget;

   oleaf = (gfs_leaf_t *) BH_DATA(obh);

   if (be32_to_cpu(oleaf->lf_magic) != GFS_MAGIC_LF) {
     printk ("GFS:  oleaf->lf_magic (%" FMT_64 "u) != GFS_MAGIC_LF\n", leaf_no);
     ASSERT (0);
   }

   len = 1 << (dip->i_di.di_depth - be16_to_cpu(oleaf->lf_depth));
   ASSERT (len != 1);

   start = (index & ~(len -1));

   /*  Change the pointers  */

   if (dip->i_di.di_height == 0) {
      /*  Find the start of pointers to the leaf.  */

      lp = (uint64 *) (BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t));
     
      /*  Change the pointers  */

      for (x = start; x < start + len / 2; x++)
	 lp[x] = cpu_to_be64 (bn);

      } else {

      lp = (uint64 *) gmalloc (len * sizeof (uint64));

      if (!lp) {
	 error = (-ENOMEM);
	 goto fail_oforget;
      }
      error = gfs_readi (dip, (char *) lp, start * sizeof (uint64), len * sizeof (uint64), TRUE);
      if (error != len * sizeof (uint64))
	 goto fail_vfree;

      /*  Find the start of pointers to the leaf.  */
     
      /*  Change the pointers  */

      for (x = 0; x < len >> 1; x++)
	 lp[x] = cpu_to_be64 (bn);

      error = gfs_writei (dip, (char *) lp, start * sizeof (uint64), len * sizeof (uint64), TRUE);
      if (error != len * sizeof (uint64))
	 goto fail_vfree;

      gfree (lp, len * sizeof(uint64));
      }

   dip->i_di.di_blocks++;

   /*  Copy the entries  */

   divider = (start + len / 2) << (32 - dip->i_di.di_depth);

   IFDEBUG (DEBUG_DIR3,
	    printk (DEBUG_HOW "  Divider = 0x%.8X\n", divider);
      );

   from = BH_DATA(obh) + sizeof(gfs_leaf_t);
   to = BH_DATA(nbh) + sizeof(gfs_leaf_t);

   /*  Note that using y instead of oleaf.lf_entries is necessary because
      oleaf.lf_entries is decremented when a dirent is moved!!!  */

   /*  write out the leaves that belong to the new block */

   dent = (gfs_dirent_t *) (BH_DATA(obh) + sizeof(gfs_leaf_t));
   prev = NULL;
   x = 0;
   old_length = be16_to_cpu(oleaf->lf_entries);

   do {

      if (!dent->de_ino) {
	 ASSERT (!x);
      	 prev = dent;
	 continue;
      }

      if (be32_to_cpu (dent->de_hash) < divider) {
	 uint32 cur_rec_len;

	 IFDEBUG (DEBUG_DIR3,
		  printk (DEBUG_HOW "  0x%.8X, %" FMT_64 "u, %s moving\n",
			  be32_to_cpu(dent->de_hash), be64_to_cpu(dent->de_ino), dent->de_name);
	    );

	 cur_rec_len = be32_to_cpu(dent->de_rec_len);
	 moved_bytes_used += GFS_DIR_REC_LEN(be32_to_cpu(dent->de_name_len));
	 memcpy (to, dent, cur_rec_len);
	 gfs_dirent_del (obh, prev, dent);

	 /* don't want gfs_dirent_next dereferencing a null pointer */
	 
	 if ((char*) dent != BH_DATA(obh) + sizeof(gfs_leaf_t))
	   dent = prev;
	
	 new_length++;

	 last_dirent = to;
	 to += cur_rec_len;
      }

      ASSERT (x < old_length);
      x++;

      prev = dent;
   } while (gfs_dirent_next (obh, &dent) == 0);

   nleaf->lf_bytes_used = cpu_to_be32(moved_bytes_used);
   nleaf->lf_leafno = cpu_to_be64(bn);
   nleaf->lf_hash = oleaf->lf_hash;
   nleaf->lf_entries = cpu_to_be16(new_length);

   oleaf->lf_depth = cpu_to_be16(be16_to_cpu(oleaf->lf_depth) + 1);
   nleaf->lf_depth = oleaf->lf_depth;

   
   oleaf->lf_entries = cpu_to_be16(be16_to_cpu(oleaf->lf_entries) - new_length);
   oleaf->lf_hash = cpu_to_be32( be32_to_cpu(oleaf->lf_hash) | (1 << (32 - be16_to_cpu(oleaf->lf_depth))) );
   
   if (be16_to_cpu(nleaf->lf_depth) == dip->i_di.di_depth)
     dip->i_di.di_same_depth += 2;

   gfs_inode_attr_in (dip);
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   if (last_dirent)
   {
     /* set the record length for last dirent in new leaf */
     dent = (gfs_dirent_t *) last_dirent;
     dent->de_rec_len = cpu_to_be32 (BH_DATA(nbh) + BH_SIZE(nbh) - (char *) dent);
   }

   else
     {
       /* initialize de_rec_len for gfs_dirent_next */

       gfs_leaf_t *leaf = (gfs_leaf_t *) BH_DATA(nbh);

       error = gfs_dirent_alloc(nbh, 0, &dent);
       ASSERT(error == 0);
     
       leaf->lf_bytes_used = 0;  /* undo gfs_dirent_alloc */
       dent->de_ino = 0;
       dent->de_hash = 0;
     }
 
   /*  Write out leaves  */

   gfs_dwrite (obh, DW_NOKEEP);

   gfs_dwrite (nbh, DW_NOKEEP);

   /* free the memory used by the blklist */
   gfs_blklist_free (blk);

   EXIT(DIR_SPLIT_LEAF, DEBUG_DIR1);
   return (0);

 fail_vfree:
   gfree (lp, len * sizeof(uint64));

 fail_oforget:
   gfs_dforget (obh);

 fail_nforget:
   gfs_dforget (nbh);

 fail_blkfree:
   gfs_blkfree (sdp, &blk);
   if (blk) {
      GARBAGE_BLOCKS ();
      gfs_blklist_free (blk);
   }

   EXIT(DIR_SPLIT_LEAF, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** dir_double_exhash()
**
** Description:
**   Double the size of the Exhash table
**
** Input(s):
**   dip - the directory to be doubled
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/
static int
dir_double_exhash (gfs_inode_t * dip)
{
   gfs_sbd_t        *sdp = dip->i_sbd;

   uint32            hsize;
   uint64           *buf;
   uint64           *from, *to;
   uint64            block;
   int               x;
   int64             error = 0;

   ENTER(DIR_DOUBLE_EXHASH, DEBUG_DIR1);
  
   /*  Sanity Checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   hsize = 1 << dip->i_di.di_depth;
   ASSERT (hsize * sizeof (uint64) == dip->i_di.di_bsize);

   if (dip->i_di.di_height == 0) {	/*  Handle the stuffed dinode case  */

      buf = (uint64 *) gmalloc (2 * hsize * sizeof (uint64));
      if (!buf) {
		EXIT(DIR_DOUBLE_EXHASH, DEBUG_DIR1);
		return (-ENOMEM);
      }
      from = (uint64 *) (BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t));
      to = buf;

      for (x = hsize; x--; from++) {
	 *to++ = *from;		/*  No endianess worries  */
	 *to++ = *from;
      }

      error = gfs_writei (dip, (char *) buf, 0, 2 * hsize * sizeof (uint64), TRUE);
      if (error != 2 * hsize * sizeof (uint64)) {
	 gfree (buf, 2 * hsize * sizeof (uint64));
	 EXIT(DIR_DOUBLE_EXHASH, DEBUG_DIR1);
    	 return (error);
      }
      gfree (buf, 2 * hsize * sizeof (uint64));
   } else {			/*  Handle the unstuffed case  */
      /*  Allocate both the "from" and "to" buffers in one big chunk  */

      buf = (uint64 *) gmalloc (3 * sdp->sd_sb.sb_bsize);
      if (!buf) {
		EXIT(DIR_DOUBLE_EXHASH, DEBUG_DIR1);
      	 return (-ENOMEM);
      }

      for (block = dip->i_di.di_bsize / sdp->sd_sb.sb_bsize; block--;) {
	 error = gfs_readi (dip, (char *) buf, block * sdp->sd_sb.sb_bsize, sdp->sd_sb.sb_bsize, TRUE);
	 ASSERT (error == sdp->sd_sb.sb_bsize);		/*  There must be a better way than this (I hope).  */

	 from = buf;
	 to = (uint64 *) ((char *) buf + sdp->sd_sb.sb_bsize);

	 for (x = sdp->sd_inptrs; x--; from++) {
	    *to++ = *from;	/*  No endianess worries  */
	    *to++ = *from;
	 }

	 error = gfs_writei (dip, (char *) buf + sdp->sd_sb.sb_bsize,
	    2 * block * sdp->sd_sb.sb_bsize, 2 * sdp->sd_sb.sb_bsize, TRUE);
	 ASSERT (error == 2 * sdp->sd_sb.sb_bsize);	/*  There must be a better way than this (I hope).  */
      }

      gfree (buf, 3 * sdp->sd_sb.sb_bsize);
   }

   dip->i_di.di_depth++;
   dip->i_di.di_same_depth = 0;
   /*  No gfs_inode_attr_in()  */
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   EXIT(DIR_DOUBLE_EXHASH, DEBUG_DIR1);
   return (0);
}


/******************************************************************************
*******************************************************************************
**
** uint64 dir_e_search()
**
*******************************************************************************
******************************************************************************/

static int
dir_e_search (gfs_inode_t * dip, struct qstr *filename, uint64 * inode)
{
   buffer_head_t *bh;
   gfs_dirent_t     *dent;
   int64             error;

   ENTER(DIR_E_SEARCH, DEBUG_DIR1);
  
   /*  Sanity Checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   if ((error = linked_leaf_search (dip, filename, &dent, NULL, &bh)) != 0) {
      if (inode)
	 *inode = 0;

      EXIT(DIR_E_SEARCH, DEBUG_DIR1);
      return error;
   }
   if (inode)
      *inode = be64_to_cpu (dent->de_ino);
   gfs_dforget (bh);
   EXIT(DIR_E_SEARCH, DEBUG_DIR1);
   return 0;
}

/******************************************************************************
*******************************************************************************
**
** uint64 dir_e_add()
**
*******************************************************************************
******************************************************************************/

static int
dir_e_add (gfs_inode_t * dip, struct qstr *filename, uint64 inode)
{
  gfs_sbd_t        *sdp = dip->i_sbd;
  buffer_head_t *bh, *nbh;
  gfs_leaf_t       *leaf, *nleaf;
  gfs_dirent_t     *dent;
  gfs_blklist_t    *blk;

  uint32            hsize, index;
  uint32            hash;
  uint64            leaf_no;
  int64             error;

  ENTER(DIR_E_ADD, DEBUG_DIR1);
  
 RESTART:

   /*  Sanity Checks  */

  ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

  hsize = 1 << dip->i_di.di_depth;
  ASSERT (hsize * sizeof (uint64) == dip->i_di.di_bsize);

  /*  Figure out the address of the leaf node.  */

  hash = dir_hash (filename);
  index = hash >> (32 - dip->i_di.di_depth);
  ASSERT (index < hsize);

  if ((error = get_leaf_nr (dip, index, &leaf_no))) {
    EXIT(DIR_E_ADD, DEBUG_DIR1);
    return error;
  }

  /*  Add entry to the leaf  */

  while (TRUE) {
    bh = gfs_dread (sdp, leaf_no, dip->i_di.di_glock);

    if (!bh) {
      EXIT(DIR_E_ADD, DEBUG_DIR1);
      return (-EIO);
    }
    leaf = (gfs_leaf_t *) BH_DATA(bh);
    if (be32_to_cpu(leaf->lf_magic) != GFS_MAGIC_LF) {
      printk ("GFS:  leaf.lf_magic (%" FMT_64 "u) != GFS_MAGIC_LF\n", leaf_no);
      ASSERT (0);
    }
    if (gfs_dirent_alloc (bh, filename->len, &dent)) {
      if (be16_to_cpu(leaf->lf_depth) < dip->i_di.di_depth) {	/*  Can we split the leaf?  */
	gfs_dforget (bh);

	if ((error = dir_split_leaf (dip, index, leaf_no))) {
	  EXIT(DIR_E_ADD, DEBUG_DIR1);
	  return (error);
	}
	goto RESTART;
      } else if (dip->i_di.di_depth < GFS_DIR_MAX_DEPTH) {	/*  Can we double the hash table?  */
	gfs_dforget (bh);

	if ((error = dir_double_exhash (dip))) {
	  EXIT(DIR_E_ADD, DEBUG_DIR1);
	  return (error);
	}
	goto RESTART;
      } else if (leaf->lf_next) {	/*  Can we try the next leaf in the list?  */
	leaf_no = be64_to_cpu(leaf->lf_next);
	gfs_dforget (bh);
	continue;
      } else {		/*  Create a new leaf and add it to the list.  */
	
	if ((error = gfs_blkalloc(dip, 1, &blk, 1))) {
	  gfs_dforget (bh);
	  EXIT(DIR_E_ADD, DEBUG_DIR1);
	  return (error);
	}
	nbh = gfs_dgetblk (sdp, blk->blk_no, dip->i_di.di_glock);
	if (!nbh) {
	  gfs_dforget (bh);
	  gfs_blkfree (sdp, &blk);
	  if (blk) {
	    GARBAGE_BLOCKS ();
	    gfs_blklist_free (blk);
	  }
	  EXIT(DIR_E_ADD, DEBUG_DIR1);
	  return (-EIO);
	}
	gfs_mark_uptodate (nbh, TRUE);
	memset (BH_DATA(nbh), 0, sdp->sd_sb.sb_bsize);

	leaf->lf_next = cpu_to_be64(blk->blk_no);

	nleaf = (gfs_leaf_t *) BH_DATA(nbh);

	nleaf->lf_leafno = cpu_to_be64(blk->blk_no);
	nleaf->lf_hash = leaf->lf_hash;
	nleaf->lf_depth = leaf->lf_depth;
	nleaf->lf_magic = cpu_to_be32(GFS_MAGIC_LF);

	gfs_blklist_free (blk);
	
	if (gfs_dirent_alloc(nbh, filename->len, &dent))
	  ASSERT(0);

	dip->i_di.di_blocks++;
	
	gfs_dwrite (bh, DW_NOKEEP);	
	bh = nbh;
	leaf = nleaf;
      }
    }

    dent->de_hash = cpu_to_be32 (hash);
    dent->de_ino = cpu_to_be64 (inode);
    memcpy (dent->de_name, filename->name, filename->len);

    leaf->lf_entries = cpu_to_be16(be16_to_cpu(leaf->lf_entries) + 1);
    
    gfs_dwrite (bh, DW_NOKEEP);

    dip->i_di.di_entries++;
    gfs_inode_attr_in (dip);
    gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
    gfs_dwrite (dip->i_didata, 0);

    EXIT(DIR_E_ADD, DEBUG_DIR1);
    return (0);
  }

  EXIT(DIR_E_ADD, DEBUG_DIR1);  
  return (-ENOENT);

}


/******************************************************************************
*******************************************************************************
**
** uint64 dir_e_del()
**
*******************************************************************************
******************************************************************************/

static int
dir_e_del (gfs_inode_t * dip, struct qstr *filename)
{
   buffer_head_t *bh;
   gfs_leaf_t        leaf;
   gfs_dirent_t     *dent, *prev;

   int64             error;

   ENTER(DIR_E_DEL, DEBUG_DIR1);
  
   /*  Sanity Checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   if ((error = linked_leaf_search (dip, filename, &dent, &prev, &bh))) {
     EXIT(DIR_E_DEL, DEBUG_DIR1);
      return error;
   }

   gfs_dirent_del (bh, prev, dent);
   gfs_leaf_in (&leaf, BH_DATA(bh));
   ASSERT (leaf.lf_entries > 0);
   leaf.lf_entries--;

   gfs_leaf_out (&leaf, BH_DATA(bh));

   /* want to make sure we have the leaf that the hash table points to */
   /* As of now we don't contract a list of linked leaves              */
   /*
   if (dip->i_di.di_depth < GFS_DIR_MAX_DEPTH)
     check_leaf_merge(dip, &bh);
   */

   gfs_dwrite (bh, DW_NOKEEP);

   if (dip->i_di.di_entries == 0)
      printk (KERN_ERR "GFS:  removing a entry from a directory with no entries!\n");
   else
      dip->i_di.di_entries--;

   gfs_inode_attr_in (dip);
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   EXIT(DIR_E_DEL, DEBUG_DIR1);
   return 0;
}

/******************************************************************************
*******************************************************************************
**
** do_filldir()
**
** Description:
**   Does the filldir given a leaf or stuffed dinode
**   six arguments stink. Ideas appreciated. (put in structure ?) 
**
** Input(s):
**   filp, dirent, filldir - exactly the same as for the vfs dir_trans function
**   bh - buffer_head of leaf or stuffed dinode
**   skip - pointer to the calling routine's skip variable
**
** Returns:
**   -EINVAL means buffer is full, 0 means keep on filling
**
*******************************************************************************
******************************************************************************/

static int
do_filldir(gfs_inode_t *dip, 
		   int64 *f_pos,
		   void *dirent,
		   filldir_t filldir,
		   buffer_head_t *bh, 
		   int64 *skip,
		   int index) {
  gfs_dirent_t *dent;
  gfs_leaf_t   *leaf;
  int type, len, error;
  
 
  ENTER(DO_FILLDIR, DEBUG_DIR1);
  type = gfs_dirent_first(bh, &dent);
  
  do {
	
    if (!dent->de_ino)
      continue;
	
    /*  Check to make sure this dirent belongs in this leaf.  */
	
    if (type == IS_LEAF) {
      leaf = (gfs_leaf_t *) BH_DATA(bh);
      len = 1 << (dip->i_di.di_depth - be16_to_cpu(leaf->lf_depth));
	  
      ASSERT ((index <= (be32_to_cpu(dent->de_hash) >> (32 - dip->i_di.di_depth))) &&
			  ((be32_to_cpu(dent->de_hash) >> (32 - dip->i_di.di_depth)) < (index + len)));
    }
	
    if (*skip == 0) {
      if ( (error = filldir (dirent, dent->de_name, be32_to_cpu(dent->de_name_len),
							 *f_pos, be64_to_cpu(dent->de_ino))) ) {
		
		ASSERT(error == -EINVAL);
		EXIT(DO_FILLDIR, DEBUG_DIR1);
      	return error;
      }
	  
      *f_pos += be32_to_cpu(dent->de_rec_len);
    } else {
      *skip -= be32_to_cpu(dent->de_rec_len);
	  
      if (*skip < 0) {
		EXIT(DO_FILLDIR, DEBUG_DIR1);
		return (-EIO);
      }
    }
	
  } while (gfs_dirent_next (bh, &dent) == 0);
  
  EXIT(DO_FILLDIR, DEBUG_DIR1);
  return 0;
}

/******************************************************************************
*******************************************************************************
**
** uint64 dir_e_trans()
**
*******************************************************************************
******************************************************************************/

static int
dir_e_trans (gfs_inode_t *dip, int64 *f_pos, void *dirent, filldir_t filldir){
  gfs_sbd_t        *sdp = dip->i_sbd;
  buffer_head_t    *bh;
  gfs_leaf_t        leaf;
  
  
  uint32            hsize, index, len = 1;
  uint64           *lp;
  uint64            leaf_no;
  uint64            bstart;
  int               bvalid;
  int64             error;
  int64             skip;
  
  ENTER(DIR_E_TRANS, DEBUG_DIR1);
  
  /*  Sanity Checks  */
  
  ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);
  
  hsize = 1 << dip->i_di.di_depth;
  ASSERT (hsize * sizeof (uint64) == dip->i_di.di_bsize);
  
  
  skip = *f_pos;
  
  if (dip->i_di.di_height == 0) {	/*  Stuffed dinode  */
	lp = (uint64 *) (BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t));

      for (index = 0; index < hsize; index += len) {
	 leaf_no = be64_to_cpu (lp[index]);

	 do {

	   if ((error = get_leaf (dip, leaf_no, &bh))) {
	     EXIT(DIR_E_TRANS, DEBUG_DIR1);
	       return error;
	   }

	    gfs_leaf_in (&leaf, BH_DATA(bh));
	    if (leaf.lf_magic != GFS_MAGIC_LF) {
	       printk ("GFS:  leaf.lf_magic (%" FMT_64 "u) != GFS_MAGIC_LF\n", leaf_no);
	       ASSERT (0);
	    }
	    len = 1 << (dip->i_di.di_depth - leaf.lf_depth);
	   
	    if ((error = do_filldir(dip, f_pos, dirent, filldir, bh, &skip, index))) {
		  /* nonzero error should be -EINVAL or -EIO */

	      if (error == (-EINVAL))  
		error = 0;

	      gfs_dforget(bh);
	      EXIT(DIR_E_TRANS, DEBUG_DIR1);
	      return error;
	    }

	    leaf_no = leaf.lf_next;
	    gfs_dforget (bh);
	 } while (leaf_no);
      }
   } else {			/*  Unstuffed dinode  */
      lp = (uint64 *) gmalloc (sdp->sd_sb.sb_bsize);
      if (!lp) {
	 EXIT(DIR_E_TRANS, DEBUG_DIR1);
	 return -ENOMEM;
      }
      bstart = 0;
      bvalid = FALSE;

      for (index = 0; index < hsize; index += len) {

	 /* make sure lp contains the right block data */
	 /* we'll leave it like this for now since calling get_first_leaf would */
	 /* require more calls to gfs_readi. Maybe it will be buffered later on */
	 /* then we could combine stuffed and unstuffed into one case */

	 while (index >= bstart + sdp->sd_inptrs) {
	    bstart += sdp->sd_inptrs;
	    bvalid = FALSE;
	 }

	 if (!bvalid) {
	    error = gfs_readi (dip, (char *) lp, bstart * sizeof (uint64), sdp->sd_sb.sb_bsize, TRUE);
	    if (error != sdp->sd_sb.sb_bsize) {
	       gfree (lp, sdp->sd_sb.sb_bsize);
      	       EXIT(DIR_E_TRANS, DEBUG_DIR1);
	       return ((error < 0) ? error : -EIO);
	    }
	    bvalid = TRUE;
	 }
	 leaf_no = be64_to_cpu (lp[index - bstart]);

	 do {
	  
	    if ((error = get_leaf(dip, leaf_no, &bh))) {
	      gfree(lp, sdp->sd_sb.sb_bsize);
	      EXIT(DIR_E_TRANS, DEBUG_DIR1);
	      return error;
	    }

	    gfs_leaf_in (&leaf, BH_DATA(bh));
	    if (leaf.lf_magic != GFS_MAGIC_LF) {
	       printk ("GFS:  leaf.lf_magic (%" FMT_64 "u) != GFS_MAGIC_LF\n", leaf_no);
	       ASSERT (0);
	    }
	    len = 1 << (dip->i_di.di_depth - leaf.lf_depth);
	
	    if ((error = do_filldir(dip, f_pos, dirent, filldir, bh, &skip, index))) {

	      if (error == -EINVAL)
		error = 0;

	      gfree(lp, sdp->sd_sb.sb_bsize);
	      gfs_dforget(bh);
	      EXIT(DIR_E_TRANS, DEBUG_DIR1);
	      return error;
	    }

	    leaf_no = leaf.lf_next;
	    gfs_dforget (bh);
	 } while (leaf_no);
      }

      gfree (lp, sdp->sd_sb.sb_bsize);
   }

   EXIT(DIR_E_TRANS, DEBUG_DIR1);
   return (0);
}


/******************************************************************************
*******************************************************************************
**
** uint64 dir_e_mvino()
**
*******************************************************************************
******************************************************************************/

static int
dir_e_mvino (gfs_inode_t * dip, struct qstr *filename, uint64 new_inode)
{
   buffer_head_t *bh;
   gfs_dirent_t     *dent;
   int64             error;

   ENTER(DIR_E_MVINO, DEBUG_DIR1);

   /*  Sanity Checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   if ((error = linked_leaf_search (dip, filename, &dent, NULL, &bh))) {
     EXIT(DIR_E_MVINO, DEBUG_DIR1);
      return error;
   }

   dent->de_ino = cpu_to_be64 (new_inode);
   gfs_dwrite (bh, DW_NOKEEP);

   EXIT(DIR_E_MVINO, DEBUG_DIR1);
   return 0;
}


/******************************************************************************
*******************************************************************************
**
** int dir_l_search()
**
*******************************************************************************
******************************************************************************/

static int
dir_l_search (gfs_inode_t * dip, struct qstr *filename, uint64 * inode)
{
  gfs_dirent_t     *dent;
  int64             error;

  ENTER(DIR_L_SEARCH, DEBUG_DIR1);
 
  /*  Sanity checks  */

  ASSERT (dip->i_di.di_regime == GFS_DIR_LINEAR);
  ASSERT (dip->i_di.di_height == 0);

  if ((error = leaf_search (dip->i_didata, filename, &dent, NULL))) {
    if (inode)
      *inode = 0;

    EXIT(DIR_L_SEARCH, DEBUG_DIR1);
    return error;
  }
  if (inode)
    *inode = be64_to_cpu(dent->de_ino);

  EXIT(DIR_L_SEARCH, DEBUG_DIR1);
  return 0;
}


/******************************************************************************
*******************************************************************************
**
** int dir_l_add()
**
*******************************************************************************
******************************************************************************/

static int
dir_l_add (gfs_inode_t * dip, struct qstr *filename, uint64 inode)
{
   gfs_dirent_t     *dent;
   uint32            hash;
   int               error = 0;

   ENTER(DIR_L_ADD, DEBUG_DIR1);
 
   /*  Sanity checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_LINEAR);
   ASSERT (dip->i_di.di_height == 0);

   hash = dir_hash (filename);

   if (gfs_dirent_alloc (dip->i_didata, filename->len, &dent)) {

      error = dir_make_exhash (dip);

      if (error == 0)
	 error = dir_e_add (dip, filename, inode);

      EXIT(DIR_L_ADD, DEBUG_DIR1);
      return (error);

   }
   dent->de_hash = cpu_to_be32 (hash);
   dent->de_ino = cpu_to_be64 (inode);
   memcpy (dent->de_name, filename->name, filename->len);

   dip->i_di.di_entries++;
   dip->i_di.di_mtime = CURRENT_TIME;
   gfs_inode_attr_in (dip);
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   EXIT(DIR_L_ADD, DEBUG_DIR1);
   return (0);
}


/******************************************************************************
*******************************************************************************
**
** int dir_l_del()
**
*******************************************************************************
******************************************************************************/

static int
dir_l_del (gfs_inode_t * dip, struct qstr *filename)
{
   gfs_dirent_t     *dent, *prev;
   int64             error;

   ENTER(DIR_L_DEL, DEBUG_DIR1);

   /*  Sanity checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_LINEAR);
   ASSERT (dip->i_di.di_height == 0);

   if ((error = leaf_search (dip->i_didata, filename, &dent, &prev))) {
     EXIT(DIR_L_DEL, DEBUG_DIR1);
     return error;
   }

   gfs_dirent_del (dip->i_didata, prev, dent);

   if (dip->i_di.di_entries == 0)
      printk (KERN_ERR "GFS:  removing a entry from a directory with no entries!\n");
   else
      dip->i_di.di_entries--;

   dip->i_di.di_mtime = CURRENT_TIME;
   gfs_inode_attr_in (dip);
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   EXIT(DIR_L_DEL, DEBUG_DIR1);
   return (0);
}


/******************************************************************************
*******************************************************************************
**
** int dir_l_trans()
**
*******************************************************************************
******************************************************************************/

static int
dir_l_trans(gfs_inode_t *dip, int64 *f_pos, void *dirent, filldir_t filldir)
{
  /*   gfs_dirent_t dent, *dent_new; */
  int64              skip;
  int                error;

   ENTER(DIR_L_TRANS, DEBUG_DIR1);

   /*  Sanity checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_LINEAR);
   ASSERT (dip->i_di.di_height == 0);

   skip = *f_pos;

   /* do not pass dip->i_data to filldir since it's persistent */

   if ((error = do_filldir(dip, f_pos, dirent, filldir, dip->i_didata, &skip, 0)))
     if (error == -EINVAL)
       error = 0;
 
   EXIT(DIR_L_TRANS, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** int dir_l_mvino()
**
*******************************************************************************
******************************************************************************/

static int
dir_l_mvino (gfs_inode_t * dip, struct qstr *filename, uint64 new_inode)
{
   gfs_dirent_t     *dent;
   int64             error;

   ENTER(DIR_L_MVINO, DEBUG_DIR1);
 
   /*  Sanity checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_LINEAR);
   ASSERT (dip->i_di.di_height == 0);

   if ((error = leaf_search (dip->i_didata, filename, &dent, NULL))) {
     EXIT(DIR_L_MVINO, DEBUG_DIR1);
     return error;
   }

   dent->de_ino = cpu_to_be64 (new_inode);

   dip->i_di.di_mtime = CURRENT_TIME;
   gfs_inode_attr_in (dip);
   gfs_dinode_out (&dip->i_di, BH_DATA(dip->i_didata));
   gfs_dwrite (dip->i_didata, 0);

   EXIT(DIR_L_MVINO, DEBUG_DIR1);
   return (0);
}   


#if DIR_EFFICIENCY_STATS
/******************************************************************************
*******************************************************************************
**
** void dir_stats()
**
** Description:
**   Print out stats about the space efficiency of this directory
**
**   Count the dinode block as part of the directory space only if the
**     directory is in the linear regime
**
** Input(s):  
**   dip - the directory
**
*******************************************************************************
******************************************************************************/

static void
print_stats (gfs_inode_t * dip)
{
   gfs_sbd_t        *sdp = dip->i_sbd;

   printk ("%u / (%" FMT_64 "u * (%u / %u))\n",
	   dip->i_di.di_entries,
	   dip->i_di.di_blocks,
	   sdp->sd_sb.sb_bsize,
	   SIZEOF_DIRENT);
}
#endif /*  DIR_EFFICIENCY_STATS  */


/******************************************************************************
*******************************************************************************
**
** int gfs_dir_search()
**
** Description:
**   This routine searches a directory for a file or another directory.
**     Assumes a glock is held on dip.
**
** Input(s):  
**   dip - Pointer to the inode structure of the directory to which
**              the entry will be added.
**   filename - "struct qstr" containing the name of the file to be added.
**
** Output(s):
**   inode - If not NULL, Inode number of the file entry.
**
** Returns:
**   Inode number if found, -EXXXX on failure.
**
*******************************************************************************
******************************************************************************/

int
gfs_dir_search (gfs_inode_t * dip, struct qstr *filename, uint64 * inode)
{
   int               error = 0;

   ENTER(GFS_DIR_SEARCH, DEBUG_DIR1);
 
   switch (dip->i_di.di_regime) {
   case GFS_DIR_LINEAR:
      error = dir_l_search (dip, filename, inode);

      break;
   case GFS_DIR_EXHASH:
      error = dir_e_search (dip, filename, inode);

      break;
   default:
      panic ("GFS:  Unknown directory regime:  %d\n", dip->i_di.di_regime);

      break;
   }

   EXIT(GFS_DIR_SEARCH, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_dir_add()
**
** Description:
**   This routine adds a new file name entry to a directory.
**     Assumes a glock is held on dip.
**
** Input(s):
**   dip - Pointer to the inode structure of the directory to which 
**           the entry will be added.
**   filename - String containing the name of the file to be added.
**   ino - Inode number of the file entry.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int
gfs_dir_add (gfs_inode_t * dip, struct qstr *filename, uint64 inode)
{
   int               error = 0;

   ENTER(GFS_DIR_ADD, DEBUG_DIR1);
  
   switch (dip->i_di.di_regime) {
   case GFS_DIR_LINEAR:
      error = dir_l_add (dip, filename, inode);

      break;
   case GFS_DIR_EXHASH:
      error = dir_e_add (dip, filename, inode);

      break;
   default:
      panic ("GFS:  Unknown directory regime:  %d\n", dip->i_di.di_regime);

      break;
   }

#if DIR_EFFICIENCY_STATS
   print_stats (dip);
#endif /*  DIR_EFFICIENCY_STATS  */

   EXIT(GFS_DIR_ADD, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_dir_del()
**
** Description:
**   This routine deletes a file name entry from a directory.
**     Assumes a glock is held on dip.
**
** Input(s):
**   dip - Pointer to the directory's inode structure.
**   filename - Name of the directory entry to be removed.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int
gfs_dir_del (gfs_inode_t * dip, struct qstr *filename)
{
   int               error = 0;

   ENTER(GFS_DIR_DEL, DEBUG_DIR1);

   switch (dip->i_di.di_regime) {
   case GFS_DIR_LINEAR:
      error = dir_l_del (dip, filename);

      break;
   case GFS_DIR_EXHASH:
      error = dir_e_del (dip, filename);

      break;
   default:
      panic ("GFS:  Unknown directory regime:  %d\n", dip->i_di.di_regime);

      break;
   }

#if DIR_EFFICIENCY_STATS
   print_stats (dip);
#endif /*  DIR_EFFICIENCY_STATS  */

   EXIT(GFS_DIR_DEL, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_dir_trans()
**
** Description:
**   This routine translates an GFS file name entry to an external
**     representation.
**
** Input(s):
**   filp - the file pointer to be read (contains the inode and offset)
**   dirent - pointer to the space to be filled
**   filldir - routine passed in to do the filling
**
** Returns:
**   0 on success, -EXXXX on success
**
*******************************************************************************
******************************************************************************/

int gfs_dir_trans(gfs_inode_t *dip, int64 *f_pos, void *dirent, filldir_t filldir)
{
   int               error = 0;

   ENTER(GFS_DIR_TRANS, DEBUG_DIR1);

   switch (dip->i_di.di_regime) {
   case GFS_DIR_LINEAR:
      error = dir_l_trans (dip, f_pos, dirent, filldir);

      break;
   case GFS_DIR_EXHASH:
      error = dir_e_trans (dip, f_pos, dirent, filldir);

      break;
   default:
      panic ("GFS:  Unknown directory regime:  %d\n", dip->i_di.di_regime);

      break;
   }

#if DIR_EFFICIENCY_STATS
   print_stats (dip);
#endif /*  DIR_EFFICIENCY_STATS  */

   EXIT(GFS_DIR_TRANS, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_dir_mvino()
**
** Description:
**   This routine changes the inode number of a directory entry.  It's used
**     by rename to change ".." when a directory is moved.
**     Assumes a glock is held on dvp.
**
** Input(s):
**   dip - directory holding the file
**   filename - the name of the file
**   new_inode - the new inode value
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int
gfs_dir_mvino (gfs_inode_t * dip, struct qstr *filename, uint64 new_inode)
{
   int               error = 0;

   ENTER(GFS_DIR_MVINO, DEBUG_DIR1);
 
   switch (dip->i_di.di_regime) {
   case GFS_DIR_LINEAR:
      error = dir_l_mvino (dip, filename, new_inode);

      break;
   case GFS_DIR_EXHASH:
      error = dir_e_mvino (dip, filename, new_inode);

      break;
   default:
      panic ("GFS:  Unknown directory regime:  %d\n", dip->i_di.di_regime);

      break;
   }

   EXIT(GFS_DIR_MVINO, DEBUG_DIR1);
   return (error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dir_free()
**
** Description:
**   Frees all leaf blocks associated with a directory
**
** Input(s):
**   dip - the directory to be freed
**
** Returns:
**   The list of blocks to be freed, NULL on failure.
**
*******************************************************************************
******************************************************************************/

gfs_blklist_t    *
gfs_dir_free (gfs_inode_t * dip)
{
   gfs_sbd_t        *sdp = dip->i_sbd;
   buffer_head_t *bh;
   gfs_leaf_t        leaf;
   gfs_blklist_t    *bhead = NULL, *btmp;

   uint32            hsize, index, len = 1;
   uint64           *lp;
   uint64            leaf_no;
   uint64            bstart;
   int               bvalid;
   int64             error;

   ENTER(GFS_DIR_FREE, DEBUG_DIR1);
 
   /*  If this is a linear directory, we don't have to free blocks  */

   if (dip->i_di.di_regime == GFS_DIR_LINEAR) {
      EXIT(GFS_DIR_FREE, DEBUG_DIR1);
     return (NULL);
   }
   /*  Sanity Checks  */

   ASSERT (dip->i_di.di_regime == GFS_DIR_EXHASH);

   hsize = 1 << dip->i_di.di_depth;
   ASSERT (hsize * sizeof (uint64) == dip->i_di.di_bsize);

   if (dip->i_di.di_height == 0) {	/*  Stuffed dinode  */
      lp = (uint64 *) (BH_DATA(dip->i_didata) + sizeof(gfs_dinode_t));

      for (index = 0; index < hsize; index += len) {
	 leaf_no = be64_to_cpu (lp[index]);

	 do {
	    bh = gfs_dread (sdp, leaf_no, dip->i_di.di_glock);
	    if (!bh) {
	       GARBAGE_BLOCKS ();
	       break;
	    }
	    gfs_leaf_in (&leaf, BH_DATA(bh));
	    if (leaf.lf_magic != GFS_MAGIC_LF) {
	       printk ("GFS:  leaf.lf_magic (%" FMT_64 "u) != GFS_MAGIC_LF\n", leaf_no);
	       ASSERT (0);
	    }
	    len = 1 << (dip->i_di.di_depth - leaf.lf_depth);

	    if (bhead && bhead->blk_no + bhead->blk_len == leaf_no)
	       bhead->blk_len++;
	    else {
	       btmp = (gfs_blklist_t *) gmalloc (sizeof (gfs_blklist_t));
	       if (btmp) {
		  btmp->blk_no = leaf_no;
		  btmp->blk_len = 1;
		  btmp->blk_next = bhead;
		  bhead = btmp;
	       } else
		  GARBAGE_BLOCKS ();
	    }

	    leaf_no = leaf.lf_next;
	    gfs_dforget (bh);
	 } while (leaf_no);
      }
   } else {			/*  Unstuffed dinode  */
      lp = (uint64 *) gmalloc (sdp->sd_sb.sb_bsize);
      if (!lp) {
	 GARBAGE_BLOCKS ();
      	 EXIT(GFS_DIR_FREE, DEBUG_DIR1);
	 return (NULL);
      }
      bstart = 0;
      bvalid = FALSE;

      for (index = 0; index < hsize; index += len) {
	 while (index >= bstart + sdp->sd_inptrs) {
	    bstart += sdp->sd_inptrs;
	    bvalid = FALSE;
	 }

	 if (!bvalid) {
	    error = gfs_readi (dip, (char *) lp, bstart * sizeof (uint64), sdp->sd_sb.sb_bsize, TRUE);
	    if (error != sdp->sd_sb.sb_bsize) {
	       GARBAGE_BLOCKS ();
	       continue;
	    }
	    bvalid = TRUE;
	 }
	 leaf_no = be64_to_cpu (lp[index - bstart]);

	 do {
	    bh = gfs_dread (sdp, leaf_no, dip->i_di.di_glock);
	    if (!bh) {
	       GARBAGE_BLOCKS ();
	       break;
	    }
	    gfs_leaf_in (&leaf, BH_DATA(bh));
	    if (leaf.lf_magic != GFS_MAGIC_LF) {
	       printk ("GFS:  leaf.lf_magic (%" FMT_64 "u) != GFS_MAGIC_LF\n", leaf_no);
	       ASSERT (0);
	    }
	    len = 1 << (dip->i_di.di_depth - leaf.lf_depth);

	    if (bhead && bhead->blk_no + bhead->blk_len == leaf_no)
	       bhead->blk_len++;
	    else {
	       btmp = (gfs_blklist_t *) gmalloc (sizeof (gfs_blklist_t));
	       if (btmp) {
		  btmp->blk_no = leaf_no;
		  btmp->blk_len = 1;
		  btmp->blk_next = bhead;
		  bhead = btmp;
	       } else
		  GARBAGE_BLOCKS ();
	    }

	    leaf_no = leaf.lf_next;
	    gfs_dforget (bh);
	 } while (leaf_no);
      }

      gfree (lp, sdp->sd_sb.sb_bsize);
   }

   EXIT(GFS_DIR_FREE, DEBUG_DIR1);
   return (bhead);
}


