/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: ondisk.c
**
**    Description:  This module converts all GFS structures that exist on
**      disk to internal processor formats.
**
*******************************************************************************
******************************************************************************/

#ifdef __KERNEL__

#include "gfs.h"
#include "gfs_debug.h"

#else  /*  __KERNEL__  */

#include "gfs_endian.h"
#include <string.h>
#undef IFDEBUG
#define IFDEBUG(x,y)

#endif  /*  __KERNEL__  */

#include "gfs_ondisk.h"



#define CPIN_8(x, y, z) {memcpy((x), (y), (z));}
#define CPOUT_8(x, y, z) {memcpy((y), (x), (z));}
#define CPIN_16(x, y) {(x) = be16_to_cpu((y));}
#define CPOUT_16(x, y) {(y) = cpu_to_be16((x));}
#define CPIN_32(x, y) {(x) = be32_to_cpu((y));}
#define CPOUT_32(x, y) {(y) = cpu_to_be32((x));}
#define CPIN_64(x, y) {(x) = be64_to_cpu((y));}
#define CPOUT_64(x, y) {(y) = cpu_to_be64((x));}



/******************************************************************************
*******************************************************************************
**
** gfs_sb_in()
**
** Description:
**   Copy the super block in from a disk buffer
**
** Input(s):
**   buf - a disk buffer
**
** Output(s):
**   sb - a superblock structure
**
*******************************************************************************
******************************************************************************/

void gfs_sb_in(gfs_sb_t *sb, char *buf)
{
  gfs_sb_t *bufsb = (gfs_sb_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_sb_in()\n"););

  CPIN_32(sb->sb_magic, bufsb->sb_magic);
  CPIN_32(sb->sb_version, bufsb->sb_version);

  CPIN_32(sb->sb_bsize, bufsb->sb_bsize);
  CPIN_32(sb->sb_length, bufsb->sb_length);
  CPIN_64(sb->sb_rootdi, bufsb->sb_rootdi);
  CPIN_32(sb->sb_rgrps, bufsb->sb_rgrps);
  
  CPIN_8(sb->sb_name, bufsb->sb_name, 12);

  CPIN_64(sb->sb_fssize, bufsb->sb_fssize);

  CPIN_64(sb->sb_glocks, bufsb->sb_glocks);
  CPIN_8(sb->sb_lockproto, bufsb->sb_lockproto, GFS_LOCKNAME_LEN + 1);
  CPIN_8(sb->sb_locktable, bufsb->sb_locktable, GFS_LOCKNAME_LEN + 1);

  CPIN_16(sb->sb_sbsize, bufsb->sb_sbsize);
  CPIN_16(sb->sb_risize, bufsb->sb_risize);
  CPIN_16(sb->sb_rgsize, bufsb->sb_rgsize);
  CPIN_16(sb->sb_disize, bufsb->sb_disize);
  CPIN_16(sb->sb_desize, bufsb->sb_desize);
  CPIN_16(sb->sb_lfsize, bufsb->sb_lfsize);

  CPIN_32(sb->sb_mounts, bufsb->sb_mounts);
  CPIN_8(sb->sb_mountids, bufsb->sb_mountids, GFS_MOUNT_BITS);
  CPIN_32(sb->sb_time, bufsb->sb_time);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_sb_in()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_sb_out()
**
** Description:
**   Copy the super block out to a disk buffer
**
** Input(s):
**   sb - a superblock structure
**
** Output(s):
**   buf - a disk buffer
**
*******************************************************************************
******************************************************************************/

void gfs_sb_out(gfs_sb_t *sb, char *buf)
{
  gfs_sb_t *bufsb = (gfs_sb_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_sb_out()\n"););

  CPOUT_32(sb->sb_magic, bufsb->sb_magic);
  CPOUT_32(sb->sb_version, bufsb->sb_version);

  CPOUT_32(sb->sb_bsize, bufsb->sb_bsize);
  CPOUT_32(sb->sb_length, bufsb->sb_length);
  CPOUT_64(sb->sb_rootdi, bufsb->sb_rootdi);
  CPOUT_32(sb->sb_rgrps, bufsb->sb_rgrps);
  
  CPOUT_8(sb->sb_name, bufsb->sb_name, 12);

  CPOUT_64(sb->sb_fssize, bufsb->sb_fssize);

  CPOUT_64(sb->sb_glocks, bufsb->sb_glocks);
  CPOUT_8(sb->sb_lockproto, bufsb->sb_lockproto, GFS_LOCKNAME_LEN + 1);
  CPOUT_8(sb->sb_locktable, bufsb->sb_locktable, GFS_LOCKNAME_LEN + 1);

  CPOUT_16(sb->sb_sbsize, bufsb->sb_sbsize);
  CPOUT_16(sb->sb_risize, bufsb->sb_risize);
  CPOUT_16(sb->sb_rgsize, bufsb->sb_rgsize);
  CPOUT_16(sb->sb_disize, bufsb->sb_disize);
  CPOUT_16(sb->sb_desize, bufsb->sb_desize);
  CPOUT_16(sb->sb_lfsize, bufsb->sb_lfsize);

  CPOUT_32(sb->sb_mounts, bufsb->sb_mounts);
  CPOUT_8(sb->sb_mountids, bufsb->sb_mountids, GFS_MOUNT_BITS);
  CPOUT_32(sb->sb_time, bufsb->sb_time);
  bufsb->sb_pad = 0;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_sb_out()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_rindex_in()
**
** Description:
**   Copy a resource index in from a disk buffer
**
** Input(s):
**   buf - a disk buffer
**
** Output(s):
**   rindex - a resource index structure
**
*******************************************************************************
******************************************************************************/

void gfs_rindex_in(gfs_rindex_t *rindex, char *buf)
{
  gfs_rindex_t *bufri = (gfs_rindex_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_rindex_in()\n"););

  CPIN_64(rindex->ri_addr, bufri->ri_addr);
  CPIN_32(rindex->ri_length, bufri->ri_length);
  CPIN_32(rindex->ri_subpool, bufri->ri_subpool);
  CPIN_64(rindex->ri_glock, bufri->ri_glock);

  CPIN_64(rindex->ri_data1, bufri->ri_data1);
  CPIN_64(rindex->ri_data, bufri->ri_data);

  CPIN_64(rindex->ri_glock1, bufri->ri_glock1);
  CPIN_64(rindex->ri_glocks, bufri->ri_glocks);

  CPIN_32(rindex->ri_bitbytes, bufri->ri_bitbytes);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_rindex_in()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_rindex_out()
**
** Description:
**   Copy a resource index out to a disk buffer
**
** Input(s):
**   rindex - a resource index structure
**
** Output(s):
**   buf - a disk buffer
**
*******************************************************************************
******************************************************************************/

void gfs_rindex_out(gfs_rindex_t *rindex, char *buf)
{
  gfs_rindex_t *bufri = (gfs_rindex_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_rindex_out()\n"););

  CPOUT_64(rindex->ri_addr, bufri->ri_addr);
  CPOUT_32(rindex->ri_length, bufri->ri_length);
  CPOUT_32(rindex->ri_subpool, bufri->ri_subpool);
  CPOUT_64(rindex->ri_glock, bufri->ri_glock);

  CPOUT_64(rindex->ri_data1, bufri->ri_data1);
  CPOUT_64(rindex->ri_data, bufri->ri_data);

  CPOUT_64(rindex->ri_glock1, bufri->ri_glock1);
  CPOUT_64(rindex->ri_glocks, bufri->ri_glocks);

  CPOUT_32(rindex->ri_bitbytes, bufri->ri_bitbytes);
  bufri->ri_pad = 0;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_rindex_out()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_rgrp_in()
**
** Description:
**   Copy a resource group header in from a disk buffer
**
** Input(s):
**   buf - a disk buffer
**
** Output(s):
**   rgrp - a resource group header
**
*******************************************************************************
******************************************************************************/

void gfs_rgrp_in(gfs_rgrp_t *rgrp, char *buf)
{
  gfs_rgrp_t *bufrg = (gfs_rgrp_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_rgrp_in()\n"););

  CPIN_32(rgrp->rg_magic, bufrg->rg_magic);
  CPIN_32(rgrp->rg_time, bufrg->rg_time);
  CPIN_64(rgrp->rg_free, bufrg->rg_free);
  CPIN_64(rgrp->rg_dinodes, bufrg->rg_dinodes);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_rgrp_in()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_rgrp_out()
**
** Description:
**   Copy a resource group header out to a disk buffer
**
** Input(s):
**   rgrp - a resource group header
**
** Output(s):
**   buf - a disk buffer
**
*******************************************************************************
******************************************************************************/

void gfs_rgrp_out(gfs_rgrp_t *rgrp, char *buf)
{
  gfs_rgrp_t *bufrg = (gfs_rgrp_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_rgrp_out()\n"););

  CPOUT_32(rgrp->rg_magic, bufrg->rg_magic);
  CPOUT_32(rgrp->rg_time, bufrg->rg_time);
  CPOUT_64(rgrp->rg_free, bufrg->rg_free);
  CPOUT_64(rgrp->rg_dinodes, bufrg->rg_dinodes); 

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_rgrp_out()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_dinode_in()
**
** Description:
**   Copy a dinode in from a disk buffer
**
** Input(s):
**   buf - a disk buffer
**
** Output(s):
**   dinode - a dinode structure
**
*******************************************************************************
******************************************************************************/

void gfs_dinode_in(gfs_dinode_t *dinode, char *buf)
{
  gfs_dinode_t *bufdi = (gfs_dinode_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_dinode_in()\n"););

  CPIN_32(dinode->di_magic, bufdi->di_magic);

  CPIN_64(dinode->di_addr, bufdi->di_addr);

  CPIN_32(dinode->di_mode, bufdi->di_mode);
  CPIN_32(dinode->di_uid, bufdi->di_uid);
  CPIN_32(dinode->di_gid, bufdi->di_gid);
  CPIN_32(dinode->di_nlink, bufdi->di_nlink);
  CPIN_64(dinode->di_bsize, bufdi->di_bsize);
  CPIN_64(dinode->di_blocks, bufdi->di_blocks);
  CPIN_32(dinode->di_atime, bufdi->di_atime);
  CPIN_32(dinode->di_mtime, bufdi->di_mtime);
  CPIN_32(dinode->di_ctime, bufdi->di_ctime);
  CPIN_32(dinode->di_major, bufdi->di_major);
  CPIN_32(dinode->di_minor, bufdi->di_minor);

  CPIN_32(dinode->di_rgrp, bufdi->di_rgrp);
  CPIN_32(dinode->di_rgrp_meta, bufdi->di_rgrp_meta);
  CPIN_32(dinode->di_rgrp_data, bufdi->di_rgrp_data);
  CPIN_64(dinode->di_last_meta, bufdi->di_last_meta);
  CPIN_64(dinode->di_last_data, bufdi->di_last_data);
  CPIN_64(dinode->di_glock, bufdi->di_glock);
  CPIN_16(dinode->di_type, bufdi->di_type);
  CPIN_16(dinode->di_height, bufdi->di_height);
  CPIN_16(dinode->di_stamp, bufdi->di_stamp);

  CPIN_16(dinode->di_nopen, bufdi->di_nopen);
  CPIN_8(dinode->di_open, bufdi->di_open, GFS_MOUNT_BITS);

  CPIN_16(dinode->di_regime, bufdi->di_regime);
  CPIN_16(dinode->di_depth, bufdi->di_depth);
  CPIN_32(dinode->di_entries, bufdi->di_entries);

  CPIN_32(dinode->di_same_depth, bufdi->di_same_depth);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_dinode_in()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_dinode_out()
**
** Description:
**   Copy a dinode out to a disk buffer
**
** Input(s):
**   dinode - a dinode structure
**
** Output(s):
**   buf - a disk buffer
**
*******************************************************************************
******************************************************************************/

void gfs_dinode_out(gfs_dinode_t *dinode, char *buf)
{
  gfs_dinode_t *bufdi = (gfs_dinode_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_dinode_out()\n"););

  CPOUT_32(dinode->di_magic, bufdi->di_magic);
  bufdi->di_pad = 0;

  CPOUT_64(dinode->di_addr, bufdi->di_addr);

  CPOUT_32(dinode->di_mode, bufdi->di_mode);
  CPOUT_32(dinode->di_uid, bufdi->di_uid);
  CPOUT_32(dinode->di_gid, bufdi->di_gid);
  CPOUT_32(dinode->di_nlink, bufdi->di_nlink);
  CPOUT_64(dinode->di_bsize, bufdi->di_bsize);
  CPOUT_64(dinode->di_blocks, bufdi->di_blocks);
  CPOUT_32(dinode->di_atime, bufdi->di_atime);
  CPOUT_32(dinode->di_mtime, bufdi->di_mtime);
  CPOUT_32(dinode->di_ctime, bufdi->di_ctime);
  CPOUT_32(dinode->di_major, bufdi->di_major);
  CPOUT_32(dinode->di_minor, bufdi->di_minor);

  CPOUT_32(dinode->di_rgrp, bufdi->di_rgrp);
  CPOUT_32(dinode->di_rgrp_meta, bufdi->di_rgrp_meta);
  CPOUT_32(dinode->di_rgrp_data, bufdi->di_rgrp_data);
  CPOUT_64(dinode->di_last_meta, bufdi->di_last_meta);
  CPOUT_64(dinode->di_last_data, bufdi->di_last_data);
  CPOUT_64(dinode->di_glock, bufdi->di_glock);
  CPOUT_16(dinode->di_type, bufdi->di_type);
  CPOUT_16(dinode->di_height, bufdi->di_height);
  CPOUT_16(dinode->di_stamp, bufdi->di_stamp);

  CPOUT_16(dinode->di_nopen, bufdi->di_nopen);
  CPOUT_8(dinode->di_open, bufdi->di_open, GFS_MOUNT_BITS);

  CPOUT_16(dinode->di_regime, bufdi->di_regime);
  CPOUT_16(dinode->di_depth, bufdi->di_depth);
  CPOUT_32(dinode->di_entries, bufdi->di_entries);

  CPOUT_32(dinode->di_same_depth, bufdi->di_same_depth);
  bufdi->di_pad1 = 0;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_dinode_out()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_dirent_in()
**
** Description:
**   Copy a directory entry in from a disk buffer
**
** Input(s):
**   buf - a disk buffer
**
** Output(s):
**   dirent - a directory entry structure
**
*******************************************************************************
******************************************************************************/

void gfs_dirent_in(gfs_dirent_t *dirent, char *buf)
{
  gfs_dirent_t *bufde = (gfs_dirent_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_dirent_in()\n"););

  CPIN_64(dirent->de_ino, bufde->de_ino);
  CPIN_32(dirent->de_hash, bufde->de_hash);
  CPIN_32(dirent->de_name_len, bufde->de_name_len);
  CPIN_32(dirent->de_rec_len, bufde->de_rec_len);

  CPIN_8(dirent->de_name, bufde->de_name, dirent->de_name_len);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_dirent_in()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_dirent_out()
**
** Description:
**   Copy a directory entry out to a disk buffer
**
** Input(s):
**   dirent - a directory entry structure
**
** Output(s):
**   buf - a disk buffer
**
*******************************************************************************
******************************************************************************/

void gfs_dirent_out(gfs_dirent_t *dirent, char *buf)
{
  gfs_dirent_t *bufde = (gfs_dirent_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_dirent_out()\n"););

  CPOUT_64(dirent->de_ino, bufde->de_ino);
  CPOUT_32(dirent->de_hash, bufde->de_hash);
  CPOUT_32(dirent->de_name_len, bufde->de_name_len);
  CPOUT_32(dirent->de_rec_len, bufde->de_rec_len);

  bufde->de_pad = 0;

  CPOUT_8(dirent->de_name, bufde->de_name, dirent->de_name_len);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_dirent_out()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_leaf_in()
**
** Description:
**   Copy a directory leaf header in from a disk buffer
**
** Input(s):
**   buf - a disk buffer
**
** Output(s):
**   leaf - a directory leaf header
**
*******************************************************************************
******************************************************************************/

void gfs_leaf_in(gfs_leaf_t *leaf, char *buf)
{
  gfs_leaf_t *buflf = (gfs_leaf_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_leaf_in()\n"););

  CPIN_32(leaf->lf_magic, buflf->lf_magic);

  CPIN_16(leaf->lf_entries, buflf->lf_entries);
  CPIN_16(leaf->lf_depth, buflf->lf_depth);
  CPIN_64(leaf->lf_next, buflf->lf_next);
  CPIN_32(leaf->lf_bytes_used, buflf->lf_bytes_used);
  CPIN_32(leaf->lf_hash, buflf->lf_hash);
  CPIN_64(leaf->lf_leafno, buflf->lf_leafno);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_leaf_in()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_leaf_out()
**
** Description:
**   Copy a directory leaf header out to a disk buffer
**
** Input(s):
**   leaf - a directory leaf header
**
** Output(s):
**   buf - a disk buffer
**
*******************************************************************************
******************************************************************************/

void gfs_leaf_out(gfs_leaf_t *leaf, char *buf)
{
  gfs_leaf_t *buflf = (gfs_leaf_t *)buf;

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Entering gfs_leaf_out()\n"););

  CPOUT_32(leaf->lf_magic, buflf->lf_magic);

  CPOUT_16(leaf->lf_entries, buflf->lf_entries);
  CPOUT_16(leaf->lf_depth, buflf->lf_depth);
  CPOUT_64(leaf->lf_next, buflf->lf_next);
  CPOUT_32(leaf->lf_bytes_used, buflf->lf_bytes_used);
  CPOUT_32(leaf->lf_hash, buflf->lf_hash);
  CPOUT_64(leaf->lf_leafno, buflf->lf_leafno);

  IFDEBUG(DEBUG_ONDISK1,
	  printk(DEBUG_HOW "Exiting gfs_leaf_out()\n"););
}

