/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_ip.o
*
*/
#include "dlrpc.h"
#include "dlip.h" /* for the ASSERT() */

/* Here we've the four encode and decode functions for the rpc transit
 * system.  The first two arn't really interesting.  But the last two are
 * much more exciting.
 */
static int dlip_xdr_evoid(struct rpc_rqst *req, u32 *p, void *dummy)
{
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}

static int dlip_xdr_dvoid(struct rpc_rqst *req, u32 *p, void *dummy)
{
	return 0;
}

static int dlip_xdr_edlock(struct rpc_rqst *req, u32 *p, CDBs *dl)
{
	
	*p++ = htonl( dl->type );
	switch ( dl->type ) {
		case 0:
			*p++ = htonl(dl->cdb_u.dlock.opcode);
			*p++ = htonl(dl->cdb_u.dlock.action);
			*p++ = htonl(dl->cdb_u.dlock.lock_number);
			*p++ = htonl(dl->cdb_u.dlock.input_version_number);
			*p++ = htonl(dl->cdb_u.dlock.allocation_length);
			*p++ = htonl(dl->cdb_u.dlock.cid);
			break;
		case 1:
			*p++ = htonl(dl->cdb_u.mode.tag);
			*p++ = htonl(dl->cdb_u.mode.max_clients);
			*p++ = htonl(dl->cdb_u.mode.num_locks);
			*p++ = htonl(dl->cdb_u.mode.lock_time_ms);
			break;
	}
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}

/*
 * For now, we expect that whatever kernel process calls this rpc call has
 * allocced the correct amout of space.
 */
static int dlip_xdr_ddlock(struct rpc_rqst *req, u32 *p, RDFs *rl)
{
	int i;
	u32 realsize=0;

	rl->type = ntohl( *p++ );
	
	switch ( rl->type ) {
		case 0:
			rl->rdf_u.dlock.result         = ntohl( *p++ );
			rl->rdf_u.dlock.act            = ntohl( *p++ );
			rl->rdf_u.dlock.expired        = ntohl( *p++ );
			rl->rdf_u.dlock.state          = ntohl( *p++ );
			rl->rdf_u.dlock.version_number = ntohl( *p++ );
			rl->rdf_u.dlock.num_holders    = ntohl( *p++ );

			realsize = (rl->rdf_u.dlock.num_holders * sizeof(u32));
#if 0 /* I think this is actually ok, because when cids is null, and
		 * dsr_len == 0, it means we don't care if there are any other
		 * clients.
		 */
			ASSERT( rl->rdf_u.dlock.cids != NULL );
#endif
			{
				u8 maxcnt = MIN(rl->rdf_u.dlock.dsr_cnt,
						rl->rdf_u.dlock.num_holders);
				for(i=0; i< maxcnt ; i++) {
					rl->rdf_u.dlock.cids[i] = ntohl( *p++ );
				}
			}
			break;
		case 1:
			rl->rdf_u.rex.return_len = ntohl( *p++ );
			{
				u8 maxsze = MIN( rl->rdf_u.rex.return_len,
						rl->rdf_u.rex.dsr_dlen );
				/* Doing the assert here, because if you call for this reply
				 * mode, you wanted the info in this pointer. So you had better
				 * have set it up!
				 */
				ASSERT( rl->rdf_u.rex.exlock_block != NULL );
				for(i=0; i< maxsze ; i++) {
					(rl->rdf_u.rex.exlock_block)[i] = ntohl( *p++ );
				}
			}
			break;
		case 2:
			rl->rdf_u.mpag.num_locks    = ntohl( *p++ );
			rl->rdf_u.mpag.max_clients  = ntohl( *p++ );
			rl->rdf_u.mpag.lock_time_ms = ntohl( *p++ );
			break;
	}
	
	return 0;
}

/*
 * Now the structures that describe this rpc program
 */
static struct rpc_procinfo dlip_procedures[2] = {
	{ "dlip_null",
		(kxdrproc_t) dlip_xdr_evoid,
		(kxdrproc_t) dlip_xdr_dvoid,
		0, 0},
	{ "dlip_do_dlock",
		(kxdrproc_t) dlip_xdr_edlock,
		(kxdrproc_t) dlip_xdr_ddlock,
		0, 0}
};

static struct rpc_version dlip_version1 = {
	1,
	sizeof(dlip_procedures)/sizeof(dlip_procedures[0]),
	dlip_procedures
};

static struct rpc_version *dlip_version[] = {
	NULL,
	&dlip_version1,
};

static struct rpc_stat dlip_stats;

struct rpc_program dlip_program = {
	"dlip",
	DLIP_PROGRAM,
	sizeof(dlip_version)/sizeof(dlip_version[0]),
	dlip_version,
	&dlip_stats,
};


/*
 * Utility function to create a new rpc client
 */
int dlip_create_client(unsigned long addr, struct rpc_clnt **client)
{
	struct rpc_clnt	* clnt;
	struct rpc_xprt	* xprt;
	struct sockaddr_in  sadr;
	char	hostname[32];
	
	sadr.sin_family      = AF_INET;
	sadr.sin_addr.s_addr = addr;
	sadr.sin_port        = 0;

	strcpy(hostname, in_ntoa(addr));

	if( ! (xprt = xprt_create_proto(IPPROTO_UDP, &sadr, NULL))) {
		return -EIO;
	}
	
	clnt = rpc_create_client(xprt, hostname,
				&dlip_program, DLIP_VERSION, 
				RPC_AUTH_NULL); /* Should look into doing some sort of
										 authentication sometime soon */
	
	if( !clnt ) {
		xprt_destroy(xprt);
		return -EIO;
	} else {
		clnt->cl_softrtry = 1;
		clnt->cl_chatty   = 1;
		clnt->cl_intr     = 1;
	}

	*client = clnt;
	return 0;
}


static struct rpc_stat dlip_rpcstat = {&dlip_program};

void setup_rpc_stats(void) {
	rpc_register_sysctl();
	rpc_proc_init();
	rpc_proc_register(&dlip_rpcstat);
}

void shutdown_rpc_stats(void) {
	rpc_proc_unregister("dlip");
}
