/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: lock_common.c
**
**    Description:
**	General, non-OS specifc device lock routines.
**
*******************************************************************************
******************************************************************************/

#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/blk.h>

#include <asm/system.h>
#include <asm/io.h>

#include "pool.h"

extern pool_t pools[POOL_MAX_DEVICES];

#if (DLOCK_VERSION == 2)
/******************************************************************************
 ******************************************************************************
 **
 ** int pool_dlock()
 **
 ** Description:
 **   This routine performs device lock operations requested by pool ioctls's.
 **   PL_NDELAY instructs pool_dlock to attempt the lock once and return the
 **   results.  PL_TRY instructs pool_dlock to attempt the lock until one
 **   quarter of pool_reset_timeout_us has expired, then the results are
 **   returned.  PL_RESET instructs pool_dlock to get the lock.  This implies
 **   that if necessary the lock will be reset after pool_reset_timeout_us has
 **   expired and no activity is present.  PL_EXP can be used with any of these
 **   options.  It instructs pool_dlock to exponentially back off if it isn't
 **   getting the lock.  The function returns the lock counter value, a boolean
 **   (timeout) indicating whether the timeout operation was performed on this
 **   lock request, and cachable, which is a boolean indicating whether the
 **   clock has changed.  Since the clock can be changed by another process on
 **   this machine and cachable would still be set, the use of cachable is
 **   depricated.  The timeout operation for PL_NDELAY is a return without the
 **   lock.  The timeout operation for PL_TRY is a return without the lock.
 **   The timeout operation for PL_RESET is a return with a lock that has been
 **   reset.
 **
 **
 ** Input(s):
 **  *pp     - 
 **  *params - 
 **
 ** Output(s):
 **
 **
 ** Returns:
 **   0 if OK, error condition on error.
 **
 ******************************************************************************
 *****************************************************************************/

int pool_dlock(pool_t *pp, dlock_params_t *params)
{
  uint32       		act;
  uint32		flg;
  uint32		clk;
  uint32		pln;
  uint32		lln;
  dlockdev_t   	       *dev;

  dlock_ret_t		ret;

  uint32		delay    = POOL_CNF_DELAY_US;
  uint32		delayed  = 0;
  uint32		timeout  = 0;
  uint32		activity = 0;
  uint32		success  = 0;
  uint32		cache    = 0;
  int			pl;	       
  time_t		ts;

  act = params->dp_action;		/* lock, unlock, ulincr */
  flg = params->dp_req->l_flags;	/* ndelay, try, reset, exp */
  lln = params->dp_req->l_id;		/* logical lock num */
  pln = params->dp_phyln;		/* physical lock num */
  dev = params->dp_dldev;		/* lock device dev_t */

  POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);
  clk = pp->p_locktab[lln].l_clock; 	/* clock for lln */
  POOL_SUNLOCK(&pp->p_dlmutex, pl);

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Entered pool_dlock: lock=%d, act=0x%x, flg=0x%x, clk=%d\n", 
		 lln, act, flg, clk););

  if (act == POOL_LOCK) {
    
    ret = device_lock(act, pln, clk, dev);
    if (ret.r_status) return KRET(EIO);
    success = ret.r_result;
    clk = ret.r_dlock.l_clock;

    if (flg & (PL_TRY | PL_RESET)) {
      
      while (!success && !timeout) {
	
	us_delay(delay);
	delayed += delay;
	if (flg & PL_EXP) {
	  delay <<= 1;
	  if (delay > 1000000) delay = 1000000; 
	}

	IFDEBUG(POOL_DEBUG_LOCK1,	
		printk("  lock conflict.  delayed=%d, activ=%d, clk=%d\n",
		       delayed, activity, clk););
	
	ret = device_lock(act, pln, clk, dev);
	if (ret.r_status) return KRET(EIO);
	success = ret.r_result;

	if (activity && (ret.r_dlock.l_clock != clk)) {
	  delayed = 0;		       
	  delay = POOL_CNF_DELAY_US;
	  activity = 0;
	  ret = device_lock(POOL_ACTIVITY_OFF, pln, clk, dev);	  
          if (ret.r_status) return KRET(EIO);
	}

	clk = ret.r_dlock.l_clock;
	activity = ret.r_dlock.l_activity;

	if (!success) {
	  if (delayed >= POOL_RESET_TIMEOUT_US / 4) {
	    if (flg & PL_TRY)
	      timeout = 1;
	    else {
	      if (delayed >= POOL_RESET_TIMEOUT_US)
		timeout = 1;
	      else if (!activity) {
		ret = device_lock(POOL_ACTIVITY_ON, pln, clk, dev);
		if (ret.r_status) return KRET(EIO);
		clk = ret.r_dlock.l_clock;
		activity = 1;
	      }	      
	    }
	  }
	}
      }		/* while */
    
      if (timeout && (flg & PL_RESET)) { 	
	printk("Resetting lock %d. clk(in)=%d, activ=%d\n", 
	       lln, clk, activity);
	ret = device_lock(POOL_RESET, pln, clk, dev);
	if (ret.r_status) return KRET(EIO);
	ret = device_lock(act, pln, clk, dev);
	if (ret.r_status) return KRET(EIO);
	if (activity) {
	  ret = device_lock(POOL_ACTIVITY_OFF, pln, clk, dev);
          if (ret.r_status) return KRET(EIO);
	  clk = ret.r_dlock.l_clock;
	}
      }          
    }
    else if (flg & PL_NDELAY) {

      if (!ret.r_result) timeout = 1;
      
    }
  }
  else {  /* action is POOL_UNLOCK or UNLOCK_INCR */
    ts = timestamp();
    POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);
    pp->p_locktab[lln].l_ultime = ts;
    POOL_SUNLOCK(&pp->p_dlmutex, pl);
    ret = device_lock(act, pln, clk, dev);
    if (ret.r_status) return KRET(EIO);
  }

  ts = timestamp();

  POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);
  if ((ts - pp->p_locktab[lln].l_ultime) > DLOCK_ROLL_TIME)
    cache = 0;
  else
    cache = (ret.r_dlock.l_clock == pp->p_locktab[lln].l_clock);
  POOL_SUNLOCK(&pp->p_dlmutex, pl);

  /* output values */

  params->dp_req->l_timeout  = timeout;
  params->dp_req->l_counter  = ret.r_dlock.l_clock;
  params->dp_req->l_cacheable = cache;

  POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);

  pp->p_locktab[lln].l_state    = ret.r_dlock.l_state;
  pp->p_locktab[lln].l_activity = ret.r_dlock.l_activity;
  pp->p_locktab[lln].l_clock    = ret.r_dlock.l_clock;

  POOL_SUNLOCK(&pp->p_dlmutex, pl);

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Exiting pool_dlock\n");
	  printk("  timeout: %d\n", timeout);
	  printk("  state:   %d\n", ret.r_dlock.l_state);
	  printk("  clock:   %d\n", ret.r_dlock.l_clock);
	  printk("  activ:   %d\n", ret.r_dlock.l_activity);
	  printk("  cache:   %d\n", params->dp_req->l_cacheable););
  

  return 0;
}


/******************************************************************************
*******************************************************************************
** NAME
**	device_lock
**
** DESCRIPTION 
**	Put together the actual scsi request and forward it to the routine
**	which messes with the scsi midlayer.
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

dlock_ret_t device_lock(uint32 action, uint32 pln, uint32 clock, 
		       dlockdev_t *dlp)
{
  u_char 	*cmd;
  dlock_ret_t	ret;
  kdev_t       	dev = dlp->dl_dev;
  

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printk("Entered device_lock: lock=%d, action=0x%x\n", 
		 pln, action););

  ASSERT(DLOCK_VERSION == 2);

  cmd = kmalloc(LOCK_CDBLEN, GFP_KERNEL);

  cmd[0] = LOCK_OPCODE;
  
  if (action == POOL_LOCK)
    cmd[1] = 0x1;
  else if (action == POOL_UNLOCK)
    cmd[1] = 0x2;
  else if (action == POOL_UNLOCK_INCR)
    cmd[1] = 0x3;
  else if (action == POOL_RESET)
    cmd[1] = 0x4;
  else if (action == POOL_ACTIVITY_ON)
    cmd[1] = 0x5;
  else if (action == POOL_ACTIVITY_OFF)
    cmd[1] = 0x6;
  
  cmd[2] = (pln >> 24) & 0xff;
  cmd[3] = (pln >> 16) & 0xff;
  cmd[4] = (pln >>  8) & 0xff;
  cmd[5] = (pln)       & 0xff;

  cmd[6] = (clock >> 24) & 0xff;
  cmd[7] = (clock >> 16) & 0xff;
  cmd[8] = (clock >>  8) & 0xff;
  cmd[9] = (clock)       & 0xff;  

  cmd[10] = LOCK_RETSIZE;
  cmd[11] = 0x00;

  ret = scsi_dlock(cmd, dev);

  kfree(cmd);
  return ret;
}



/******************************************************************************
 ******************************************************************************
 **
 ** static int lock_mapping()
 **
 ** Description:
 **   This routine maps logical lock numbers to physical device lock numbers.
 **
 ** Input(s):
 **  *pp     - 
 **   logln  - 
 **  *params - 
 **
 ** Output(s):
 **
 **
 ** Returns:
 **   0 if OK, error condition on error.
 **
 ******************************************************************************
 *****************************************************************************/

int lock_mapping(pool_t *pp, dlock_params_t *params) 
{
  dlockdev_t		*dlockdevp;
  subpool_t		*sp;
  uint32		logln;
  int			i,j;
  int			error;

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Entering lock_mapping()\n"););

  if ((logln = params->dp_req->l_id) >= pp->p_dlocks) {
    printk("lock_mapping: lock number too large\n");
    error = EINVAL;
    goto fail;
  }


  for (i=0; i<pp->p_num_spools; i++) {
    sp = &(pp->p_spools[i]);
    
    if (logln < sp->sp_dlstart + sp->sp_dlocks) {
      logln -= sp->sp_dlstart;

      for (j=0; j<sp->sp_dldevs; j++) {
	dlockdevp = &(sp->sp_dldevlist[j]);

	if (logln < dlockdevp->dl_start + dlockdevp->dl_nlocks) {
	  logln -= dlockdevp->dl_start;
	  params->dp_spool = i;
	  params->dp_devnum = j;
	  params->dp_dldev = dlockdevp;
	  params->dp_phyln = logln;

	  IFDEBUG(POOL_DEBUG_LOCK1,
		  printk("Exiting lock_mapping()\n"););

	  return (0);
	}
      }
    }
  }
  
  printk("pool: lock_mapping: unknown\n");
  error = EINVAL;

 fail:

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Exiting lock_mapping()\n"););

  return KRET(error);
}

#endif /* DLOCK_VERSION == 2 */



#if (DLOCK_VERSION == 3)

/******************************************************************************
 ******************************************************************************
 **
 ** static int lock_mapping()
 **
 ** Description:
 **   This routine maps logical lock numbers to physical device lock numbers.
 **
 ** Input(s):
 **  *pp     - 
 **   logln  - 
 **  *params - 
 **
 ** Output(s):
 **
 **
 ** Returns:
 **   0 if OK, error condition on error.
 **
 ******************************************************************************
 *****************************************************************************/

int lock_mapping(pool_t *pp, dlock_params_t *params) 
{
  dlockdev_t		*dlockdevp;
  subpool_t		*sp;
  uint32		logln = 0;
  pool_dlock_t          *lock = params->dp_dlock;
  int			i,j;
  int			error;
  
  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Entering lock_mapping()\n"););

  
  if (lock->dl_lnoffset + lock->dl_lnsize > 16) {
    error = EINVAL;
    goto fail;
  }

  logln = lock->dl_ln;


  if (logln  >= pp->p_dlocks) {
    printk("lock_mapping: lock number too large\n");
    error = EINVAL;
    goto fail;
  }


  for (i=0; i<pp->p_num_spools; i++) {
    sp = &(pp->p_spools[i]);
    
    if (logln < sp->sp_dlstart + sp->sp_dlocks) {
      logln -= sp->sp_dlstart;

      for (j=0; j<sp->sp_dldevs; j++) {
	dlockdevp = &(sp->sp_dldevlist[j]);

	if (logln < dlockdevp->dl_start + dlockdevp->dl_nlocks) {
	  params->dp_dldev = dlockdevp;
	  logln -= dlockdevp->dl_start;

	  for (i = lock->dl_lnsize; i--; )
	  {
	    lock->dl_cdb[lock->dl_lnoffset + i] = logln & 0xFF;
	    logln >>= 8;
	  }

	  IFDEBUG(POOL_DEBUG_LOCK1,
		  printk("Exiting lock_mapping()\n"););

	  return (0);
	}
      }
    }
  }

  printk("lock_mapping: unknown\n");
  error = EINVAL;

 fail:

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Exiting lock_mapping()\n"););

  return KRET(error);
}

#endif /*  DLOCK_VERSION == 3  */



