/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: pool_incore.h
**
**    Description:
**	Network storage pool shared header file for in-core structures.
**
*******************************************************************************
******************************************************************************/

#ifndef __POOL_INCORE_H__
#define __POOL_INCORE_H__

#include <asm/spinlock.h>

#include "global.h"
#include "pool_ondisk.h"
#include "lock.h"
#include "pool_stats.h"


typedef struct device_s
{
  uint64		d_start;      	/* starting logical blk in spool */
  uint64		d_nblks;	/* number of blocks in device */
  kdev_t                d_dev;		/* device number */
  uint32		d_adev;		/* user space device number */
  struct inode          *inode;
  struct file           *file;
} device_t;


typedef struct subpool_s
{
  uint64		sp_start;	/* lba of first block */
  uint32		sp_len;		/* number of blocks in subpool */
  uint32		sp_stripe;	/* stripe size for subpool */
  uint64		sp_dlstart;	/* starting lock in logical pl array */
  uint32		sp_dlocks;	/* number of dlocks in this spool */
  uint32		sp_devs;      	/* number of devices in subpool */
  device_t	       	*sp_devlist;	/* list of devices */
  uint32		sp_dldevs;    	/* number of dlock devices */
  dlockdev_t		*sp_dldevlist;	/* list of dlock devices */  
} subpool_t;


#define POOL_DSTATS_MAX    	16

typedef struct pool_s
{
  /* read-only values: not protected, except by POOL_VALID 
   * which prevents opening when the structure is invalid
   * (during pool adds, first open, and final close) */

  uint64		p_pool_id;	/* pool id (on-disk) */
  uint32		p_version;	/* pool version */
  uint64		p_capacity;	/* number of blocks in pool */
  uint32		p_dlocks;	/* number of dlocks in pool */
  uint32		p_num_spools;	/* number of subpools */
  char			p_name[POOL_NAME_SIZE]; /* name of pool */
  subpool_t		*p_spools;	/* subpools */

  spinlock_t		p_mutex;	/* for reference counts and flags */
  uint32	      	p_oblk;		/* # times opened as block dev */
  uint32	        p_flags;	/* status flags */

#if (DLOCK_VERSION == 2)
  spinlock_t		p_dlmutex;	/* for lock table struct */
  dlock_t		*p_locktab;	/* table of lock states */
#endif

  spinlock_t	        p_statslock;	/* just for stats */
  pool_stat_t	        p_stats;	/* total io stats */
  pool_dstat_t	      	p_dstats[POOL_DSTATS_MAX]; /* dev stats */
  pool_dlstat_t	    	p_dlstats;	/* device lock stats */
} pool_t;



typedef struct pool_add_s 
{
  uint64		pa_pool_id;	/* pool id (on-disk) */
  uint32		pa_version;	/* pool version */
  uint64		pa_capacity;	/* number of blocks in pool */
  uint32		pa_dlocks;	/* number of dlocks */
  uint32		pa_num_spools;	/* number of subpools */
  char			pa_name[POOL_NAME_SIZE]; /* name of pool */
  subpool_t    		*pa_spools;	/* subpool list */
  uint32	       	pa_pool_dev;	/* pool device number (returned) */
} pool_add_t;



typedef struct pool_status_s
{
  uint32		ps_pool_major;
  uint32		ps_pool_minor;
  uint64		ps_pool_id;
  uint32		ps_in_use;
  uint64		ps_capacity;
  uint32		ps_dlocks;
  char			ps_pool_name[POOL_NAME_SIZE];
} pool_status_t;



typedef struct pool_list_s 
{
  uint32		pl_count;
  pool_status_t		*pl_list;
} pool_list_t;



typedef struct pool_spstatus_s
{
  uint64		sps_boffset;
  uint64		sps_bcount;
  uint32		sps_stripe;
  uint64		sps_dlstart;
  uint32		sps_dlocks;
  uint32		sps_numdevs;
  uint32		sps_dldevs;
  uint32		sps_devs[64];
} pool_spstatus_t;



typedef struct pool_splist_s
{
  uint32		spl_count;
  pool_spstatus_t	*spl_list;  
} pool_splist_t;


#endif /* __POOL_INCORE_H__ */



/* Storage pool structure 
 *	One per device, maintains important info.  The offset of a pool_t
 *	structure into the array is the minor number of the pool block
 *	special file.
 *
 * Locking protocol:
 *	Since the pools[] structure is of a fixed, compile time length there is
 *	no need to syncronize access to the array as a whole.  (This is very
 *	similar to the kernel proc table.)  Syncronization to individual
 *	records is achieved through three constructs.  First: p_mutex is a
 *	spinlock (non-sleeping lock) that once obtained (at software priority
 *	spldisk) gives the holder permission to modify any field provided the
 *	following conditions are met.  Second: the p_oblk, p_olyr, and p_ochr
 *	are reference count fields.  The reference count in lyr is incremented
 *	for every layered open we get from another driver.  ochr and oblk are
 *	boolean values signifying that there is still some process that has the
 *	devices opened as a character or block device, respectively.  If any of
 *	these counts are nonzero none of the static pool information can be
 *	changed (even if p_mutex is held.)  Static pool information basically
 *	includes information such as the number of subpools, the size of the
 *	subpools, the number of dlock devices, the label_device dev_t
 *	structure, the pool label, etc.  In addition, the pool can NOT be
 *	unloaded if the the reference counts are nonzero.  Third: the p_flags
 *	field has bit masks that indicate this pool_t field is in various
 *	states.  The structure can be PF_VALID or not PF_VALID.  If the
 *	structure is not valid it essentially means it is being intialized.
 *	Since intialization requires the driver to sleep it can not hold
 *	p_mutex, so the pool is marked invalid.  Since this happens only at
 *	initialization time, the only situation where we should see a pool_t
 *	without PF_VALID is if two pool_adds try to access the same pool_t
 *	structure. If this is the case the second passemble's call to pool_add
 *	will see the missing PF_VALID and return EBUSY.  The second interesting
 *	flag is PF_FREE.  This indiates whether this pool_t structure actually
 *	represents a pool that is in use.  Empty pool_t's (PF_FREE set) can be
 *	used by pool_add to dynamically create new pools.  pool_remove sets
 *	PF_FREE in the corresponding pool_t upon sucess.  (cjs)
 */










