/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: lock_linux.c
**
**    Description:
**	Linux specific device lock routines.
**
*******************************************************************************
******************************************************************************/

#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/blk.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>

#include "scsi.h"
#include "sd.h"
#include "pool.h"

extern Scsi_Disk *rscsi_disks;



/******************************************************************************
*******************************************************************************
**
** void pool_notify()
**
** Description:
**   This routine serves as an interrupt handler for the requests.
**
** Input(s):
**  
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

static void pool_notify(Scsi_Cmnd *SCpnt)
{

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Entered pool_notify_sd\n"););

  SCpnt->request.rq_status = RQ_SCSI_DONE;

  if (SCpnt->request.sem != NULL)
    up(SCpnt->request.sem);

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Exiting pool_notify_sd\n"););
  
  return;
}


/******************************************************************************
 ******************************************************************************
 **
 ** void us_delay()
 **
 ** Description:  This function is meant to imitate the same IRIX function.
 **               It may sleep. 
 **
 ** Input(s):
 **
 ** Output(s):
 **
 **
 ** Returns:
 **
 ******************************************************************************
 *****************************************************************************/

void us_delay(int delay) 
{
  int i, j, b, time=0, jif_dly=0;
  struct timeval tv1, tv2;
  int last = 0;

  do_gettimeofday(&tv1);
  
  if (delay <= 1000) {
    udelay(delay);
    return;
  }

  if (delay > 2000000) {
    b = delay / 2000000;
    if ((last = delay % 2000000))
      b++;
    delay = 2000000;    
  }
  else
    b = 1;
  
  for (i=0; i<b; i++) {

    if (last && i==(b-1)) 
      delay = last;
    
    jif_dly = delay * HZ / 1000000;

    if (jif_dly > 0) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(jif_dly);
    }
    else {
      for(j=0; j< delay/1000; j++)
	udelay(1000);
    }

    do_gettimeofday(&tv2);    

    if (tv2.tv_sec > tv1.tv_sec)
      time = 1000000 * (tv2.tv_sec - tv1.tv_sec) + tv2.tv_usec - tv1.tv_usec;
    else
      time = tv2.tv_usec - tv1.tv_usec;
    
    IFDEBUG(POOL_DEBUG_LOCK_TIME,
	    printk("  delay: %8d  time: %8d  b: %d jif_dly: %3d last: %d\n",
		   delay, time, b, jif_dly, last););
  }
}


#if (DLOCK_VERSION == 2)
/******************************************************************************
*******************************************************************************
**
** static dlock_ret_t scsi_dlock()
**
** Description:
**   This directs a SCSI device lock command to the scsi midlayer.
**   This needs to be generalized so it can be used for both Dlock versions.
**
** Input(s):
**
**
** Output(s):
**
**
** Returns:
**	
**
*******************************************************************************
******************************************************************************/

dlock_ret_t scsi_dlock(u_char *cmd, kdev_t dev)
{

  Scsi_Cmnd		*SCpnt;
  Scsi_Device           *device;
  char			*buff;
  int 			wait = 1;
  dlock_ret_t		ret;
  struct semaphore  	sem;
  unsigned char	        byte4, byte3, byte2, byte1, byte0;
  unsigned long 	flags;

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Entered scsi_dlock()\n"););

  /* hack - rscsi_disks exported from sd */
  device = rscsi_disks[MINOR(dev) >> 4].device;

  if (!(buff = (u_char *)scsi_malloc(512))) {
    ret.r_status = -1;
    ret.r_result = 0;
    ret.r_dlock.l_state = 0;
    ret.r_dlock.l_activity = 0;
    ret.r_dlock.l_clock = 0;
    return(ret);
  }

  spin_lock_irqsave(&io_request_lock, flags);

  if (!(SCpnt=scsi_allocate_device(NULL, device, wait))) {
    spin_unlock_irqrestore(&io_request_lock, flags);
    scsi_free(buff, 512);
    printk("dlockopen: not SCSIALLOCOK\n");
    ret.r_status = -1;
    ret.r_result = 0;
    ret.r_dlock.l_state = 0;
    ret.r_dlock.l_activity = 0;
    ret.r_dlock.l_clock = 0;
    return(ret);
  }

  SCpnt->tag = SC_TAG_HEAD;  /* overwritten by HBA driver? */

  sem = MUTEX_LOCKED;
  SCpnt->request.sem = &sem;
  SCpnt->request.rq_status = RQ_SCSI_BUSY;

  scsi_do_cmd( SCpnt,           /* Scsi_Cmnd pointer */
               cmd,             /* CDB */
               buff,            /* buffer for returned lock data */
	       LOCK_RETSIZE,    /* buff length */
	       pool_notify, 	/* scsi cmd callback function */
	       DL_TIMEOUT,      /* timeout value */
	       DL_MAXRETRY);	/* number of retried attempted */

  spin_unlock_irqrestore(&io_request_lock, flags);

  down(&sem);
  SCpnt->request.sem = NULL; 

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("device_lock: SCpnt->result = %d\n", SCpnt->result););
  
  IFDEBUG(POOL_DEBUG_LOCK3,
	  {
	    int i;
	    
	    printk("Command:\n");
	    for (i = 0; i < SCpnt->cmd_len; i++) {
	      printk("%x  ", SCpnt->cmnd[i]);
	    }
	    printk("\n");
	    
	    printk("Return:\n");
	    for (i = 0; i < SCpnt->bufflen; i++) {
	      printk("%x  ", buff[i]);
	    }
	    printk("\n");
	  }
	  );

	  
  /* fill in ret struct from buff data */

  ret.r_status = status_byte(SCpnt->result);  /* status of scsi cmd */
  
  byte4 = buff[0];
  byte3 = buff[1];
  byte2 = buff[2];
  byte1 = buff[3];
  byte0 = buff[4];
  
  ret.r_result = byte4 >> 7;
  ret.r_dlock.l_state = (byte4 >> 6) & 0x1;
  ret.r_dlock.l_activity = (byte4 >> 5) & 0x1;
  ret.r_dlock.l_clock = byte3 << 24 | byte2 << 16 | byte1 << 8 | byte0;;


  if (status_byte(SCpnt->result) != 0) {
    unsigned char *buf = SCpnt->sense_buffer;
    printk("device_lock_sd: DLOCK command failed, status: %d\n", 
	   status_byte(SCpnt->result));
    printk("device_lock_sd: sense: 0: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x "
	   "0x%02x 0x%02x 0x%02x\n"
	   "device_lock_sd: sense: 1: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x "
	   "0x%02x 0x%02x 0x%02x\n",
	   buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7],
	   buf[8], buf[9], buf[10], buf[11], buf[12], buf[13], buf[14], 
	   buf[15]);
  }

  scsi_free(buff, 512);

  wake_up(&SCpnt->device->device_wait);
  scsi_release_command(SCpnt);

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Exiting scsi_dlock\n"););

  return(ret);

}
#endif 


/******************************************************************************
*******************************************************************************
** NAME
**	timestamp
**
** DESCRIPTION
**     
**
*******************************************************************************
******************************************************************************/

time_t timestamp(void)
{
  struct timeval tv;

  do_gettimeofday(&tv);

  return( tv.tv_sec );
}
