/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1998 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: pool.c
**
**    Description:
**	Network storage pool device driver routines.
**
*******************************************************************************
******************************************************************************/

#include <sys/types.h>
#include <sys/buf.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/ddi.h>
#include <sys/conf.h>
#include <sys/cred.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>
#include <sys/ktime.h>
#include <sys/mips_addrspace.h>
#include <sys/open.h>
#include <sys/param.h>
#include <sys/scsi.h>
#include <sys/sema.h>
#include <sys/sysmacros.h>
#include <sys/systm.h>
#include <sys/scsiInitiator.h>
#include <sys/scsiInitiator_api.h>
#include <sys/xlv_ioctl.h>

#include "pool_debug.h"
#include "pool.h"
#include "pool_head.h"
#include "lock.h"

int                     pooldevflag = D_MP;		/* REQUIRED */

extern pool_t  		pools[POOL_MAX_DEVICES];	/* lookup table */

/* Timer variables all const after init, no syncronization needed */
volatile void        *iotimer_addr;		/* physical address of timer */
uint64	 	      iotimer_max;
uint64                iotimer_rate_ps;          /* picoseconds per tick */
uint                  iotimer_size;


void poolinit(void);
int  open_devices(pool_t *pp, int oflag, cred_t *crp);
int  close_devices(pool_t *pp, int oflag, cred_t *crp);
static int  poolopen(kdev_t *devp, int oflag, int otyp, cred_t *crp);
static int  poolclose(kdev_t dev, int oflag, int otyp, cred_t *crp); 
static int  poolread(kdev_t dev, uio_t *uiop, cred_t *crp); 
static int  poolwrite(kdev_t dev, uio_t *uiop, cred_t *crp);
static void pool_done(buf_t *bp);
static int  print_buf(buf_t *bp);
static int  pool_split(pool_queue_t *qp);
static int  pool_readcap(pool_t *pp, uint32 *size);


/* why are these two prototypes in pool.h ? */
extern int  poolstrategy(buf_t *bp);
extern int  poolioctl(kdev_t dev, int cmd, void *arg, int mode, cred_t *crp, int *rvalp);

extern int lock_mapping(pool_t *pp, dlock_params_t *params); 
extern int pool_dlock(pool_t *pp, dlock_params_t *params);
extern int pool_add(pool_add_t *add);
extern int pool_remove(uint32_t dev);
extern int pool_list(pool_list_t *list);
extern int pool_splist(pool_t *pool, pool_splist_t *list);
extern int pool_mapping(pool_t *pp, uint64 psector, uint32 pbcount, 
		 uint64 *dev_sector, uint32 *dev_bcount, device_t **out_dev);


/*****************************************************************************
 *****************************************************************************
 **
 ** void poolinit()
 **
 ** Description:
 **   This routine is run at system startup, thus we use it to setup values
 **   that should be available to the rest of the pool driver.  First poolinit
 **   queries the system for information on the real time clock using the
 **   cyclecntr() interfaces.  The prisa KPI function pointers are also
 **   intialized to non null values if PRISA is defined as 1 in
 **   /var/sysgen/master.d/pool. Memory is allocated for only the first pool_t
 **   object (minor number == 0).  This is the pool communication device.  It
 **   is used exclusively by passemble for user process / pool communication.
 **   The rest of the pointers in pool[] are intialized by POOL_ADD ioctl's
 **   made by passemble.  See the definition of pool_t in pool_head.h for an
 **   introduction to the locking protocol used for manipulating the pools
 **   array
 **
 *****************************************************************************
 *****************************************************************************/

void 
poolinit(void)
{
  uint64 	iotimer_phy_addr; 
  int 		i;
  uint 		timer_rate;

  printk("POOL v%d.%d.%d: %s at %s installed%s\n", POOL_VERSION >> 16, 
	 (POOL_VERSION >> 8) & 0xFF, (POOL_VERSION) & 0xFF, __DATE__,__TIME__,
	 pool_fc_open ? " with Prisa FC extensions" : "");
  
  /* Get the timer's address and precision.  Then map timer's physical address
   * to kseg1 (uncached physical memory) and setup timer dependent values */
  
  if ((iotimer_phy_addr = query_cyclecntr(&timer_rate)) == 0) {
    printk("GFS ERROR: This machine has no cycle counter.");
    return;
  }

  iotimer_rate_ps = timer_rate;
  iotimer_size = query_cyclecntr_size();
  iotimer_addr = (void *)PHYS_TO_K1(iotimer_phy_addr);
  
  if (iotimer_size >= 64) {
    iotimer_max = 0xFFFFFFFFFFFFFFFFLL;	/* 2^64-1, close enough */
  } else {
    iotimer_max = 1LL << iotimer_size; /* 2^32 */
  }

  /* set up the prisa interface, ugly hack */

  psiOpen = (int(*)(scsiPortID, u_int, struct siEndpoint **))pool_fc_open;
  psiClose = (void(*)(struct siEndpoint *))pool_fc_close;
  psiGetRequest = (struct siRequest *(*)(struct siEndpoint *))
    pool_fc_get_request;
  psiFreeRequest = (void(*)(struct siEndpoint *, struct siRequest *))
    pool_fc_free_request;
  psiExecuteCommand = (void(*)(struct siRequest *))pool_fc_execute_command;
  
  /* initialize pools[] array */
  
  for (i=0; i < POOL_MAX_DEVICES; i++) 
  {
    bzero(&pools[i], sizeof(pool_t));
    LOCK_INIT(&pools[i].p_mutex, -1, pldisk,(lkinfo_t *)-1L);
    LOCK_INIT(&pools[i].p_statslock, -1, pldisk,(lkinfo_t *)-1L);
    LOCK_INIT(&pools[i].p_dlmutex, -1, pldisk, (lkinfo_t *)-1L);
    pools[i].p_flags = PF_VALID | PF_FREE;
  }
  
  pools[0].p_flags = PF_VALID;

  /* Initialize the debugging flags  */
  
  pool_dflags = 0;
}


/******************************************************************************
*******************************************************************************
** NAME
**	poolopen
**
** DESCRIPTION
**	This gets called for every open.  We basically just ensure the call is
**	to a valid minor number and increment the open counts accordingly.  In
**	addition, if this is the first open on the device, all the lower level
**	device comprising the pool are sent open calls.  (This ensures once a
**	user has the pool open lower level driver modules are not unloaded.)  A
**	special case is the opening of minor device 0.  This is the poolcomm
**	device and is the only pool_t that accepts assemble and remvove ioctls.
**	It is the only minor number that always exists and is not dynamically
**	assigned.  In addtion, access to it is serialized at the open level.
**	(That means we don't let more than two processes open it at the same
**	time.)  This prevents multiple copies of passemble from interfering.
**	Note that I believe it is still possible for a process to open poolcomm
**	and fork.  This essentially bypasses the precaution taken here.  DO NOT
**	do this.
**
** RETURN VALUE
**	possible errno
**
*******************************************************************************
******************************************************************************/

static int 
poolopen(dev_t *devp, int oflag, int otyp, cred_t *crp) 
{
  pool_t	*pp;
  minor_t	minor = minor(*devp);
  int		s;
  int		pl;
  
  IFDEBUG(POOL_DEBUG_OPEN,
	  printk("poolopen: minor=%d\n", minor););

  if (minor >= POOL_MAX_DEVICES)
  {
    printk("POOL--poolopen: minor number out of range - %d\n", minor);
    return KRET(ENXIO);
  }

  pp = &pools[minor];

  /* if this is an open on the comm device handle it specially.  Access to
   * the comm device is exclusive */

  if (minor == 0) { 
    POOL_SLOCK(pl, &pp->p_mutex, pldisk);

    /* make sure the comm device exists */
    ASSERT((pp->p_flags & PF_FREE) == 0);

    /* check to see if device is busy */
    if (!(pp->p_flags & PF_VALID))
      {
	POOL_SUNLOCK(&pp->p_mutex, pl);
	return KRET(EBUSY);
      }
    if (pp->p_oblk > 0 ) {
      POOL_SUNLOCK(&pp->p_mutex, pl);
      return KRET(EBUSY);
    }

    if (otyp == OTYP_BLK)
      pp->p_oblk = 1;
    else if (otyp == OTYP_CHR)
      pp->p_ochr = 1;
    else if (otyp == OTYP_LYR)
      pp->p_olyr++;
    
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return(0);
  }

  /* regular open */

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);

  if (pp->p_flags & PF_FREE ) 
  {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(ENXIO);
  }
  if (!(pp->p_flags & PF_VALID))
  {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(EBUSY);
  }

  /* verify that all pointers are valid */

  ASSERT(pp->p_subpoollist != NULL);
  ASSERT(pp->p_dlockdevs != NULL);

  /* if the driver is already opened, set the proper flags and return */
  if ((pp->p_oblk) || (pp->p_ochr) || (pp->p_olyr))
  {
    IFDEBUG(POOL_DEBUG_OPEN,
	    printk("poolopen: not first open\n"););

    if (otyp == OTYP_BLK)
      pp->p_oblk = 1;
    else if (otyp == OTYP_CHR)
      pp->p_ochr = 1;
    else if (otyp == OTYP_LYR)
      pp->p_olyr++;

    POOL_SUNLOCK(&pp->p_mutex, pl);
    return(0);
  }

  pp->p_flags &= ~PF_VALID;	/* switch protection type */
  POOL_SUNLOCK(&pp->p_mutex, pl);

  /* Initialize the lock table */

  if (pp->p_locktab == 0)
  {
    pp->p_locktab = (locktab_t *)
      kmem_alloc(pp->p_label.pl_dlocks*sizeof(locktab_t), KM_SLEEP);
    memset(pp->p_locktab, 0, pp->p_label.pl_dlocks*sizeof(locktab_t));
  }

  IFDEBUG(POOL_DEBUG_OPEN,
	  printk("\n");
	  printk("SSSSSSSSSSSSSSSSSSS\n");
	  printk("    Pool Open %2d \n", minor);
	  printk("SSSSSSSSSSSSSSSSSSS\n");
	  printk("\n"););


  /* open all lower devices */


  if ((s = open_devices(pp, oflag, crp)) == 0) 

  {
    pp->p_flags |= PF_LOWEROPEN;
  }
  else
  {
    kmem_free(pp->p_locktab, pp->p_label.pl_dlocks*sizeof(locktab_t));
    pp->p_flags |= PF_VALID;
    return (s);
  }

  /* The driver is not already opened, so set flags and initialize */

  if (otyp == OTYP_BLK)
    pp->p_oblk = 1;
  else if (otyp == OTYP_CHR)
    pp->p_ochr = 1;
  else if (otyp == OTYP_LYR)
    pp->p_olyr++;

  /* Device is now opened */

  pp->p_flags |= PF_VALID;	/* XXX assuming integer write is atomic */
  return (0);
}

/******************************************************************************
*******************************************************************************
** NAME
**	poolclose
**
** DESCRIPTION
**	This routine bassically decrements the reference counts for the pool
**	device being closed.  If all these counts reach zero the lower level
**	drivers are closed.  This allows them to be unloaded if they are
**	loadable modules.  
**
** RETURN VALUE 
**	possible errno
**
*******************************************************************************
******************************************************************************/

static int 
poolclose(kdev_t dev, int oflag, int otyp, cred_t *crp) 
{
  pool_t	*pp = &pools[minor(dev)];
  int		pl;

  IFDEBUG(POOL_DEBUG1,
	  printk("poolclose: %d, otyp = %d\n", minor(dev), otyp););

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);

  ASSERT((pp->p_flags & PF_VALID) != 0);
  ASSERT((pp->p_flags & PF_FREE) == 0);


  if (otyp == OTYP_BLK) 
  {
    ASSERT(pp->p_oblk == 1);
    pp->p_oblk = 0;
  }
  else if (otyp == OTYP_CHR) 
  {
    ASSERT(pp->p_ochr == 1);
    pp->p_ochr = 0;
  }
  else if (otyp == OTYP_LYR) 
  {
    ASSERT(pp->p_olyr > 0);
    pp->p_olyr--;
  } else 
    {
      printk("unknown otype: %d\n", otyp);
    }

  /* Is this truly the last one? */

  if (!((pp->p_oblk == 0) && (pp->p_ochr == 0) && (pp->p_olyr == 0)))
  {
    IFDEBUG(POOL_DEBUG1,
	    printk("  This is not the last close\n"););

    POOL_SUNLOCK(&pp->p_mutex, pl);
    return(0);
  }

  IFDEBUG(POOL_DEBUG_OPEN,
	  printk("\n");
	  printk("SSSSSSSSSSSSSSSSS\n");
	  printk("    Pool Close   \n");
	  printk("SSSSSSSSSSSSSSSSS\n");
	  printk("\n"););

  /* Free the lock table */

  if (minor(dev) != 0) 
    {
      kmem_free(pp->p_locktab, pp->p_label.pl_dlocks*sizeof(locktab_t));
      pp->p_locktab = NULL;
  
  /* Close all the lower devices */

      pp->p_flags &= ~PF_VALID;
      POOL_SUNLOCK(&pp->p_mutex, pl);
  
      if (pp->p_flags & PF_LOWEROPEN)
	{
	  close_devices(pp, oflag, crp);
	  pp->p_flags &= ~PF_LOWEROPEN;
	}
  
      pp->p_flags |= PF_VALID;
      return (0);
    }
  else
    {
      POOL_SUNLOCK(&pp->p_mutex, pl);
      return (0);
    }
}


/******************************************************************************
*******************************************************************************
** NAME
**	poolread
**
** DESCRIPTION
**	verify alignment at pass on to poolstrategy
**
** RETURN VALUE
**	possible errno
**
*******************************************************************************
******************************************************************************/

/* ARGSUSED */
static int 
poolread(kdev_t dev, uio_t *uiop, cred_t *crp) 
{
  if (minor(dev) == 0) 
    {
      printk("poolread: Can't write to poolcomm\n");
      return (EPERM);
    }

  /* block count and word aligned */

  if ((uiop->uio_resid % BBSIZE) || 
      ((unsigned int) uiop->uio_offset & 0x3))
  {
    printk("poolread: Error with offset or size\n");
    return (EIO);
  }

  return(uiophysio(poolstrategy, NULL, dev, B_READ, uiop));
}

/******************************************************************************
*******************************************************************************
** NAME
**	poolwrite
**
** DESCRIPTION
**	verify alignment and pass request onto poolstrategy
**
** RETURN VALUE
**	possible errno
**
*******************************************************************************
******************************************************************************/

/* ARGSUSED */
static int 
poolwrite(kdev_t dev, uio_t *uiop, cred_t *crp) 
{
  if (minor(dev) == 0) 
    {
      printk("poolwrite: Can't write to poolcomm\n");
      return (EPERM);
    }

  /* block count and word aligned */

  if ((uiop->uio_resid % BBSIZE) || 
      ((unsigned int) uiop->uio_offset & 0x3))
  {
    printk("poolwrite: Error with offset or size\n");
    return (EIO);
  }

  return uiophysio(poolstrategy, 0, dev, B_WRITE, uiop);
}

/******************************************************************************
*******************************************************************************
** NAME
**	poolstrategy
**
** DESCRIPTION
**	This routine gets a new or reused queue element and breaks up the buf_t
**	pased in into requests to the various devices comprising the pool.  It
**	then intializes the queue to point to these (possibly smaller) buf_t
**	requests.  The buf_t's are intialized with callbacks to our interrupt
**	handler, pool_done.  The new buf_t's are passed on to the lower level
**	devices.  When each IO completes it calls pool_donewhich updates
**	statistics and counts.  When the last lower IO completes, pool_done
**	deallocates itself and calls biodone on the orginal buf_t.  This calls
**	the upper level callback and completes the IO.
**
** RETURN VALUE
**	possible errno
**
*******************************************************************************
******************************************************************************/

int 
poolstrategy(buf_t *ubp) 
{
  pool_t	*pp	= NULL;
  pool_queue_t	*qp	= NULL;
  buf_t		*lbp	= NULL;
  minor_t	 minor	= minor(ubp->b_edev);
  int			i;
  int			pl;	/* for LOCK */
#if POOL_STATS
  uint64		start;
  int			zero;
#endif

  IFDEBUG(POOL_DEBUG1,
	  printk("Entered poolstrategy()\n");
	  printk("  Buffer pointer: 0x%08x\n", ubp);
	  printk("  Device: (%lu) %d/%d\n", 
		 ubp->b_edev, emajor(ubp->b_edev), eminor(ubp->b_edev)););

  if (minor == 0)  	/* handle writes to poolcomm */
    {
      printk("poolstrategy: can not write to poolcomm.\n");
      return (EPERM);
    }

#if POOL_STATS
  TIMER(start);		/* Start timing */
#endif

  pp = &pools[minor];

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);
  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID) || 
      !(pp->p_flags & PF_LOWEROPEN))
  {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    cmn_err(CE_NOTE, "poolstrategy: No such device (%lu) %d/%d\n", 
	    ubp->b_edev, emajor(ubp->b_edev), eminor(ubp->b_edev));
    bioerror(ubp, ENXIO);
    biodone(ubp);
    return(0);
  }
  POOL_SUNLOCK(&pp->p_mutex, pl);

  /* Simple bound test */

  if ((ubp->b_blkno + ubp->b_bcount / BBSIZE) > pp->p_label.pl_capacity)
  {
    IFDEBUG(POOL_DEBUG_MAP,
	    printk("poolstrategy: Error with offset or size\n");
	    printk("poolstrategy: blkno=%u, bcount=%u\n",
		   (uint32)ubp->b_blkno, ubp->b_bcount););

    bioerror(ubp, EIO);
    biodone(ubp);
    return (0);
  }

  /* Make new queue element (only one queue element, for now) */

  qp = (pool_queue_t *)kmem_zalloc(sizeof(pool_queue_t), KM_SLEEP);
  LOCK_INIT(&qp->q_mutex, -1, pldisk, (lkinfo_t*)-1L);
  
  qp->q_pp = pp;
  qp->q_ubp = ubp;

#if POOL_STATS
  qp->q_start = start;
#endif

  /* Fill in the lower-level buffer structure */

  pool_split(qp);

  /* Send lower-level buffers to their respective devices */

  lbp = qp->q_lbp;

  while (lbp != 0)
  {
    IFDEBUG(POOL_DEBUG2,
	    printk("sending command: lbp=0x%llx, private=0x%llx\n", 
		   (uint64)lbp, (uint64)(lbp->b_private));
	    printk("fsprivate=0x%llx\n", (uint64)(lbp->b_fsprivate)););

#if POOL_STATS2
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);

    zero = -1;
    for (i=0; i<POOL_DSTATS_MAX; i++)
    {
      IFDEBUG(POOL_DEBUG2,
	      printk("poolstrategy: i=%d, b_edev=%d, s_dev=%d\n", 
		     i, lbp->b_edev, (pp->p_dstats)[i].s_dev););
	    
      if ((zero == -1) && ((pp->p_dstats)[i].s_dev == 0))
      {
	zero = i;
	continue;
      }

      if (lbp->b_edev == (pp->p_dstats)[i].s_dev)
	break;
    }

    if (i == POOL_DSTATS_MAX)
    {
      if (zero == -1)
      {
	cmn_err(CE_NOTE, "poolstrategy: No such device (%lu) %d/%d\n", 
		lbp->b_edev, emajor(lbp->b_edev), eminor(lbp->b_edev));
      }
      else
      {
	(pp->p_dstats)[zero].s_dev = lbp->b_edev;

	if (lbp->b_flags & B_READ)
	  (pp->p_dstats)[zero].s_out_reads++;
	else
	  (pp->p_dstats)[zero].s_out_writes++;
      }
    }
    else
    {
      if (lbp->b_flags & B_READ)
	(pp->p_dstats)[i].s_out_reads++;
      else
	(pp->p_dstats)[i].s_out_writes++;
    }
	
    POOL_SUNLOCK(&pp->p_statslock, pl);
#endif
#ifdef CELL_IRIX
    bdstrat(get_bdevsw(lbp->b_edev), lbp);
#else
    bdstrat(major(lbp->b_edev), lbp);
#endif
    lbp = lbp->b_forw;
  }
  return(0);
}

/******************************************************************************
 ******************************************************************************
 **
 ** void pool_done()
 **
 ** Description:
 **   This routine services all q_lower requests, then calls biodone, and then
 **   cleans up. This routine serves as the functional equivalent of an inter-
 **   rupt handler for the driver.
 **
 ** Input(s):
 **  *bp - Pointer to the buffer structure.
 **
 ** Output(s):
 **
 **
 ******************************************************************************
 *****************************************************************************/

static void 
pool_done(buf_t *bp)
{
  pool_queue_t	*qp;
  buf_t		*ubp;
  buf_t		*lbp, *lbp_tmp;
  int		pl;	/* for LOCK */
#if POOL_STATS
  int		pl2;	/* for LOCK */
  uint64	start, stop, usec, diff;
  pool_t	*pp;
  uint32       	b_flags, b_flags2, b_bcount;
  int		i;
#endif


  IFDEBUG(POOL_DEBUG1,
	  printk("pool_done: bp=0x%llx\n",(uint64)bp););

  qp = bptoq(bp);
  ASSERT(qp);

  /* could have multiple calls on this pool queue, so lock */

  POOL_SLOCK(pl,&qp->q_mutex, pldisk);

  qp->q_donecnt++;
  ubp = qp->q_ubp;


  /* XXX fix this. Perform some error checking */

  if (bp->b_error)
  {
    printk("pool_done: error - %d\n", bp->b_error);
    print_buf(bp);
    ubp->b_error = bp->b_error;
    bioerror(ubp, bp->b_error);
  }

  IFDEBUG(POOL_DEBUG_BUFT,
	  print_buf(bp););

#if POOL_STATS
  start = qp->q_start;
  pp = qp->q_pp;
#endif

#if POOL_STATS2
  POOL_SLOCK(pl2, &pp->p_statslock, pldisk);

  for (i=0; i<POOL_DSTATS_MAX; i++)
  {
    IFDEBUG(POOL_DEBUG2,
	    printk("pool_done: i=%d, b_edev=%d, s_dev=%d\n", 
		   i, bp->b_edev, (pp->p_dstats)[i].s_dev););
	
    if (bp->b_edev == (pp->p_dstats)[i].s_dev)
      break;
  }

  if (i == POOL_DSTATS_MAX)
  {
    printk("pool_done: device not found (%d,%d)\n", 
	   emajor(bp->b_edev), minor(bp->b_edev));
  }
  else
  {
#if 1
    if (bp->b_flags & B_READ)
    {
      if ((pp->p_dstats)[i].s_out_reads > 0)
	(pp->p_dstats)[i].s_out_reads--;
    }
    else
    {
      if ((pp->p_dstats)[i].s_out_writes > 0)
	(pp->p_dstats)[i].s_out_writes--;
    }
#endif
  }
  POOL_SUNLOCK(&pp->p_statslock, pl2);
#endif

  /* Transfer amount -- scary */

  ubp->b_resid += bp->b_resid;

  /* Still more to come */

  if (qp->q_donecnt < qp->q_lower)
  {
    IFDEBUG(POOL_DEBUG2,
	    printk("  More to come...\n"););
    POOL_SUNLOCK(&qp->q_mutex, pl);
    return;
  }

  /* Free buffers */

  lbp = qp->q_lbp;

  while (lbp != 0)
  {
    lbp_tmp = lbp;
    lbp = lbp->b_forw;
#if 1
    kmem_free(lbp_tmp, sizeof(buf_t));
#else
    freerbuf(lbp_tmp);
#endif
  }

  /* Remove a queue element from the queue */

  POOL_SUNLOCK(&qp->q_mutex, pl);
  kmem_free(qp, sizeof(pool_queue_t));

#if POOL_STATS
  /* Save the statistics data before releasing ubp */

  b_flags = ubp->b_flags;
  b_flags2 = ubp->b_flags2;
  b_bcount = ubp->b_bcount;
#endif

  /* Biodone - errors and counts should be included */

  biodone(ubp);

#if POOL_STATS
  /* Stop timing, calculate microseconds */

  TIMER(stop);

  if (stop < start) 
    {
      diff = ((uint64)(iotimer_max - start)) + 1 + stop;
    } 
  else
    {
      diff = stop - start;
    }
      
  usec = (diff*iotimer_rate_ps)/1000000;

  IFDEBUG(POOL_DEBUG_TIME,
	  printk("diff  = %lld\n", stop-start);
	  printk("start = %lld\n", start);
	  printk("stop  = %lld\n", stop);
	  printk("diff  = %lld\n", diff);
	  printk("usecs  = %lld\n", usec););

  /* Update statistics */

  POOL_SLOCK(pl,&pp->p_statslock, pldisk);

  if (b_flags & B_READ)
  {
    pp->p_stats.s_reads++;
    pp->p_stats.s_readblks += b_bcount/BBSIZE;
    pp->p_stats.s_readtime += usec;

    if (b_flags2 & B_GFS_INVMISS)
    {
      IFDEBUG(POOL_DEBUG2,
	      printk("  Invalidated!\n"););

      pp->p_stats.s_invmiss++;
      pp->p_stats.s_invtime += usec;
    }
  }
  else
  {
    pp->p_stats.s_writes++;
    pp->p_stats.s_writeblks += b_bcount/BBSIZE;
    pp->p_stats.s_writetime += usec;
  }
  POOL_SUNLOCK(&pp->p_statslock, pl);
#endif
}

/*****************************************************************************
******************************************************************************
**
** int poolioctl()
**
** Description:
**   Implements all pool ioctl's.  This includes dlock/unlock, assemble/remove,
**   statistics gathering, and other miscellaneous ioctls.  Note there should
**   be only one retrn from this function to handle the return protocol
**   correctly.  If an ioctl causes an internal routine to be called, the
**   parameters are copied to kernel space before the call is made.  NOTE:
**   structures that are in user space and are pointed to by the structure
**   passed as the ioctl arguement are not copied into kernel space before the
**   call.  Some ioctls will only work on the /dev/poolcomm device (71,0)
**
** Input(s):
**   dev   - Device number to which ioctl command is to be passed.
**   cmd   - Command that is to be issued to the device.
**  *arg   - arg is IOCTL specific value passed in from user space
**   mode  - 
**  *cr    - Pointer to the user's credentials.
**  *rvalp - Pointer to storage associated with the return value.
**
** Output(s):
**  *rvalp - Pointer to storage containing the return value.
**
** Returns:
**   0 if OK, error condition on error.
**
******************************************************************************
*****************************************************************************/

/* ARGSUSED */
int 
poolioctl(kdev_t dev, int cmd, void *arg, int mode, cred_t *cr, int *rvalp) 
{
  pool_t	*pp    = NULL;
  int            errno = 0;
  int            minor = minor(dev);
  int		 pl    = 0;			/* for LOCK */
  int		 s     = 0;
  int		 i     = 0;

  IFDEBUG(POOL_DEBUG1,
	  printk("Entered poolioctl()\n");
	  printk("  Device:  (%lu) %d/%d\n", dev, emajor(dev), minor(dev));
	  printk("  Command: 0x%x\n", cmd););

  if (minor < 0 || minor >= POOL_MAX_DEVICES)
    {
      *rvalp = -1;
      return ENXIO;
    }

  pp = &pools[minor];
  *rvalp = -1;

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);
  if (pp->p_flags & PF_FREE) 
  {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return ENXIO;
  }
  ASSERT(pp->p_ochr > 0 || pp->p_olyr > 0 || pp->p_oblk > 0);
  ASSERT(pp->p_flags & PF_VALID);
  POOL_SUNLOCK(&pp->p_mutex, pl);

  /* Do commands */

  switch(cmd)
  {
  case POOL_LOCK:
  case POOL_UNLOCK:
  case POOL_UNLOCK_INCR:
  {
    lock_ctl_t		lock;
    dlock_params_t	params;
    
    if (s = gencopy(arg, &lock, sizeof(lock), 1))
    {
      printk("poolioctl: Error (un)locking.\n");
      return EFAULT;
    }
      
    /* Build params */
      
    params.dp_action = cmd;
    params.dp_ctlarg = &lock;
          
    /* Logical to physical mapping */
      
    if (s = lock_mapping(pp, &params))
    {
      return s;
    }
      
    IFDEBUG(POOL_DEBUG2,
	    printk("  action  = %u\n", params.dp_action);
	    printk("  logln   = %u\n", params.dp_ctlarg->l_id);
	    printk("  phyln   = %u\n", params.dp_phyln);
	    printk("  devnum  = %u\n", params.dp_devnum);
	    printk("  major   = %d\n", emajor(params.dp_dldev->dl_dev));
	    printk("  minor   = %d\n", minor(params.dp_dldev->dl_dev)););
      
    /* Perform the device lock command */
      
    errno = pool_dlock(pp, &params);
    if (s = gencopy(&lock, arg, sizeof(lock), 1))
    {
      if (errno == 0)
	errno = EFAULT;
    }
    break;
  }
  case POOL_CLEAR_STATS:
  {
    IFDEBUG(POOL_DEBUG2,
	    printk("  Clearing statistics...\n"););
    
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);

    bzero(&(pp->p_stats), sizeof(pool_stat_t));
    bzero(&(pp->p_dlstats), sizeof(pool_dlstat_t));

    for (i=0; i<POOL_DSTATS_MAX; i++)
    {
      (pp->p_dstats)[i].s_out_reads = 0;
      (pp->p_dstats)[i].s_out_writes = 0;
    }

    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
  }
  case POOL_DUMP_TOTALSTATS:
  {
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);
    
    /* Copy out statistics structure */
    
    if (copyout(&(pp->p_stats), arg, sizeof(pool_stat_t)))
    {
      *rvalp = -1;
      errno = EFAULT;
    }
    else
      *rvalp = sizeof(pool_stat_t);
    
    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
  }
  case POOL_DUMP_DISKSTATS:
  {
#if POOL_STATS2
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);
    
    for (i=0; i<POOL_DSTATS_MAX; i++)
    {
      if ((pp->p_dstats)[i].s_dev == 0)	
	break;
    }
    if ((i == 0)||(i == POOL_DSTATS_MAX))
      *rvalp = 0;			
    else
      *rvalp = (int)(i*sizeof(pool_dstat_t));			
    
    /* Copy out statistics structure */
    
    if (copyout(pp->p_dstats, arg, *rvalp))
    {
      *rvalp = -1;
      errno = EFAULT;
    }

    POOL_SUNLOCK(&pp->p_statslock, pl);
#endif
    break;
  }
  case POOL_DUMP_LOCKSTATS:
  {
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);

    /* Copy out device lock statistics structure */
	
    if (copyout(&(pp->p_dlstats), arg, sizeof(pool_dlstat_t)))
    {
      *rvalp = -1;
      errno = EFAULT;
    }
    else
      *rvalp = sizeof(pool_dlstat_t);

    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
  }
  case POOL_READ_CAPACITY:
    
    /* Copy out the capacity */
    
    if (copyout((void *)&(pp->p_label.pl_capacity), arg, 
		sizeof(uint64)))
    {
      errno = EFAULT;
    }
    else
      *rvalp = sizeof(uint64);
    
    break;
  case POOL_DLOCK_COUNT:
    /* Copy out the dlock count */
    
    if (copyout((void *)&(pp->p_label.pl_dlocks), arg, sizeof(uint32)))
      { 
	errno = EFAULT;
      }
    else
      *rvalp = sizeof(uint32);

    break;
  case POOL_DEBUG_ON:  /*  Turn on debug bits.  */
  {
    uint64 x;
    
    if (copyin(arg,&x,sizeof(uint64)))
    {
      return EFAULT;
    }
#ifdef DEBUG
    printk("Pool driver debugging enabled.\n");
    printk("  Flags: 0x%16llX + 0x%16llX = ", pool_dflags, x);
#endif
    pool_dflags |= x;
#ifdef DEBUG
    printk("0x%16llX\n", pool_dflags);
#endif
    break;
  }
  case POOL_DEBUG_OFF:  /*  Turn off debug bits.  */
  {
    uint64 x;
	  
    if (copyin(arg,&x,sizeof(uint64)))
    {
      return EFAULT;
    }

#ifdef DEBUG
    printk("Pool driver debugging disabled.\n");
    printk("  Flags:  0x%16llX - 0x%16llX = ", pool_dflags, x);	
#endif
    pool_dflags &= ~x;
#ifdef DEBUG
    printk("0x%16llX\n", pool_dflags);
#endif
    break;
  }
  case POOL_ADD:
  {
    pool_add_t	add;
    pool_add_t	save;
    if ( minor) {	/* root on minor(pooldev) == 0 ? */
      errno = EINVAL;
      break;
    }
    if (copyin(arg, (void *)&add, sizeof(add)) != 0)
    {
      errno = EFAULT;
      break;
    }

    /* FIXME */
    /* minus one because the disk_partition_header is 1 BB after the label */
    ASSERT( add.pa_label->pl_labelsize <= (POOL_SIZE_PHEADER(0) - 1) );

    /* save a copy with user space pointers intact (this only matters in an
     * IRIX64 kernel */

    save = add;

    /* convert addresses from 32bit to kernel internal address length
     * if we're in a 64bit kernel */

#if _MIPS_SIM == _ABI64

    add.pa_label = (pool_labels_t *)((app64_ptr_t)add.pa_ulabel.u_user);
    add.pa_dlockdevs = (dlockdev_t *)((app64_ptr_t)add.pa_udlockdevs.u_user);
    add.pa_spools = (subpool_t *)((app64_ptr_t)add.pa_uspools.u_user);

#endif 

    errno = pool_add(&add);

    /* copy the returned values into the save structure  and copy it 
     * back out */

    save.pa_pool_dev = add.pa_pool_dev;
    save.pa_status = add.pa_status;
    if (copyout((void *)&save, arg, sizeof(save)) != 0)
    {
      if (!errno)
	errno = EFAULT;
    }
    break;
  }
  case POOL_REMOVE:
  {
    kdev_t	dev;

    if (minor) {	/* root on minor(pooldev) == 0 ? */
      errno = EINVAL;
      break;
    }
    if (copyin(arg, (void *)&dev, sizeof(dev)) != 0) 
    {
      errno = EFAULT;
      break;
    }
    errno = pool_remove(dev);
    break;
  }
  case POOL_LIST:
  {
    pool_list_t	list;
    pool_list_t	save;
    
    if (minor) {	/* root on minor(pooldev) == 0 ? */
      errno = EINVAL;
      break;
    }
    if (copyin(arg, (void *)&list, sizeof(list)) != 0) 
    {
      errno = EFAULT;
      break;
    }

    /* save user space pointer values */

    save = list;

    /* generate kernel pointers for the structure */

#if _MIPS_SIM == _ABI64   
    list.pl_list = (pool_status_t *)((app64_ptr_t)list.pl_ulist.u_user);
#endif    

    errno = pool_list(&list);

    save.pl_count = list.pl_count;

    if (copyout((void *)&save, arg, sizeof(save)) != 0)
    {
      if (!errno)
	errno = EFAULT;
    }
    break;
  }
  case POOL_SPLIST:
    {
      pool_splist_t list;
      pool_splist_t save;

      if (copyin(arg, &list, sizeof(list)) != 0)
	{
	  errno = EFAULT;
	  break;
	}

      if (list.spl_count <= 0) {
	errno = EFAULT;
	break;
      }

      /* save user space addresses */

      save = list;

      /* generate kernel addresses */

#if _MIPS_SIM == _ABI64
      list.spl_list = (pool_spstatus_t *)((app64_ptr_t)list.spl_ulist.u_user);
#endif

      errno = pool_splist(pp, &list);

      /* copy return values back out */

      save.spl_devlocks = list.spl_devlocks;
      save.spl_count = list.spl_count;

      if (copyout((void *)&save, arg, sizeof(save)) != 0)
	{
	  if (!errno)
	    errno = EFAULT;
	}
      break;
    }

  /* The rest of the ioctl's are for compatablity with dksc and xlv.  These 
   * allow us to make an xfs filesystem on top of pool. They should not 
   * be compiled in for a port.  All functionality should also be accessable 
   * through POOL_ ioctl's. */

  /* XXX needless to say they are not working yet. */

  case DIOCREADCAPACITY:
    {
      uint32	size;
      errno = pool_readcap(pp, &size);
      if (copyout((void *)&size, arg, sizeof(uint32)) != 0 && errno == 0)
	errno = EFAULT;
      break;
    }
  case DIOCGETXLVDEV:
    {
      xlv_getdev_t	getdev;
      getdev.version = XLV_GETDEV_VERS;
      getdev.data_subvol_dev = dev;
      getdev.log_subvol_dev = makedev(0,0);
      getdev.rt_subvol_dev = makedev(0,0);
      if (copyout(&getdev, arg, sizeof(xlv_getdev_t)) != 0) 
	{
	  errno = EFAULT;
	}
      break;
    }
  default:
    *rvalp = -1;
    errno = EINVAL;
  }

  if (errno == 0 && *rvalp == -1)
    *rvalp = 0;
  return (errno);
}

/******************************************************************************
*******************************************************************************
** NAME
**	print_buf
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

static int
print_buf(buf_t *bp)
{
  printk("Flags:	0x%x\n", bp->b_flags);
  printk("Flags2:	0x%x\n", bp->b_flags2);
  printk("Forward: 	%p\n", bp->b_forw);
  printk("Back: 	%p\n", bp->b_back);
  printk("Av Forward:	%p\n", bp->av_forw);
  printk("Av Back:	%p\n", bp->av_back);
  printk("Device:	(%d,%d)\n", emajor(bp->b_edev), minor(bp->b_edev));
  printk("Error:	%d\n", bp->b_error);
  printk("Vnode Offset	%lld\n", bp->b_offset);
  printk("Transfer Cnt	%u\n", bp->b_bcount);
  printk("Residual	%u\n", bp->b_resid);
  printk("Remaining	%u\n", bp->b_remain);
  printk("Bufsize	%u\n", bp->b_bufsize);
  printk("b_un		%p\n", bp->b_un.b_daddr);
  printk("b_obj		%p\n", bp->b_obj.proc);
  printk("Blk #		%lld\n", bp->b_blkno);
  printk("Start Clock   %u\n", bp->b_start);
  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	open_devices
**
** DESCRIPTION
**      open all lower level devices described in labels include device lock
** 	device.
**
** RETURN VALUE
**	Standard unix errno values
**
*******************************************************************************
******************************************************************************/

/*
 *   Incorporate backout on failure capability from Linux routine. 
 */

int 
open_devices(pool_t *pp, int oflag, cred_t *crp) 
{
  pool_labels_t		*label = &pp->p_label;
  subpool_t           	*spoolp;
  device_t            	*devp;
  dlockdev_t		*dlockdevp;
  dev_t 		dev;
  int			i,j;
  int			error;
#ifdef CELL_IRIX
  struct bdevsw		*devswp;
#endif
  
  IFDEBUG(POOL_DEBUG_OPEN,
	  printf("open_devices:\n"););
  
  for (j=0; j<label->pl_subpools; j++)
  {
    spoolp = &(pp->p_subpoollist[j]);

    for (i=0; i<spoolp->sp_devices; i++)
    {
      devp = &(spoolp->sp_devlist[i]);
      dev = devp->d_dev;

      IFDEBUG(POOL_DEBUG_OPEN,
	      printf("open_devices: opening %d,%d (%d)\n",
		     emajor(dev), minor(dev), dev););

#ifdef CELL_IRIX
      /* XXX we assume that the dev passed in corresponds to the block device
       * special file.  dev_t's for character and block are the same under
       * 6.2/6.3, but under 6.4 the vhdl's for block and char differ. */

      devswp = get_bdevsw(dev);
      if (devswp == NULL) {
	IFDEBUG(POOL_DEBUG_OPEN,
		printf("open_devices: get_bdevsw returned null\n"););
	return ENXIO;
      }
      error = bdhold(devswp);
      IFDEBUG(POOL_DEBUG_OPEN,
	      if (error != 0) {
		printf("open_devices bdhold failed!\n");
	      });
#else
      error = bdhold(major(dev));
#endif

      if (!error)
      {
#ifdef CELL_IRIX
	error = bdopen(devswp, &dev, oflag, OTYP_LYR, crp);
	bdrele(devswp);
#else
	error = bdopen(major(dev), &dev, oflag, OTYP_LYR, crp);
	bdrele(major(dev));
#endif
      }
      if (error)
      {
	printf("open_devices: error %d opening %d,%d (%d)\n",
	       error, emajor(dev), minor(dev), dev);
	return (error);
      }
    }
  }
  
  for (j=0; j<label->pl_dldevices; j++)
  {
    dlockdevp = &(pp->p_dlockdevs[j]);
    if (emajor(dlockdevp->dl_dev) != POOL_PRISA)
    {
      dev = dlockdevp->dl_dev;

      IFDEBUG(POOL_DEBUG_OPEN,
	      printf("open_devices: opening %d,%d (%d)\n",
		     emajor(dev), minor(dev), dev););
#ifdef CELL_IRIX
      devswp = get_bdevsw(dev);
      if (devswp == NULL)
	return ENXIO;
      error = bdhold(devswp);
#else
      error = bdhold(major(dev));
#endif
      if (error)
      {
	printf("open_devices: error bdhold\n");
	return (error);
      }
#ifdef CELL_IRIX
      error = bdopen(devswp, &dev, oflag, OTYP_LYR, crp);
      bdrele(devswp);
#else
      error = bdopen(major(dev), &dev, oflag, OTYP_LYR, crp);
      bdrele(major(dev));
#endif
      if (error)
      {
	printf("open_devices: error bdopen\n");
	return (error);
      }
    }
    else 
    {
      IFDEBUG(POOL_DEBUG_OPEN,
	      printf("open_devices: opening %llx\n", dlockdevp->dl_wwn););

      error = psiOpen(dlockdevp->dl_wwn, siOpen_Shared, 
			 (struct siEndpoint **)&(dlockdevp->dl_data));

      if (error) 
      {
	printf("open_devices: error opening %llx\n", dlockdevp->dl_wwn);
	printf("open_devices: error = %d\n", error);
	return (error);
      }
    }
  }
  
  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	close_devices
**
** DESCRIPTION
**	Close all lower level devices opened for this pool_t structure
**
** RETURN VALUE
**	Standard unix errno values
**
*******************************************************************************
******************************************************************************/

int 
close_devices(pool_t *pp, int oflag, cred_t *crp) 
{
  pool_labels_t		*label = &pp->p_label;
  subpool_t           	*spoolp;
  dlockdev_t		*dlockdevp;
  device_t           	*devp;
  int                 	i,j;
  int			error;
  kdev_t 		dev;

  IFDEBUG(POOL_DEBUG2,
	  printk("close_devices:\n"););

  /* Close all storage devices */

  for (j=0; j<label->pl_subpools; j++)
  {
    spoolp = &(pp->p_subpoollist[j]);

    for (i=0; i<spoolp->sp_devices; i++)
    {
      devp = &(spoolp->sp_devlist[i]);
      dev = devp->d_dev;

      IFDEBUG(POOL_DEBUG_OPEN,
	      printk("close_devices: closing %d,%d (%d)\n",
		     emajor(dev), minor(dev), dev););
#ifdef CELL_IRIX
      error = bdclose(get_bdevsw(dev), dev, oflag, OTYP_LYR, crp);
#else
      error = bdclose(major(dev), dev, oflag, OTYP_LYR, crp);
#endif

      if (error)
      {
	printk("pool: close_devices: error %d\n", -error);
      }
    }
  }

  /*
   *  close device lock devices
   */

  for (j=0; j<label->pl_dldevices; j++)
  {
    dlockdevp = &(pp->p_dlockdevs[j]);
    dev = dlockdevp->dl_dev;

    if (emajor(dev) != POOL_PRISA) 
    {
      
      IFDEBUG(POOL_DEBUG_OPEN,
	      printk("close_devices: closing %d,%d (%d)\n",
		     emajor(dev), minor(dev), dev););

#ifdef CELL_IRIX      
      error = bdclose(get_bdevsw(dev), dev, oflag, OTYP_LYR, crp);
#else
      error = bdclose(major(dev), dev, oflag, OTYP_LYR, crp);
#endif

      if (error)
	return (error);
    }
    else
    {
      
      IFDEBUG(POOL_DEBUG_OPEN,
	      printk("close_devices: closing %llx\n", dlockdevp->dl_wwn););

      psiClose((struct siEndpoint *)dlockdevp->dl_data);
    }

  }
  return (0);
}

/******************************************************************************
 ******************************************************************************
 **
 ** static int pool_split()
 **
 ** Description:
 **   [IRIX specific]
 **   This routine maps upper-level driver requests to lower-level driver
 **   requests.  The pool_queue_t contains a pointer to the upper level buf_t
 **   request structure and the associated pool_t structure.  It also is the
 **   head of a linked list containing all the lower level buf_t structures the
 **   upper level buf_t structure is translated into.  This routine allocates
 **   the lower buf_t structures, fills them in, and links them together.
 **
 ** Input(s):
 **  *qp - Pointer to the pool queue structure.
 **
 ** Output(s):
 **
 **
 ** Returns:
 **   0 if OK, error condition on error.
 **
 ******************************************************************************
 *****************************************************************************/

static int 
pool_split(pool_queue_t *qp) 
{
  pool_labels_t		*labels = &qp->q_pp->p_label;
  buf_t			*ubp = qp->q_ubp;
  buf_t			*lbp = NULL;  
  
  subpool_t           	*spoolp;
  device_t            	*devp;
  int			i,j, rc;
  
  uint32		current_spool;
  uint32		current_dev;
  
  caddr_t		caddr	= ubp->b_dmaaddr;
  uint64	        blkno	= ubp->b_blkno;
  uint32		bcount = 0;
  uint64		blk;
  uint32		bc;
  
  unsigned		b,s;
  uint64		o;
  uint32		devnum;
  
  /* Loop until all requests are created */
  
  current_spool = 0;
  
  /* Invalidate cache - R4x000 */
  
  if (ubp->b_flags & B_READ)
  {
    dki_dcache_inval(ubp->b_dmaaddr, ubp->b_bcount);
  }
  else
  {
    dki_dcache_wb(ubp->b_dmaaddr, ubp->b_bcount);
  }
  
  while (bcount < ubp->b_bcount)
  {
    /* map the request */

    rc=pool_mapping(qp->q_pp, blkno, ubp->b_bcount - bcount, &blk, &bc, &devp);
    if (rc)
      return rc;

    if (lbp == NULL)
    {
#if 1
      lbp = (buf_t *)kmem_zalloc(sizeof(buf_t), KM_SLEEP);
#else
      lbp = getrbuf(KM_SLEEP);
#endif
      qp->q_lbp = lbp;
      qp->q_lower = 1;
    }
    else
    {
      IFDEBUG(POOL_DEBUG2,
	      printk("  Creating the 2nd buffer...\n"););

#if 1
      lbp->b_forw = (buf_t *)kmem_zalloc(sizeof(buf_t), KM_SLEEP);
#else
      lbp->b_forw = getrbuf(KM_SLEEP);
#endif
      lbp->b_forw->b_back = lbp;
      lbp = lbp->b_forw;
      qp->q_lower++;
    }

    /* Fill in the buffer structure */

    lbp->b_edev = devp->d_dev;

    /* setup start address.  Translate from subpool device offset to 
     * partion offset (i.e. account for the partion header) */

    lbp->b_blkno = (int64)
      (blk + POOL_SIZE_PHEADER(sizeof(disk_partition_header_t)));
    lbp->b_bcount = bc;
    lbp->b_bufsize = bc;  

    lbp->b_error = 0;
    lbp->b_iodone = pool_done;

    lbp->b_private = (void *)qp; /* prisa trashes b_private */
    lbp->b_fsprivate = (void *)qp;  
    lbp->b_fsprivate2 = (void *)qp;  

    lbp->b_pages	= ubp->b_pages;
    lbp->b_vp 	= ubp->b_vp;
    lbp->b_flags = ubp->b_flags;

    if (BP_ISMAPPED(qp->q_ubp))
    {
      IFDEBUG(POOL_DEBUG2, printk("  Is mapped\n"););

      lbp->b_flags &= ~(B_MAPPED | B_PAGEIO);
      lbp->b_dmaaddr = caddr;
    }
    else
    {
      IFDEBUG(POOL_DEBUG2, printk("  Is NOT mapped\n"););
	  
      lbp->b_un.b_pfdat = ubp->b_pages;
      lbp->b_remain = ubp->b_remain;
      /* lbp->b_dmaaddr = 0; */
    }

    /* Increment counters */

    blkno += bc/BBSIZE;
    bcount += bc;
    caddr += bc;
  } 
  return(0);
}



/******************************************************************************
*******************************************************************************
** NAME
**	pool_readcap
**
** DESCRIPTION
**	return the pool volume capacity in basic blocks in *size.
**
** RETURN VALUE
**	0 on success, errno value on error
**
*******************************************************************************
******************************************************************************/

static int 
pool_readcap(pool_t *pp, uint32 *size)
{
  subpool_t	*spool;
  uint64	 big_size;

  /* get the last subpool. */

  spool = &pp->p_subpoollist[pp->p_label.pl_subpools - 1];

  /* determine the pool's size. */

  big_size = spool->sp_start + spool->sp_len;

  if (big_size >= (1ULL << 32)) 
    {
      return EFBIG;
    }
  else
    {
      *size = (uint32)big_size;
      return 0;
    }
}


/******************************************************************************
*******************************************************************************
** NAME
**	timestamp
**
** DESCRIPTION
**     
**
*******************************************************************************
******************************************************************************/

time_t
timestamp(void)
{
  /* FIXME !!!!! */


  return( 0 );
}











