/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: pool_linux.c
**
**    Description:
**	Linux-specific pool driver routines.
**
*******************************************************************************
******************************************************************************/

#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/blk.h>
#include <linux/hdreg.h>
#include <linux/sched.h>


#include "pool.h"


extern pool_t pools[POOL_MAX_DEVICES];

static int pool_blocksizes[POOL_MAX_DEVICES];
static int pool_maxreadahead[POOL_MAX_DEVICES];
static int pool_size[POOL_MAX_DEVICES];  	/* in KB */



/******************************************************************************
*******************************************************************************
** NAME
**	do_pool_request
**
** DESCRIPTION
**	Stub routine to let Linux know we are a block device and we actually do
**	exist
**
*******************************************************************************
******************************************************************************/

static void do_pool_request(void)
{
  printk("pool: request function called, this should never be called\n");
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_read
**
** DESCRIPTION
**
**
*******************************************************************************
******************************************************************************/

static ssize_t pool_read(struct file *file, char *buf, size_t count, 
			 loff_t *ppos)
{
  int minor=MINOR(file->f_dentry->d_inode->i_rdev);

  if (minor > POOL_MAX_DEVICES)
    return (-ENXIO);

  if (minor == 0)  /* prevent reads / writes to pool comm device */
    return (-EPERM);

  return block_read(file, buf, count, ppos);
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_write
**
** DESCRIPTION
**
**
*******************************************************************************
******************************************************************************/

static ssize_t pool_write(struct file *file, const char *buf, size_t count, 
			  loff_t *ppos)
{
  int minor=MINOR(file->f_dentry->d_inode->i_rdev);

  if (minor > POOL_MAX_DEVICES)
    return (-ENXIO);

  if (minor == 0)  /* prevent reads / writes to pool comm device */
    return (-EPERM);

  return block_write(file, buf, count, ppos);
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_open
**
** DESCRIPTION
**	This gets called for every open.  We basically just ensure the call is
**	to a valid minor number and increment the open counts accordingly.  In
**	addition, if this is the first open on the device, all the lower level
**	device comprising the pool are sent open calls.  (This ensures once a
**	user has the pool open lower level driver modules are not unloaded.)  A
**	special case is the opening of minor device 0.  This is the poolcomm
**	device and is the only pool_t that accepts assemble and remvove ioctls.
**	It is the only minor number that always exists and is not dynamically
**	assigned.  In addtion, access to it is serialized at the open level.
**	(That means we don't let more than two processes open it at the same
**	time.)  This prevents multiple copies of passemble from interfering.
**	Note that I believe it is still possible for a process to open poolcomm
**	and fork.  This essentially bypasses the precaution taken here.  DO NOT
**	do this. (cjs)  Only one open (locally) is allowed on a pool to 
**	prevent unaccounted cache blocks.
**
** RETURN VALUE
**	possible errno
**
*******************************************************************************
******************************************************************************/

static int pool_open(struct inode *inode, struct file *file) 
{
  int 		minor = MINOR(inode->i_rdev);
  pool_t	*pp;
  int		pl;
  
  IFDEBUG(POOL_DEBUG1,
	  printk("poolopen: minor=%d\n", minor););

  if (minor >= POOL_MAX_DEVICES) {
    printk("POOL--poolopen: minor number out of range - %d\n", minor);
    return (-ENXIO);
  }

  pp = &pools[minor];

  /* if this is an open on the comm device handle it specially. */

  if (minor == 0) { 
    POOL_SLOCK(pl, &pp->p_mutex, pldisk);

    /* make sure the comm device exists */
    ASSERT((pp->p_flags & PF_FREE) == 0);

    /* check to see if device is busy */
    if (!(pp->p_flags & PF_VALID)) {
      POOL_SUNLOCK(&pp->p_mutex, pl);
      return (-EBUSY);
    }

    if (pp->p_oblk > 0 ) {
      POOL_SUNLOCK(&pp->p_mutex, pl);
      return (-EBUSY);
    }

    pp->p_oblk++;
    
    POOL_SUNLOCK(&pp->p_mutex, pl);
    MOD_INC_USE_COUNT;
    return(0);
  }

  /* regular open */

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);

  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_LOWEROPEN)) {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return (-ENXIO);
  }

  if (!(pp->p_flags & PF_VALID) || pp->p_oblk > 0) {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return (-EBUSY);
  }

  ASSERT(pp->p_spools != NULL);

  pp->p_oblk++;

  POOL_SUNLOCK(&pp->p_mutex, pl);
  MOD_INC_USE_COUNT;
  return(0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_release
**
** DESCRIPTION
**
** RETURN VALUE 
**	possible errno
**
*******************************************************************************
******************************************************************************/

static int pool_release(struct inode *inode, struct file *file) 
{
  int 		minor = MINOR(inode->i_rdev);
  pool_t	*pp;
  int		pl;

  IFDEBUG(POOL_DEBUG1,
	  printk("poolrelease: minor=%d\n", minor););

  pp = &pools[minor];

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);

  ASSERT((pp->p_flags & PF_VALID) != 0);
  ASSERT((pp->p_flags & PF_FREE) == 0);
  ASSERT(pp->p_oblk > 0);

  pp->p_oblk--;

  POOL_SUNLOCK(&pp->p_mutex, pl);
  MOD_DEC_USE_COUNT;
  return(0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_map
**
** DESCRIPTION   
**	rsector and size are in basic blocks and are relative to the
**	logical pool device.  pool_blkmap() is called to translate the 
**	logical device and logical block into the actual device and block
**	for the request.
**
** RETURN VALUE
**	The parameters rdev and rsector are changed to reflect the mapping.
**
*******************************************************************************
******************************************************************************/

static int pool_map(int minor, kdev_t *rdev, unsigned long *rsector, 
		    unsigned long size)
{
  pool_t	*pp = NULL;
  device_t	*p_dev;
  int	       	i;
  int	       	pl;	/* for LOCK */
  int	       	bc;
  uint64      	dev_sector;

#if POOL_STATS
  int			zero;
#endif

  IFDEBUG(POOL_DEBUG1,
	  printk("Entered pool_map()  minor=%d, rsector=%ld, size=%ld\n",
		 minor, *rsector, size););

  if (minor == 0) {  	/* handle writes to poolcomm */    
    printk("pool_map: can not write to poolcomm.\n");
    return (-EPERM);
  }

  pp = &pools[minor];

  POOL_SLOCK(pl, &pp->p_mutex, pldisk);

  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID) || 
      !(pp->p_flags & PF_LOWEROPEN)) {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    printk("pool_map: No such minor device %d\n", minor);
    return (-EIO);
  }

  POOL_SUNLOCK(&pp->p_mutex, pl);


  /* Simple bound test */

  if ((*rsector + size ) > pp->p_capacity) {
    printk("pool_map: Error with offset or size\n");
    printk("pool_map: blkno=%u, bcount=%lu\n", (uint32)*rsector, size);
    return (-EIO);
  }

  /* Fill in the lower-level buffer structure */

  if (pool_blkmap(pp, *rsector, size << BBSHIFT, &dev_sector, &bc, &p_dev)) {
    return (-EIO);
  }

  if (sizeof(unsigned long) <= 4 && dev_sector > UINT_MAX) {
    printk("poolmap: attempted access beyond 32 bit sector limit.\n");
    return (-EIO);
  }

  *rdev = p_dev->d_dev;
  *rsector = dev_sector;


  /* do statistics */

#if POOL_STATS2
  POOL_SLOCK(pl,&pp->p_statslock, pldisk);

  zero = -1;
  for (i=0; i<POOL_DSTATS_MAX; i++) {
    if ((zero == -1) && ((pp->p_dstats)[i].s_dev == 0)) {
      zero = i;	
      continue;
    }

    if (*rdev == (pp->p_dstats)[i].s_dev)
      break;
  }

  if (i == POOL_DSTATS_MAX) {
    if (zero == -1) {
      printk("poolmap: No such device (%u) %d/%d\n", *rdev, 
	     MAJOR(*rdev), MINOR(*rdev));
    } else {
      (pp->p_dstats)[zero].s_dev = *rdev;
      (pp->p_dstats)[zero].s_out_reads++;
    }
  } else {
    (pp->p_dstats)[i].s_out_reads++;
  }
	
  POOL_SUNLOCK(&pp->p_statslock, pl);
#endif

  return(0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_ioctl
**
** DESCRIPTION
**   Implements all pool ioctl's.  This includes dlock/unlock, assemble/remove,
**   statistics gathering, and other miscellaneous ioctls.  Note there should
**   be only one retrn from this function to handle the return protocol
**   correctly.  If an ioctl causes an internal routine to be called, the
**   parameters are copied to kernel space before the call is made.  NOTE:
**   structures that are in user space and are pointed to by the structure
**   passed as the ioctl arguement are not copied into kernel space before the
**   call.  Some ioctls will only work on the /dev/poolcomm device (minor=0)
**
** RETURN VALUE
**   0 if OK, error condition on error.
**
*******************************************************************************
******************************************************************************/

int pool_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	       unsigned long arg) 
{
  kdev_t      	dev 	= inode->i_rdev;
  int         	minor 	= MINOR(dev);
  pool_t	*pp    	= NULL;
  int           rc 	= 0;
  int		pl    	= 0;
  int	        s     	= 0;
  int	        i;
  int	        len;

  IFDEBUG(POOL_DEBUG1,
	  printk("Entered poolioctl()\n");
	  printk("  Device:  (%u) %d/%d\n", dev, MAJOR(dev), MINOR(dev));
	  printk("  Command: 0x%x\n", cmd););


  if (minor < 0 || minor >= POOL_MAX_DEVICES)
    return (-ENXIO);


  pp = &pools[minor];



  POOL_SLOCK(pl,&pp->p_mutex, pldisk);
  if (pp->p_flags & PF_FREE) {
    POOL_SUNLOCK(&pp->p_mutex, pl);
    return (-ENXIO);
  }
  ASSERT(POOL_BUSY(pp));
  ASSERT(pp->p_flags & PF_VALID);
  POOL_SUNLOCK(&pp->p_mutex, pl);



  /* Do commands */

  switch(cmd) {


#if (DLOCK_VERSION == 2)
  case POOL_LOCK:
  case POOL_UNLOCK:
  case POOL_UNLOCK_INCR: 
  {
    dlock_req_t		lock;
    dlock_params_t	params;
    
    if ((s = gencopy(&lock, (void *)arg, sizeof(lock)))) {
      printk("poolioctl: error copying in data.\n");
      return (-EFAULT);
    }
      
    params.dp_action = cmd;
    params.dp_req = &lock;
          
      
    if ((s = lock_mapping(pp, &params))) {
      return s;
    }
      
    IFDEBUG(POOL_DEBUG_LOCK2,
	    printk("  action  = 0x%x\n", params.dp_action);
	    printk("  logln   = %u\n", params.dp_req->l_id);
	    printk("  phyln   = %u\n", params.dp_phyln);
	    printk("  devnum  = %u\n", params.dp_devnum);
	    printk("  major   = %d\n", MAJOR(params.dp_dldev->dl_dev));
	    printk("  minor   = %d\n", MINOR(params.dp_dldev->dl_dev)););
      
      
    rc = pool_dlock(pp, &params);

    if ((s = gencopy((void *)arg, &lock, sizeof(lock)))) {
      if (rc == 0)
	rc = -EFAULT;
    }
    break;
  }
#endif


  case POOL_CLEAR_STATS:
  {
    IFDEBUG(POOL_DEBUG2,
	    printk("  Clearing statistics...\n"););
    
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);

    memset(&(pp->p_stats), 0, sizeof(pool_stat_t));
    memset(&(pp->p_dlstats), 0, sizeof(pool_dlstat_t));

    for (i=0; i<POOL_DSTATS_MAX; i++) {
      (pp->p_dstats)[i].s_out_reads = 0;
      (pp->p_dstats)[i].s_out_writes = 0;
    }

    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
  }

  case POOL_DUMP_TOTALSTATS:
  {
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);
    
    if (copyout(&(pp->p_stats), (void *)arg, sizeof(pool_stat_t)))
      rc = -EFAULT;
    
    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
  }

  case POOL_DUMP_DISKSTATS:
  {
#if POOL_STATS2
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);
    
    for (i=0; i<POOL_DSTATS_MAX; i++) {
      if ((pp->p_dstats)[i].s_dev == 0)	
	break;
    }

    if ((i == 0)||(i == POOL_DSTATS_MAX))
      len = 0;			
    else
      len = (int)(i*sizeof(pool_dstat_t));			
    
    if (copyout(pp->p_dstats, (void *)arg, len))
      rc = -EFAULT;

    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
#endif
  }

  case POOL_DUMP_LOCKSTATS:
  {
    POOL_SLOCK(pl,&pp->p_statslock, pldisk);

    if (copyout(&(pp->p_dlstats), (void *)arg, sizeof(pool_dlstat_t))) {
      rc = -EFAULT;
    }

    POOL_SUNLOCK(&pp->p_statslock, pl);
    break;
  }

  case BLKGETSIZE: 
  {
    if (!arg) 
      return (-EINVAL);

    if (copyout(&(pp->p_capacity), (void *)arg, sizeof(uint64)))
      rc = -EFAULT;
    break;       
  }

  /* Linux specific IOCTL's */

  case HDIO_GETGEO:
    printk("poolioctl: HDIO_GETGEO\n");
    return -ENXIO;
  case BLKRASET:
    printk("poolioctl: BLKRASET\n");
    return -ENXIO;
  case BLKRAGET:
    printk("poolioctl: BLKRAGET\n");
    return -ENXIO;
  case BLKFLSBUF:
    printk("poolioctl: BLKFLSBUF\n");
    return -ENXIO;
  case BLKRRPART:
    printk("poolioctl: BLKRRPART\n");
    return -ENXIO;
   
    RO_IOCTLS(dev, arg);

  case POOL_DLOCK_COUNT:
  {    
    if (copyout(&(pp->p_dlocks), (void *)arg, sizeof(uint))) {
      rc = -EFAULT;
    } else
      rc = sizeof(uint);

    break;
  }

  case POOL_DEBUG_ON:  /*  Turn on debug bits.  */
  {
    uint64 x;
    
    if (copyin((void *)arg, &x, sizeof(uint64))) {
      return (-EFAULT);
    }
#ifdef DEBUG
    printk("Pool driver debugging enabled.\n");
    printk("  Flags: 0x%16"FMT_64"X + 0x%16"FMT_64"X = ", pool_dflags, x);
#endif
    pool_dflags |= x;
#ifdef DEBUG
    printk("0x%16"FMT_64"X\n", pool_dflags);
#endif
    break;
  }

  case POOL_DEBUG_OFF:  /*  Turn off debug bits.  */
  {
    uint64 x;
	  
    if (copyin((void *)arg, &x, sizeof(uint64))) {
      return -EFAULT;
    }

#ifdef DEBUG
    printk("Pool driver debugging disabled.\n");
    printk("  Flags:  0x%16"FMT_64"X - 0x%16"FMT_64"X = ", pool_dflags, x);
#endif
    pool_dflags &= ~x;
#ifdef DEBUG
    printk("0x%16"FMT_64"X\n", pool_dflags);
#endif
    break;
  }

  case POOL_ADD: 
  {
    pool_add_t	add;

    if (minor) {
      return (-EINVAL);
    }

    if (copyin((void *)arg, &add, sizeof(add)) != 0) {
      return (-EFAULT);
    }
    
    rc = add_pool(&add);

    minor = MINOR(add.pa_pool_dev);

    pool_size[minor] = (pools[minor].p_capacity >> 1);

    if (copyout(&add, (void *)arg, sizeof(add)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }

  case POOL_REMOVE:
  {
    uint32 m;   /* minor number of device to remove */

    if (minor) { 
      return (-EINVAL);
    }

    if (copyin((void *)arg, (void *)&m, sizeof(m)) != 0) {
      rc = -EFAULT;
      break;
    }

    rc = remove_pool(m);
    break;
  }

  case POOL_LIST:
  {
    pool_list_t	list;
    
    if (minor) {
      return (-EINVAL);
    }

    if (copyin((void *)arg, (void *)&list, sizeof(list)) != 0) {
      return (-EFAULT);
    }

    rc = list_pool(&list);

    if (copyout((void *)&list, (void *)arg, sizeof(list)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }

  case POOL_SPLIST:
  {
    pool_splist_t list;

    if (copyin((void *)arg, &list, sizeof(list)) != 0) {
      return -EFAULT;
    }

    if (list.spl_count <= 0) {
      rc = -EFAULT;
      break;
    }

    rc = list_spool(pp, &list);

    if (copyout((void *)&list, (void *)arg, sizeof(list)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }

  case POOL_IND_CACHE:
  {
    fsync_dev(dev);
    invalidate_buffers(dev);
    break;
  }

  default:
    rc = -EINVAL;
  }

  return (rc);
}


/******************************************************************************
*******************************************************************************
** NAME
**	bdopen
**
** DESCRIPTION
**	Open a lower level block device...
**
** RETURN VALUE
**	Unix errno
**
*******************************************************************************
******************************************************************************/

int bdopen(kdev_t dev, struct inode **inodep, struct file **filep)
{
  int error;

  IFDEBUG(POOL_DEBUG_OPEN,
	  printk("Entering bdopen\n"););

  if (!inodep || !filep) {
    panic("pool: NULL pointer %d\n", __LINE__);
  }
  *filep = get_empty_filp();
  if (!*filep)
    return (-ENFILE);
  *inodep = get_empty_inode();

  (*inodep)->i_dev = dev;
  (*inodep)->i_rdev = dev;
  (*inodep)->i_op=&blkdev_inode_operations;
  (*filep)->f_op=blkdev_inode_operations.default_file_ops;

  if ((*filep)->f_op->open) {
    error = (*filep)->f_op->open(*inodep, *filep);
  } else {
    panic("pool: blkdev_inode_operations has no open member!\n");
    error = 0;	/* avoid compiler warnings */
  }
      
  if (error) {
    printk("pool: open_devices: open failed: %d\n", -error);
    put_filp(*filep);
    clear_inode(*inodep);
    *filep = NULL;
    *inodep = NULL;
  }

  return (error);
}


/******************************************************************************
*******************************************************************************
** NAME
**	bdclose
**
** DESCRIPTION
**	Close a lower level block device
**
** RETURN VALUE
**	Unix errno
**
*******************************************************************************
******************************************************************************/

int bdclose(struct inode **inodep, struct file **filep)
{
  int error;

  IFDEBUG(POOL_DEBUG_OPEN,
	  printk("Entering bdclose\n"););

  if (!inodep || !filep) {
    panic("pool: NULL pointer %d\n", __LINE__);
  }
  if (*inodep && *filep && (*filep)->f_op && (*filep)->f_op->release) {
    error = (*filep)->f_op->release(*inodep, *filep);
    put_filp(*filep);
    clear_inode(*inodep);
  } else {
    panic("pool: NULL pointer %d\n", __LINE__);
  }
  *filep = NULL;
  *inodep = NULL;

  return (error);
}


/******************************************************************************
*******************************************************************************
** NAME
**	gencopy
**
** DESCRIPTION
**	i386   TASK_SIZE default:  0xC0000000
**	alpha  TASK_SIZE default:  0x40000000000		       
**
** RETURN VALUE	
**	number of bytes *not* transfered
**
*******************************************************************************
******************************************************************************/

unsigned long gencopy(void *to, void *from, unsigned long size)
{
  if ((unsigned long)from <= TASK_SIZE){    			
    if ((unsigned long)to <= TASK_SIZE)
      memcpy(to, from, size);			       	/* user -> user */
    else
      return copy_from_user(to, from, size);		/* user -> kernel */
  }
  else {
    if ((unsigned long)to <= TASK_SIZE)
      return copy_to_user(to, from, size);		/* kernel -> user */
    else
      memcpy(to, from, size);				/* kernel -> kernel */
  }
  return 0;
}



static struct file_operations pool_fops=
{
  NULL,
  pool_read,
  pool_write,
  NULL,
  NULL,
  pool_ioctl,
  NULL,
  pool_open,
  NULL,
  pool_release,
  block_fsync
};


/******************************************************************************
*******************************************************************************
** NAME
**	pool_init
**
** DESCRIPTION
**	This routine is called by init_module().  It does the 
**   important elements required for Linux block driver initialization
**   including registering itself as a block device in the kernel and
**   setting up the blk_dev[] function pointers.  Pool locks are also
**   initialized. 
**
** RETURN VALUE
**	zero if OK
**
*******************************************************************************
******************************************************************************/

static int pool_init(void)
{
  int i;

  printk(KERN_INFO "Pool v%d.%d.%d (%s release) (%s %s) installed\n",
	 POOL_VERSION >> 16, (POOL_VERSION >> 8) & 0xFF, (POOL_VERSION) & 0xFF,
	 GFS_RELEASE_NAME,
	 __DATE__, __TIME__);

  if (register_blkdev(POOL_MAJOR, "pool", &pool_fops)) {
    printk ("Unable to get major %d for pool\n", POOL_MAJOR);
    return (-1);
  }

  blk_dev[POOL_MAJOR].request_fn=do_pool_request;
  blk_dev[POOL_MAJOR].makerq_fn=NULL;
  blk_dev[POOL_MAJOR].map_fn=pool_map;
  blk_dev[POOL_MAJOR].current_request=NULL;
  read_ahead[POOL_MAJOR]=INT_MAX;

  for (i=0; i < POOL_MAX_DEVICES; i++) {
    pool_blocksizes[i] = 512;
    pool_maxreadahead[i] = 0;
    pool_size[i] = 0;
  }

  blksize_size[POOL_MAJOR]=pool_blocksizes;
  blk_size[POOL_MAJOR]=pool_size;
  max_readahead[POOL_MAJOR]=pool_maxreadahead;

  /* initialize pools[] array */
  memset(pools, 0, POOL_MAX_DEVICES * sizeof(pool_t)); 
 
  for (i=0; i < POOL_MAX_DEVICES; i++) {
    spin_lock_init(&pools[i].p_mutex);
    spin_lock_init(&pools[i].p_statslock);
    spin_lock_init(&pools[i].p_dlmutex);
    pools[i].p_flags = PF_VALID | PF_FREE;
  }
  
  pools[0].p_flags = PF_VALID;

  /* Initialize the debugging flags  */
  
  pool_dflags = (0ll);

  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	init_module
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

#ifdef MODULE
int init_module(void)
{
  int rc;

  rc = pool_init();
  if (rc)
    return rc;

  return(0);
}
#endif


/******************************************************************************
*******************************************************************************
** NAME
**	cleanup_module
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

#ifdef MODULE
void cleanup_module(void)
{
  int i, j, k;
  int pl;

  printk(KERN_INFO "Pool uninstalled\n");

  for (i=0; i < POOL_MAX_DEVICES; i++) {
    POOL_SLOCK(pl, &pools[i].p_mutex, hi);
    if (pools[i].p_flags & PF_FREE) {
      POOL_SUNLOCK(&pools[i].p_mutex, pl);
      continue;
    }

    /* sync operations to flush any dirty buffers */

    for (j=0; j < pools[i].p_num_spools; j++) {
      for (k=0; k < pools[i].p_spools[j].sp_dldevs; k++) {
	fsync_dev(pools[i].p_spools[j].sp_dldevlist[k].dl_dev);
      }
    }

    for (j=0; j < pools[i].p_num_spools; j++) {
      for (k=0; k < pools[i].p_spools[j].sp_devs; k++)
	fsync_dev(pools[i].p_spools[j].sp_devlist[k].d_dev);
    }

    if (!(pools[i].p_flags & PF_VALID)) {
      POOL_SUNLOCK(&pools[i].p_mutex, pl);
      printk("ARRGG unloading module while still in use\n");
    }

    if (pools[i].p_flags & PF_LOWEROPEN) {
      close_devices(&pools[i]);
    }
    
    free_pool(&pools[i]);
    pools[i].p_flags = PF_VALID | PF_FREE;
    POOL_SUNLOCK(&pools[i].p_mutex, pl);
  }

  unregister_blkdev(POOL_MAJOR, "pool");

  blk_dev[POOL_MAJOR].request_fn=NULL;
  blk_dev[POOL_MAJOR].makerq_fn=NULL;
  blk_dev[POOL_MAJOR].map_fn=NULL;
  read_ahead[POOL_MAJOR] = 0;
  blk_size[POOL_MAJOR] = NULL;
  blksize_size[POOL_MAJOR] = NULL;
}
#endif



EXPORT_SYMBOL(gencopy);
EXPORT_SYMBOL(pool_ioctl);





