/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Andrew Barry <barry@lcse.umn.edu>
** 
**    Module name: gfs_ipd
**
**    Description: partition_opps.c controls the state behavior for 
**    the ip dlock server 
**
*******************************************************************************
******************************************************************************/


#ifndef __PARTITION_OPPS_C__
#define __PARTITION_OPPS_C__

#include "partition_opps.h"
#include "print.h"
#include <assert.h>

/*
  This file includes most of the serv'ers behavior. 
Most of these functions are named identically to their 
name in the dlock specification of the scsi dlock command. 
*/

/*******************************************************
 *All actions passed with the dlock command descriptor block
 *are handled here. This function merely checks to see if the
 *lock involved is expired, decodes the command field, and 
 *launches one of the functions bellow that actually deal
 *with it. 
 *******************************************************/
/*extern int sig_int_flag;*/ 



void part_actions(partition_locks *pal, 
		  dlock_cdb *cdb, 
		  dlock_rdf *rdf, 
		  send_req_t *sr){

  if((cdb->lock_number >= pal->dlocks_per_partition) && 
     (cdb->lock_number!=0xffffffff)){
    bad_lock_number         (pal, cdb, rdf);
  }
  else{
    if(cdb->lock_number!=0xffffffff){
      deal_expired          (pal, cdb, rdf);
    }
    if(cdb->lock_number!=0xffffffff){
      pthread_mutex_lock(&(pal->sdl[cdb->lock_number]->pmut));
    }
    switch(cdb->action){
    case 0:
      no_operation          (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 1:
      lock_shared           (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 2:
      lock_exclusive        (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 3:
      force_lock_exclusive  (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 4:
      touch_lock            (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 5:
      unlock                (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 6:
      unlock_increment      (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 7:
      activity_on           (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    case 8:
      activity_off          (pal, cdb, rdf, sr);
      all_actions           (pal, cdb, rdf);
      break;
    default:
      break;
    }
    if(cdb->lock_number!=0xffffffff){
      pthread_mutex_unlock(&(pal->sdl[cdb->lock_number]->pmut));
    }
  }
  return;  
}
  

/****************************************************************
 *unique to mode sense and mode select commands
 ****************************************************************/
void  mode_actions       (partition_locks *pal, mode_cdb  *cdb,
			  mode_page_rdf *rdf){

  if(cdb->code==CDBMODER){
    rdf->max_clients=pal->max_lock_holds;
    rdf->num_locks=pal->dlocks_per_partition;
    rdf->lock_time_ms=pal->lock_timeout_interval;  
  }
  else{
    pal->max_lock_holds=cdb->max_clients;
    pal->dlocks_per_partition=cdb->num_locks;
    pal->lock_timeout_interval=cdb->lock_time_ms;
    rdf->max_clients=pal->max_lock_holds;
    rdf->num_locks=pal->dlocks_per_partition;
    rdf->lock_time_ms=pal->lock_timeout_interval;
  }  
}

/****************************************************************
 *unique to report expired commands
 ****************************************************************/
void expired_actions     (partition_locks *pal, dlock_cdb *cdb, 
			  rexpired_rdf *rex){
  int i;
  time_t currenttime, *ctp;
  ctp=NULL;
  for(i=0; i<pal->dlocks_per_partition; i++){
    currenttime=time(ctp);
    if((pal->sdl[i]->expire_time < (int)currenttime)
       && (pal->sdl[i]->state != STATE_UNLOCKED)){
      pal->sdl[i]->expired=pal->sdl[i]->state;
      pal->sdl[i]->state=STATE_UNLOCKED;
      pal->sdl[i]->version_number++;
      sdl_clear_cid(pal->sdl[cdb->lock_number]);
    }
    rex->return_length=(pal->dlocks_per_partition)/8; 
    if(pal->sdl[i]->expired!=EXPIRED_NOT){
      sdl_clear_q(pal->sdl[i]);
      re_rdf_set_exlock(rex, i); 
    }
    else{
      re_rdf_clear_exlock(rex, i); 
    }
  }
  report_expired(pal, cdb, rex);
} 

/********************************************************
 *as the name implies. This one does nothing. 
 ********************************************************/
void no_operation(partition_locks *pal, dlock_cdb *cdb, 
		  dlock_rdf       *rdf, 
		  send_req_t      *sr){
  rdf->result=1;}


/*******************************************************************
 * -If the lock is expired_exclusive lock exclusively so that the 
 *initiator can do any error correction.
 * -If the lock is unlocked && not expired_exclusive, share-lock it. 
 *Add the initiator's cid to the cid lis. reset the expire time. 
 * -If the lock is locked_shared && room in the cid list. Add 
 *the initiator's cid to the cid list. Reset the expire time. 
 * -If there is no room in the cid list then return a failure result.
 * -If the lock is locked_exclusive not held by the initiator, nil. 
 *a failure result.
 * -If locked exclusive held by this initiator, reset expire time 
 *and change to locked_shared.
 *******************************************************************/
void lock_shared(partition_locks *pal, dlock_cdb *cdb, 
		 dlock_rdf       *rdf, 
		 send_req_t      *sr){
  server_dlock *lockp;
  int i, c;

  lockp=pal->sdl[cdb->lock_number];
  if(lockp->state==STATE_UNLOCKED){
    rdf->result=1;
    if(lockp->expired==EXPIRED_E){
      lockp->state=STATE_LOCKED_E;
    }
    else{
      lockp->state=STATE_LOCKED_S;
    }
    add_dlock_cid(lockp, cdb->cid, sr);
    sdl_reset(lockp);
  }
    
  else if(lockp->state==STATE_LOCKED_S){
    if(lockp->num_holders<pal->max_lock_holds){
      rdf->result=1;
      add_dlock_cid(lockp, cdb->cid, sr);
      sdl_reset(lockp);
    }
    else{
      rdf->result=0;
      for(i=0;i<lockp->num_holders;i++){
	c=send_drop(cdb->lock_number, CB_NEED_S, (get_cid_list(i, lockp)->soc));
	if(c<0){
	  dbprintf("connection failed");}
      }
    }
  }  
  
  else{
    if(cid_in_dlock(lockp, cdb->cid)){
      rdf->result=1;
      lockp->state=STATE_LOCKED_S;
      if(lockp->qhead->state==STATE_LOCKED_S){ 
	c=send_drop(cdb->lock_number, CB_FREED_S, lockp->qhead->soc);
	if(c<0){
	  dbprintf("connection failed");}
      }
      sdl_reset(lockp);
    }
    else{
      rdf->result=0;
      if(lockp->qsize==0){
	add_q_cid(lockp, cdb->cid, STATE_LOCKED_S, sr);
      }
      c=send_drop(cdb->lock_number, CB_NEED_S, get_cid_list(0, lockp)->soc);
      if(c<0){
	dbprintf("connection failed");}
    }
  }
  dbprintf("q_size:%d\n",lockp->qsize);
}


/*****************************************************************
 * -If lock is unlocked, lock it exclusively, add initiator's cid
 *to the cid list and reset the expire time.
 * -If lock is locked_shared && held by only this initiator, lock
 *exclusively and reset the expire time.
 * -If lock is locked shared by any other initiators, return failure.
 * -If lock is locked exclusively return failure.
 *****************************************************************/
void lock_exclusive(partition_locks *pal, dlock_cdb *cdb, 
		    dlock_rdf       *rdf, 
		    send_req_t      *sr){
  server_dlock *lockp;
  int i,c;

  lockp=pal->sdl[cdb->lock_number];
  if(lockp->state==STATE_UNLOCKED){
    rdf->result=1;
    lockp->state=STATE_LOCKED_E;
    add_dlock_cid(lockp, cdb->cid, sr);
    sdl_reset(lockp);
  }

  else if(lockp->state==STATE_LOCKED_S){
    if(lockp->num_holders==1 &&
	 cid_in_dlock(lockp, cdb->cid)) {
      rdf->result=1;
      lockp->state=STATE_LOCKED_E;
      sdl_reset(lockp);
    }

    else{
      rdf->result=0;
      if(lockp->qsize==0){
	add_q_cid(lockp, cdb->cid, STATE_LOCKED_E, sr);
      }
      for(i=0; i<lockp->num_holders; i++){
	assert( get_cid_list(i, lockp)->soc != NULL );
	c=send_drop(cdb->lock_number, CB_NEED_E, (get_cid_list(i, lockp)->soc));
	if(c<0){dbprintf("connection failed");}
      }
    }
  }
  else{   
    rdf->result=0;
    if(lockp->qsize==0){
      add_q_cid(lockp, cdb->cid, STATE_LOCKED_E, sr);
    }
    c=send_drop(cdb->lock_number, CB_NEED_E, (get_cid_list(0, lockp)->soc));
    if(c<0){dbprintf("connection failed");}
  }
  dbprintf("q_size:%d\n",lockp->qsize);
}

/***********************************************************************
 * -If unlocked, add cid to list, lock exclusive, reset expire time. 
 * -If locked and input version is the same as the lock version
 *number, clear cid list, add cid, lock exclusively, reset the
 *the expire time, increment the version number.
 * -If locked but the input version number is not the same as
 *the lock version number return failure.
 ***********************************************************************/
void force_lock_exclusive(partition_locks *pal, dlock_cdb *cdb, 
			  dlock_rdf       *rdf, 
			  send_req_t      *sr){
  server_dlock *lockp;

  lockp=pal->sdl[cdb->lock_number];
  if(lockp->state==STATE_UNLOCKED){
    rdf->result=1;
    lockp->state=STATE_LOCKED_E;
    add_dlock_cid(lockp, cdb->cid, sr);
    sdl_reset(lockp);
  }
  
  else if(((lockp->version_number << 8) >> 8)==
	  ((cdb->input_version_number << 8) >> 8)){
    rdf->result=1;
    lockp->expired=pal->sdl[cdb->lock_number]->state;
    lockp->state=STATE_LOCKED_E;
    sdl_clear_cid(lockp);
    sdl_clear_q(lockp);
    add_dlock_cid(lockp, cdb->cid, sr);
    lockp->version_number++;
    sdl_reset(lockp);
  }

  else{
    rdf->result=0;
    if(lockp->qsize==0){
      add_q_cid(lockp, cdb->cid, STATE_LOCKED_E, sr);
    }
    send_drop(cdb->lock_number, CB_NEED_E, get_cid_list(0, lockp)->soc);
  }
}

/**************************************************************
 * -If unlocked return failure.
 * -If locked && the cid is in the list && lock number 
 *is 0xffffffff touch all locks held by that cid.
 * -If locked && the cid is in the list reset expire time.
 * -If locked && the cid is not in the list return failure.
 **************************************************************/
void touch_lock(partition_locks *pal, dlock_cdb *cdb, 
		dlock_rdf       *rdf, 
		send_req_t      *sr){
  server_dlock *lockp;
  int i;

  if(cdb->lock_number==0xffffffff){
    for(i=0; i<pal->dlocks_per_partition; i++){
      if(cid_in_dlock(pal->sdl[i], cdb->cid)){
	sdl_reset(pal->sdl[i]);
      }
    }
  }
  else{
    lockp=pal->sdl[cdb->lock_number];
    if(lockp->state==STATE_UNLOCKED){
      rdf->result=0;}
    
    else if(cid_in_dlock(lockp, cdb->cid)){
      rdf->result=1;
      sdl_reset(lockp);
    }
    
    else{
      rdf->result=0;
    }
  }
}


/*********************************************************
 * -If the lock is unlocked return failure.
 * -If locked && the cid is in the list, remove the cid from
 *the list, unlock if no more holders, increment version if
 *activity is on. 
 * -If cid not in list return failure.
 *********************************************************/
void unlock(partition_locks *pal, 
	    dlock_cdb       *cdb, 
	    dlock_rdf       *rdf, 
	    send_req_t      *sr){
  server_dlock *lockp;

  lockp=pal->sdl[cdb->lock_number];
  if(lockp->state==STATE_UNLOCKED){
    rdf->result=0;}
  else if(cid_in_dlock(lockp, cdb->cid)){
    rdf->result=1;
    rem_dlock_cid(lockp, cdb->cid);
    if(lockp->act==ACTIVITY_ON){
      lockp->version_number++;}
    if(lockp->num_holders==0){
      lockp->state=STATE_UNLOCKED;
      if(lockp->qsize!=0){
	send_drop(cdb->lock_number, CB_FREED_E, lockp->qhead->soc);
	rem_q_cid(lockp, get_q_list(0, lockp)->cid);
      }
    }
    else{
      if((lockp->qsize!=0)&&(lockp->qhead->state==STATE_LOCKED_S)){
	send_drop(cdb->lock_number, CB_FREED_S, lockp->qhead->soc);
	rem_q_cid(lockp, get_q_list(0, lockp)->cid);
      }
    }
  }
  else{
    rdf->result=0;}
}


/*******************************************************************
 * Behaves exactly like unlock, except that it will increment the
 * version number regardless of the activity bit's state. 
 *******************************************************************/
void unlock_increment(partition_locks *pal, 
		      dlock_cdb       *cdb, 
		      dlock_rdf       *rdf, 
		      send_req_t      *sr){
  server_dlock *lockp;

  lockp=pal->sdl[cdb->lock_number];
  if(lockp->state==STATE_UNLOCKED){
    rdf->result=0;}
  
  else if(cid_in_dlock(lockp, cdb->cid)){
    rdf->result=1;
    rem_dlock_cid(lockp, cdb->cid);

    lockp->version_number++;
    if(lockp->num_holders==0){
      lockp->state=STATE_UNLOCKED;
      if(lockp->qsize!=0){
	send_drop(cdb->lock_number, CB_FREED_E, lockp->qhead->soc);
	rem_q_cid(lockp, get_q_list(0, lockp)->cid);
	assert(lockp->qhead==NULL);/**/
      }
    }
    else{
      if((lockp->qsize!=0)&&(lockp->qhead->state==STATE_LOCKED_S)){
	send_drop(cdb->lock_number, CB_FREED_S, lockp->qhead->soc);
	rem_q_cid(lockp, get_q_list(0, lockp)->cid);
      }
    }    
  }
  else{
    rdf->result=0;}
  dbprintf("\t\tqsize:%d\n", lockp->qsize);
}


/***********************************************************
 *set the activity bit of the lock.
 ***********************************************************/
void activity_on(partition_locks *pal, dlock_cdb *cdb, 
		  dlock_rdf *rdf, send_req_t      *sr){
  rdf->result=1;
  pal->sdl[cdb->lock_number]->act=ACTIVITY_ON;
 }



/***************************************************************
 *clear the activity bit of the lock. 
 ***************************************************************/
void activity_off(partition_locks *pal, dlock_cdb *cdb, 
		   dlock_rdf *rdf, send_req_t      *sr){
  rdf->result=1;
  pal->sdl[cdb->lock_number]->act=ACTIVITY_OFF;
  pal->sdl[cdb->lock_number]->version_number++;
}



/*****************************************************************
 *searches the entire list of locks and checks each one to see if 
 *it has expired. Sets exp to report all
 *of the expired locks. Also sets the expired field of the lock. If
 *not expired clears bits in lock and exp.
 *
 *broken still
 *****************************************************************/ 
void report_expired(partition_locks *pal, 
		    dlock_cdb *cdb, 
		    rexpired_rdf *exp){
 
  dbprintf("report_expired\n");
}


/***************************************************************
 *checks the lock to see if it has expired. sets the expired bit
 *of the lock if it has. 
 ***************************************************************/
void deal_expired(partition_locks *pal, dlock_cdb *cdb, 
		  dlock_rdf       *rdf){

  int i;
  time_t currenttime, *ctp;
  server_dlock *lockp;

  lockp=pal->sdl[cdb->lock_number];
  ctp=NULL;
  currenttime=time(ctp);
  i=cdb->lock_number;
  if((lockp->expire_time < (int)currenttime)
     && (lockp->state != STATE_UNLOCKED)){
    lockp->expired=lockp->state;
    lockp->state=STATE_UNLOCKED;
    lockp->version_number++;
    sdl_clear_cid(lockp);
    sdl_clear_q(lockp);
  } 
} 


/**************************************************************
 *every action does this. sets the return data based on the 
 *lock status. including returning all the cidames.
 *zero return values for touchlock(all);
 **************************************************************/
void all_actions(partition_locks *pal, dlock_cdb *cdb, 
		 dlock_rdf       *rdf){
  server_dlock *lockp;

  if(cdb->lock_number==-1){
    rdf->act =           0;
    rdf->state =         0;
    rdf->version_number= 0;
    rdf->expired =       EXPIRED_NOT;
    rdf->num_holders=    0;
  }
  else{
    lockp=pal->sdl[cdb->lock_number];
    rdf->act =           lockp->act;
    rdf->state =         lockp->state;
    rdf->version_number= lockp->version_number;
    rdf->expired=        lockp->expired;
    rdf->num_holders=    lockp->num_holders;
  }
  return;
}
   
/***************************************************************
 *if the lock number is out of bounds. Error case.
 ***************************************************************/
void bad_lock_number(partition_locks *pal, dlock_cdb *cdb, 
		     dlock_rdf       *rdf){
  rdf->result=0;
  rdf->act=0;
  rdf->expired=0;
  rdf->state=0;
  rdf->version_number=0;
}


/********************************************************
 *
 ********************************************************/
int send_drop(u32 lock_number, u32 flags, send_req_t *sr){
  drop_rdf drp;
  int c;

  drp.code=CALLBACK;
  drp.lock_number=htonl(lock_number);
  drp.flags=htonl(flags);
  c=send_request(sr, (char *)&drp, sizeof(drop_rdf));
  if(c<0)dbprintf("\nOh!!! Fuck!!!!\n");
  /*dbprintf("\tlock:%u\tflags:%x\n",drp.lock_number, drp.flags);*/
  return c;
}



#endif /* __PARTITION_OPPS_C__ */


