/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Manish M. Agarwal <agarwal@lcse.umn.edu>
** 
**    Module name: gfs_tool.c
**
**    Description:
**      Utility functions - used by gfs_tool and gfs_balancer.
**
*******************************************************************************
******************************************************************************/

#ifdef linux
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#endif

#include "global.h"
#include "gfs_ioctl.h"

/******************************************************************************
*******************************************************************************
**
** void PrintParams(gfs_lnkinfo_t * param)
**
** Description:
**   This routine is used for debugging. It prints all the paramters as 
**   read from the input file.
**
** Input(s):
**   param - pointer to the initialised gfs_lnkinfo_t structure.
**
** Returns:
**   void
**
*******************************************************************************
******************************************************************************/

void PrintParams(gfs_lnkinfo_t * param)
{
  int i;

  fprintf(stdout, "\nDestination Subpool = %"FMT_32"d\n", param->d_sp);
  fprintf(stdout, "Number of Links     = %"FMT_32"d\n", param->lnk_cnt);
  fprintf(stdout, "The Links are :\n\n");


  for(i = 0; i < param->lnk_cnt; i++)
    {
      fprintf(stdout, "%d. ", i+1);
      fprintf(stdout, "file = %s",param->l_name[i].dname);
      fprintf(stdout,"/");
      fprintf(stdout, "%s\n\n",param->l_name[i].fname);
    }

  return;
}



/******************************************************************************
*******************************************************************************
**
** void fill_params()
**
** Description:
**   This routine takes a filename (with or without path) and separates its
**     	it into filename and pathname.
**
** Input(s):
**   istr - filename (with or without path).
**
** Output:
**   lnk  - gfs_lnkname_t structure pointer.
**
** Returns:
**   void
**
*******************************************************************************
******************************************************************************/

void fill_params(gfs_lnkname_t * lnk, char * istr)
{
  int len = strlen(istr);
  int i = len -1;
  int j = 0;


  while((istr[i] != '/') && (i > 0) ) i--;


  if(i==0) /* no directory name */
    {
      lnk->dname[0] = '.';  /* current directory */
      lnk->dname[1] = '\0';
      lnk->dlen = 1;
      strcpy(lnk->fname,istr);
      lnk->flen = len;
      return;
    }

  i++;


  lnk->flen = len -i;
  lnk->dlen = i - 1;


  while(j < lnk->dlen)
    {
      lnk->dname[j] = istr[j];
      j++;
    }
  lnk->dname[j] = '\0';


  j = 0;


  while( j < lnk->flen)
    {
      lnk->fname[j] = istr[i];
      i++;j++;
    }


  lnk->fname[j] = '\0';

  return;
}





/******************************************************************************
*******************************************************************************
**
** int gfs_show_subpools_i()
**
** Description:
**   This routine uses an ioctl command to gather subpool information.
**   PS: Incase of an error - An error message is printed and the allocated
**	 memory is freed.
**
** Input(s):
**   fd - File descriptor for a file on the file system.
**
** Output  :
**   spstatlistp - the subpool information
**
** Returns :
**   int  - 0 if successful.
**
*******************************************************************************
******************************************************************************/

int gfs_show_subpools_i(gfs_spstatlist_t *spstatlistp, int fd)
{
  char		buff[256];
  uint32 	i;
  int 		error;


  spstatlistp->sp_cnt = i = 16;

  do
    {
      i = spstatlistp->sp_cnt;
      

      if((spstatlistp->sp_list = (gfs_spstat_t *)malloc(sizeof(gfs_spstat_t) * i)) != NULL)
	memset((char *)spstatlistp->sp_list, 0, (sizeof(gfs_spstat_t) * i));
      else
	{
	  fprintf(stderr, "%s: unable to allocate memory on line %d.\n", __FILE__, __LINE__); 
	  return(EXIT_FAILURE); 
	}


      if ((error = ioctl(fd,GFS_SHOW_SUBPOOLS,spstatlistp )))
	{
	  strcpy(buff, "*gfs_tool: Error doing ioctl");
	  perror(buff);
	  free(spstatlistp->sp_list);
	  return(EXIT_FAILURE);
	}


      /* print the results onto the screen */
      if (spstatlistp->sp_cnt <= i)
	  break; /* exit the while loop */

      free(spstatlistp->sp_list);

    }  while (i != spstatlistp->sp_cnt); /* try until it succeeds */
  
  return (0);
}



/******************************************************************************
*******************************************************************************
**
** void gfs_mv_file()
**
** Description:
**   This routine actually moves the files. Do not report any errors.
**
** Input(s):
**   filep - file to be moved.
**
*******************************************************************************
******************************************************************************/

void gfs_mv_file(gfs_lnkinfo_t *filep, int verbose)
{
  int 			i,j,fd,error;
  char 			filename[GFS_MAXPATHLEN],buff[256];

  if(verbose)
    PrintParams(filep); 
 

  /* Open the file */
  for(i = 0; i < filep->l_name[0].dlen; i++)
    filename[i] = filep->l_name[0].dname[i];


  filename[i] = '/';
  i++;


  for(j = 0; j < filep->l_name[0].flen; i++,j++)
    filename[i] = filep->l_name[0].fname[j];

  filename[i] = '\0'; 
  

  if ((fd = open(filename, O_RDWR)) < 0) 
    {
      sprintf(buff, "Can't open %s", filename);
      perror(buff);
      return;
    }


  /* ioctl Call */
  if( (error = ioctl(fd,GFS_MOVE_TO_SUBPOOL,filep)) )
    {
      if(verbose)
	{
	  strcpy(buff, "**gfs_tool: Error doing ioctl");
	  perror(buff);
	}

      close(fd);
      return;
    }

  close(fd);
  if(verbose)
    fprintf(stdout, " File Successfully Moved\n\n");

  return;
}

