/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: rbtree.c
**
**    Description:
**	Binary / Red-Black Tree implementation for filesystem checker.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gfsck.h"

extern char *prog_name;

node_t nil;				/* universal nil[T] */



node_t *New_Treenode(uint64 key, char *data)
{
  node_t *x;

  type_zalloc(x, node_t, sizeof(node_t));

  x->key = key;
  x->data = data;
  x->left = x->right = x->p = &nil;
  
  return (x);
}


/*************************
 *    Tree Operations
 *************************/

/* Left and right rotate used by Insert and Delete to preserve RB tree */

void Left_Rotate(node_t **pT, node_t *x)
{
  node_t *y;

  y = x->right;
  x->right = y->left;
  
  if (y->left != &nil)
    y->left->p = x;

  y->p = x->p;
  
  if (x->p == &nil)
    *pT = y;
  else {
    if (x == x->p->left)
      x->p->left = y;
    else
      x->p->right = y;
  }

  y->left = x;
  x->p = y;

  return;
}


void Right_Rotate(node_t **pT, node_t *x)
{
  node_t *y;

  y = x->left;
  x->left = y->right;
  
  if (y->right != &nil)
    y->right->p = x;

  y->p = x->p;
  
  if (x->p == &nil)
    *pT = y;
  else {
    if (x == x->p->right)
      x->p->right = y;
    else
      x->p->left = y;
  }

  y->right = x;
  x->p = y;

  return;
}

  
/* Binary search tree insert used by RB_Insert */

void Tree_Insert(node_t **pT, node_t *z)
{
  node_t *x, *y;

  y = &nil;
  x = *pT;
  
  while (x != &nil) {
    y = x;
    if (z->key < x->key)
      x = x->left;
    else
      x = x->right;    
  }
  z->p = y;

  if (y == &nil) {
    *pT = z;
  }
  else {
    if (z->key < y->key)
      y->left = z;
    else
      y->right = z;
  }
  return;
}


/* Insert x into tree and then rotate/recolor to preserve RB properties */

void RB_Insert(node_t **pT, node_t *x)
{
  node_t *y;

  Tree_Insert(pT, x);

  x->color = RED;

  while ((x != *pT) && (x->p->color == RED)) {
    
    if (x->p == x->p->p->left) {
      y = x->p->p->right;
      if (y->color == RED) {
	x->p->color = BLACK;		/* Case 1 */
	y->color = BLACK;
	x->p->p->color = RED;
	x = x->p->p;
      }
      else {
	if (x->p->right == x) {
	  x = x->p;			/* Case 2 */
	  Left_Rotate(pT, x);
	}
	x->p->color = BLACK;	      	/* Case 3 */
	x->p->p->color = RED;

	Right_Rotate(pT, x->p->p);
      }
    }
    else {				/* Symmetric with above */
      y = x->p->p->left;
      if (y->color == RED) {
	x->p->color = BLACK;
	y->color = BLACK;
	x->p->p->color = RED;
	x = x->p->p;
      }
      else {
	if (x->p->left == x) {
	  x = x->p;
	  Right_Rotate(pT, x);
	}
	x->p->color = BLACK;
	x->p->p->color = RED;
	Left_Rotate(pT, x->p->p);
      }
    }
  }

  (*pT)->color = BLACK;

  return;
}


/* Delete node z from ordinary binary tree and return z */

node_t *Tree_Delete(node_t **pT, node_t *z)
{
  node_t *y, *x;
  uint64 tmpk;
  char *tmpd;

  if ((z->left == &nil) || (z->right == &nil))
    y = z;
  else
    y = Tree_Successor(z);

  if (y->left != &nil)
    x = y->left;
  else
    x = y->right;

  if (x != &nil)
    x->p = y->p;

  if (y->p == &nil)
    *pT = x;
  else {
    if (y == y->p->left)
      y->p->left = x;
    else
      y->p->right = x;
  }

  if (y != z) {
    tmpk = z->key;	/* copying these fields is easier than relinking */
    tmpd = z->data;

    z->key = y->key;
    z->data = y->data;

    y->key = tmpk;
    y->data = tmpd;
  }    

  return (y);
}


/* Delete node z from RB tree and return z */

node_t *RB_Delete(node_t **pT, node_t *z)
{
  node_t *y, *x;
  uint64 tmpk;
  char *tmpd;

  if ((z->left == &nil) || (z->right == &nil))
    y = z;
  else
    y = Tree_Successor(z);

  if (y->left != &nil)
    x = y->left;
  else
    x = y->right;

  x->p = y->p;
  
  if (y->p == &nil)
    *pT = x;
  else {
    if (y == y->p->left)
      y->p->left = x;
    else
      y->p->right = x;
  }

  if (y != z) {
    tmpk = z->key;	/* copying these fields is easier than relinking */
    tmpd = z->data;

    z->key = y->key;
    z->data = y->data;

    y->key = tmpk;
    y->data = tmpd;
  }
 
  if (y->color == BLACK)
    RB_Delete_Fixup(pT, x);

  return (y);
}


/* Restory Red-Black properties to tree x after deletion */

void RB_Delete_Fixup(node_t **pT, node_t *x)
{
  node_t *w;

  while ((x != *pT) && (x->color == BLACK)) {

    if (x == x->p->left) {
      w = x->p->right;
      
      if (w->color == RED) {
	w->color = BLACK;			/* Case 1 */
	x->p->color = RED;
	Left_Rotate(pT, x->p);
	w = x->p->right;
      }
      
      if ((w->left->color == BLACK) && (w->right->color == BLACK)) {
	w->color = RED;				/* Case 2 */
	x = x->p;
      }
      else {
	if (w->right->color == BLACK) {
	  w->left->color = BLACK;		/* Case 3 */
	  w->color = RED;
	  Right_Rotate(pT, w);
	  w = x->p->right;
	}
	w->color = x->p->color;			/* Case 4 */
	x->p->color = BLACK;
	w->right->color = BLACK;
	Left_Rotate(pT, x->p);
	x = *pT;
      }
    }
    else {					/* Symmetric with above */
      w = x->p->left;
      
      if (w->color == RED) {
	w->color = BLACK;
	x->p->color = RED;
	Right_Rotate(pT, x->p);
	w = x->p->left;
      }
      
      if ((w->right->color == BLACK) && (w->left->color == BLACK)) {
	w->color = RED;
	x = x->p;
      }
      else {
	if (w->left->color == BLACK) {
	  w->right->color = BLACK;
	  w->color = RED;
	  Left_Rotate(pT, w);
	  w = x->p->left;
	}
	w->color = x->p->color;
	x->p->color = BLACK;
	w->left->color = BLACK;
	Right_Rotate(pT, x->p);
	x = *pT;
      }
    }
  }
  x->color = BLACK;
}


/* return node succeeding/preceeding x */

node_t *Tree_Successor(node_t *x)
{
  node_t *y;

  if (x->right != &nil)
    return ( Tree_Minimum(x->right) );
  
  y = x->p;

  while ((y != &nil) && (x == y->right)) {
    x = y;
    y = y->p;
  }
  return (y);
}


node_t *Tree_Predecessor(node_t *x)
{
  node_t *y;
  
  if (x->left != &nil)
    return ( Tree_Maximum(x->left) );

  y = x->p;

  while ((y != &nil) && (x == y->left)) {
    x = y;
    y = y->p;
  }
  return (y);
}


/* return the node with minimum/maximum key in tree x */

node_t *Tree_Minimum(node_t *x)
{
  while (x->left != &nil)
    x = x->left;
  
  return (x);
}


node_t *Tree_Maximum(node_t *x)
{
  while (x->right != &nil)
    x = x->right;
  
  return (x);
}


/* print out tree keys inorder (LxR) and preorder (xLR). '*' means RED */

void Inorder_Tree_Walk(node_t *x)
{
  if (x != &nil) {
    Inorder_Tree_Walk(x->left);
    printf("%"FMT_64"d", x->key);
    /* (x->color == RED) ? printf("* ") : printf("  "); */
    printf(" ");
    Inorder_Tree_Walk(x->right);
  }
  return;
}


void Preorder_Tree_Walk(node_t *x)
{
  if (x != &nil) {
    printf("%"FMT_64"d", x->key);
    (x->color == RED) ? printf("* ") : printf("  ");
    Preorder_Tree_Walk(x->left);
    Preorder_Tree_Walk(x->right);
  }
  return;
}    


/* free all memory used by nodes in a tree */

void Free_Tree(node_t *x)
{
  if (x->left != &nil)
    Free_Tree(x->left);
  if (x->right != &nil)
    Free_Tree(x->right);

  if (x->data)
    free(x->data);

  if (x != &nil)
    free(x);

  return;
}


/* free memory of a node */

void Free_Node(node_t *x)
{
  if (x && x->data)
    free(x->data);
  if (x)
    free(x);  
  return;
}


/* returns node with key equal to k */

node_t *Tree_Search(node_t *x, uint64 k)
{
  if ((x == &nil) || (k == x->key))
    return (x);

  if (k < x->key)
    return ( Tree_Search(x->left, k) );
  else
    return ( Tree_Search(x->right, k) );
}


/* returns first node with key greater than or equal to k in tree x */

node_t *Tree_Search_GE(node_t *x, uint64 k)
{
  if ((x != &nil) && (x->key == k))
    return (x);

  if ((x != &nil) && (x->key > k)) {
    if (x->left == &nil)
      return (x);
    else
      return ( Tree_Search_GE(x->left, k) );
  }

  if ((x != &nil) && (x->key < k)) {
    if (x->right == &nil)
      return ( Tree_Successor(x) );
    else
      return ( Tree_Search_GE(x->right, k) );
  }

  return (x);
}
































