/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: hexedit.c
**
**    Description:
**	Dumps file or device data out in hex and ASCII.
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/param.h>

#include "global.h"
#include "gfs_ondisk.h"
#include "hexedit.h"
#include "gfshex.h"

/* Preprocessor Defines */

#define SCREEN          16	       /* print 256 bytes per screen */


/* Function Macros */

#ifdef linux
long long int llseek(int, long long int, int);
#define LLSEEK(x, y, z) llseek((x), (y), (z))
#endif  /*  linux  */

#ifdef sgi
#define LLSEEK(x, y, z) lseek64((x), (y), (z))
#endif  /*  sgi  */


/* Type Definitions */


/* Global Variables */

int	fd;
char   *buffer = NULL;
int     bsize = GFS_BASIC_BLOCK;
int64   off = 0;


int main(int argc, char *argv[])
{
  char line[80];
  char cmd[5];
  char arg1[20];
  char arg2[20];


  if (argc < 2)
    {
      fprintf(stderr, "Usage: %s <device | file system>\n", argv[0]);
      return(1);
    }

  buffer = (char *)malloc(bsize);
  bzero(buffer, bsize);
  
  if ((fd = open(argv[1], O_RDWR)) < 0)
    {
      fprintf(stderr, "Error opening file: %s\n", argv[1]);
      perror("Reason");
      return(1);
    }

  display(0, SCREEN);
  
  printf("\n> ");

  while (fgets(line, 79, stdin) != NULL)
    {
      sscanf(line, "%s %s %s", cmd, arg1, arg2);
      
      if (!strncmp(cmd, "", 1))
	{
	  /* Do nothing */ 
	}
      else
	{
	  printf("\n");
	  run_command(cmd, arg1, arg2);
	}

	printf("\n> ");
	
	bzero(line, 80);
	bzero(cmd, 5);
	bzero(arg1, 20);
	bzero(arg2, 20);
    }
  printf("\n\n");

  exit(EXIT_SUCCESS);
}


/******************************************************************************
*******************************************************************************
**
** void print_usage()
**
** Description:
**   This routine prints out the appropriate commands for this application.
**
*******************************************************************************
******************************************************************************/

void print_usage(void)
{
  fprintf(stderr, "\n\nSupported commands:\n");
  fprintf(stderr, "l\t\t- More Lines\n");
  fprintf(stderr, "n\t\t- Next\n");
  fprintf(stderr, "p\t\t- Previous\n");
  fprintf(stderr, "q\t\t- Quit\n");
  fprintf(stderr, "r\t\t- Reprint\n");
  fprintf(stderr, "s <bn>\t\t- Search\n");
  fprintf(stderr, "b <bytes>\t- Block size\n");
  fprintf(stderr, "d\t\t- Display GFS structure\n");
  fprintf(stderr, "dx\t\t- Display GFS extended structure\n");
  fprintf(stderr, "e\t\t- edit GFS structure\n");
  fprintf(stderr, "ex\t\t- edit extended GFS structure\n");
}


/******************************************************************************
*******************************************************************************
**
** int run_command()
**
** Description:
**   This routine...
**
** Input(s):
**  *cmd  - 
**  *arg1 - 
**  *arg2 - 
**
** Output(s):
**
**
** Returns:
**   0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int run_command(char *cmd, char *arg1, char *arg2)
{
  static int          start = 0;
  static int          end = SCREEN;
  
  /* Next */

  if (!strncmp(cmd, "n", 1))
    {
      start = 0;
      end = SCREEN;
      off++;
      display(start, end);
    }

  /* Previous */

  else if (!strncmp(cmd, "p", 1))
    {
      start = 0;
      end = SCREEN;
      off--;
      if (off < 0)
	off = 0;
      display(start, end);
    }

  /* Next few lines */

  else if (!strncmp(cmd, "l", 1))
    {
      start = end;
      end = start + SCREEN;
      
      if (end * 16 > bsize)
	{
	  start = 0;
	  end = SCREEN;
	  off++;
	}
      display(start, end);
    }

  /* Search */

  else if (!strncmp(cmd, "s", 1))
    {
      start = 0;
      end = SCREEN;
      off = atoi(arg1);
      display(start, end);
    }

  /* Reprint */ 

  else if (!strncmp(cmd, "r", 1))
    {
      start = 0;
      end = SCREEN;
      display(start, end);
    }

  /*  Change block size */

  else if (!strncmp(cmd, "b", 1))
    {
      bsize = atoi(arg1);
      printf("Block Size = %d bytes\n", bsize);
      
      free(buffer);
      buffer = (char *)malloc(bsize);
      bzero(buffer, bsize);
    }

  /* Quit */ 

  else if (!strncmp(cmd, "q", 1))
    {
      printf("\n\n");
      exit(0);
    }

  /* Display GFS extended structure */

  else if (!strncmp(cmd, "dx", 2))
    {
      int64 s;
      
      if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) < 0)
	{
	  fprintf(stderr, "Error seeking from file\n");
	  perror("Reason");
	  exit(1);
	}
      
      if ((s = read(fd, buffer, bsize)) < 0)
	{
	  fprintf(stderr, "Error reading from file\n");
	  perror("Reason");
	  exit(1);
	}
      
      display_gfs(buffer, 1);
    }

  /* Display GFS structure */

  else if (!strncmp(cmd, "d", 1))
    {
      int64  s;

      if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) < 0)
	{
	  fprintf(stderr, "Error seeking from file\n");
	  perror("Reason");
	  return(1);
	}

      if ((s = read(fd, buffer, bsize)) < 0)
	{
	  fprintf(stderr, "Error reading from file\n");
	  perror("Reason");
	  exit(1);
	}
      
      display_gfs(buffer, 0);
    }
  else if (!strncmp(cmd, "ex", 2))  /* Edit extended GFS structure */
    {
      int64 s;

      if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) < 0)
	{
	  fprintf(stderr, "Error seeking from file\n");
	  perror("Reason");
	  return(1);
	}

      if ((s = read(fd, buffer, bsize)) < 0)
	{
	  fprintf(stderr, "Error reading from file\n");
	  perror("Reason");
	  exit(1);
	}

      edit_gfs(buffer, 1);

      if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) < 0)
	{
	  fprintf(stderr, "Error seeking from file\n");
	  perror("Reason");
	  return(1);
	}

      if (write(fd, buffer, bsize) < 0)
	{
	  fprintf(stderr, "Error writing to file\n");
	  perror("Reason");
	  exit(1);
	}
    }
  else if (!strncmp(cmd, "e", 1))  /* Edit GFS structure */
    {
      int64 s;

      if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) < 0)
	{
	  fprintf(stderr, "Error seeking from file\n");
	  perror("Reason");
	  return(1);
	}

      if ((s = read(fd, buffer, bsize)) < 0)
	{
	  fprintf(stderr, "Error reading from file\n");
	  perror("Reason");
	  exit(1);
	}

      edit_gfs(buffer, 0);      

      if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) < 0)
	{
	  fprintf(stderr, "Error seeking from file\n");
	  perror("Reason");
	  return(1);
	}

      if (write(fd, buffer, bsize) < 0)
	{
	  fprintf(stderr, "Error writing to file\n");
	  perror("Reason");
	  exit(1);
	}
    }
  else
    {
      fprintf(stderr, "%s: Command not found.\n", cmd);
      print_usage();
    }

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** int display()
**
** Description:
**   This routine...
**
** Input(s):
**   start - 
**   end   - 
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

int display(int start, int end)
{
  int	i, j;
  int64 s;
  
  if ((s = LLSEEK(fd, off * bsize, SEEK_SET)) != off * bsize)
    {
      fprintf(stderr, "Error seeking from file\n");
      perror("Reason");
      return(1);
    }

  if (start == 0)
    {
      if ((s = read(fd, buffer, bsize)) != bsize)
	{
	  for (i = s; i < bsize; i++)
	    buffer[i] = 0;
	}
      if (s < 0)
	{
	  fprintf(stderr, "Error reading from file\n");
	  perror("Reason");
	  return(1);
	}
    }

  printf("\n\n%03"FMT_64"d (0x%03"FMT_64"x) (%010"FMT_64"d)\n", off, off, off*bsize + start*16);
  
  for (i = start; i < end; i++)
    {
      for (j = 0; j < 16; j++)
	{
	  printf("%02x ", (unsigned char)buffer[16 * i + j]);
	}
      printf("    ");
      for (j = 0; j < 16; j++)
	{
	  if (isprint(buffer[16 * i + j]))
	    printf("%c", buffer[16 * i + j]);
	  else
	    printf(" ");
	}
      printf("\n");
    }
  printf("\n");

  return(0);
}
