/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: mkfs_gfs.c
**
**    Description:
**	Called by mkfs(1M) or alone
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>
#include <time.h>

/*  Needed for BLKGETSIZE ioctl().
    Replace it with whatever the appropriate thing is for your OS.  */
#if defined(__linux__)
#include <linux/fs.h>
#elif defined(__FreeBSD__)
#include <pool_fbsd.h>
#include <sys/disklabel.h>
#include <ctype.h>
#define BLKGETSIZE DIOCGDINFO
#define atoll(x) strtoq(x,(char **)NULL,10)
#endif

#include "global.h"
#include "pool.h"
#include "GFS.cf"
#include "gfs_ondisk.h"



/*  Architecture Dependent macros  */

#ifdef linux
long long int llseek(int, long long int, int);
#define LLSEEK(x, y, z) llseek((x), (y), (z))
#endif  /*  linux  */

#ifdef IRIX
#define LLSEEK(x, y, z) lseek64((x), (y), (z))
#endif  /*  IRIX  */

#if defined(__FreeBSD__)
#define LLSEEK(x,y,z) lseek((x),(y),(z))
#endif


/*  Constants  */

#define MKFS_ROOT_MODE	        (0755)
#define MKFS_DEFAULT_BSIZE	(4096)
#define MKFS_FRACTION_LOCKS	(0.04)

#define OPTION_STRING ("b:df:hqr:Vzp:t:l:")

#define LINE_SIZE (1024)



/*  Function Macros  */

#define type_zalloc(ptr, type, count) \
{ \
  if (((ptr) = (type *)malloc(sizeof(type) * (count))) != NULL) \
    memset((char *)(ptr), 0, sizeof(type) * (count)); \
  else \
  { \
    fprintf(stderr, "%s: unable to allocate memory on line %d.\n", prog_name, __LINE__); \
    exit(EXIT_FAILURE); \
  } \
}

#define type_alloc(ptr, type, count) \
{ \
  if (((ptr) = (type *)malloc(sizeof(type) * (count))) == NULL) \
  { \
    fprintf(stderr, "%s: unable to allocate memory on line %d.\n", prog_name, __LINE__); \
    exit(EXIT_FAILURE); \
  } \
}

#ifndef BTOBB
#define BTOBB(x) (((x) + GFS_BASIC_BLOCK - 1) / GFS_BASIC_BLOCK)
#endif  /*  BTOBB  */

#define do_lseek(fd, off) \
{ \
  if (LLSEEK((fd), (off), SEEK_SET) != (off)) \
  { \
    fprintf(stderr, "%s: bad seek: %s\n", prog_name, strerror(errno)); \
    exit(EXIT_FAILURE); \
  }\
}

#define do_write(fd, buff, len) \
{ \
  if (write((fd), (buff), (len)) != (len)) \
  { \
    fprintf(stderr, "%s: bad write: %s\n", prog_name, strerror(errno)); \
    exit(EXIT_FAILURE); \
  } \
}



typedef struct commandline_s
{
  int debug;            /*  Print out debugging information?  */
  int quiet;            /*  Be Quiet?  */
  int zero;             /*  Zero the filesystem  */

  uint32 bsize;         /*  The block size of the FS  */
  uint32 rgrps;         /*  The number of resource groups in the FS  */

  char *paramfile;      /*  Parameter file name  */
  char *device;         /*  Device file  */

  char lockproto[GFS_LOCKNAME_LEN + 1];
  char locktable[GFS_LOCKNAME_LEN + 1];

  uint64 locks;
} commandline_t;

typedef struct filesystem_s
{
  gfs_sb_t sb;          /*  The Superblock  */
  gfs_rindex_t *ri;     /*  Array of resource indexes  */
  gfs_rgrp_t *rg;       /*  Array of resource group headers  */
  gfs_dinode_t di;      /*  root dinode structure  */
  gfs_dirent_t de[2];   /*  "." and ".." in the root dinode  */
} filesystem_t;

typedef struct param_s
{
  int subpools;         /*  The number of subpools in this pool  */
  uint32 *rgrps;        /*  The number of resource groups in each subpool  */
} param_t;

typedef struct geometry_s
{
  uint32 subpool;       /*  The subpool who holds this resource group  */

  uint64 rg_offset;     /*  The offset of the beginning of this resource group  */
  uint64 rg_length;     /*  The length of this resource group  */

  uint64 gl_offset;     /*  The offset of the first Glock in this resource group  */
  uint64 gl_length;     /*  The Number of glocks in this resource group  */
} geometry_t;



char *prog_name;





/******************************************************************************
*******************************************************************************
**
** print_usage()
**
** Description:
**   Print usage
**
*******************************************************************************
******************************************************************************/

void print_usage()
{
  printf("Usage: \n\n");
  
  printf("%s [-V|-h]\n", prog_name);
  printf("%s [-dq] -f <ParameterFile>\n", prog_name);
  printf("%s [-dqz] [-b BlockSize] [-r ResourceGroups]\n\t[-p LockProtoName] [-t LockTableName] <CharacterDevice>\n\n",
		 prog_name);

  printf("  -b\tSet filesystem block size\n");
  printf("  -d\tEnable debugging code.\n");
  printf("  -f\tGenerate filesystem from configuration file.\n");
  printf("  -h\tPrint this help, then exit.\n");
  printf("  -q\tQuiet operation; program output is supressed.\n");
  printf("  -r\tSet the number of resource groups.\n");
  printf("  -V\tPrint program version information, then exit.\n");
  printf("  -z\tZero the filesystem before writting it.\n");
  printf("  -p\tThe name of the locking protocol to use.\n");
  printf("  -t\tThe name of the lock table to use.\n");
  printf("  -l\tThe number of locks for the FS (if not on a pool)\n");
}


/******************************************************************************
*******************************************************************************
**
** decode_arguments()
**
** Description:
**   Decode the command line arguments
**
** Input(s):
**   argc - well, it's argc...
**   argv - well, it's argv...
**
** Output(s):
**   comline - the returned arguments
**
*******************************************************************************
******************************************************************************/

void decode_arguments(int argc, char *argv[], commandline_t *comline)
{
  int optchar;


  while((optchar = getopt(argc, argv, OPTION_STRING)) > 0)
    switch (optchar)
	  {
	  case 'b':
		if (comline->paramfile)
		  {
			fprintf(stderr, "%s: -b is incompatible with -f\n", prog_name);
			print_usage();
			exit(EXIT_FAILURE);
		  }
		comline->bsize = (unsigned int)atoi(optarg);
		break;
	  case 'd':
		comline->debug = TRUE;
		break;
	  case 'f':
		if (comline->bsize || comline->rgrps || comline->zero)
		  {
			fprintf(stderr, "%s: -f is incompatible with -b and -r\n", prog_name);
			print_usage();
			exit(EXIT_FAILURE);
		  }
		comline->paramfile = optarg;
		break;
	  case 'h':
		print_usage();
		exit(EXIT_SUCCESS);
		break;
	  case 'q':
		comline->quiet = TRUE;
		break;
	  case 'r':
		if (comline->paramfile)
		  {
			fprintf(stderr, "%s: -r is incompatible with -f\n", prog_name);
			print_usage();
			exit(EXIT_FAILURE);
		  }
		comline->rgrps = (unsigned int)atoi(optarg);
		break;
	  case 'V':
		printf("mkfs_gfs -- %s Release\n", GFS_RELEASE_NAME);
		printf("  Copyright 1995-1999 Regents of the University of Minnesota\n");
		printf("  Compiled for GFS Version %d.%d.%d at %s %s\n",
			   GFSMajorVersion, GFSMinorVersion, GFSUpdateLevel,
			   __DATE__, __TIME__);
		exit(EXIT_SUCCESS);
		break;
	  case 'z':
		if (comline->paramfile)
		  {
			fprintf(stderr, "%s: -z is incompatible with -f\n", prog_name);
			print_usage();
			exit(EXIT_FAILURE);
		  }
		comline->zero = TRUE;
		break;
	  case 'p':
		if (strlen(optarg) > GFS_LOCKNAME_LEN)
		  {
			fprintf(stderr, "%s: lock protocol name %s is too long\n", prog_name, optarg);
			exit(EXIT_FAILURE);
		  }
		strcpy(comline->lockproto, optarg);
		break;
	  case 't':
		if (strlen(optarg) > GFS_LOCKNAME_LEN)
		  {
			fprintf(stderr, "%s: lock table name %s is too long\n", prog_name, optarg);
			exit(EXIT_FAILURE);
		  }
		strcpy(comline->locktable, optarg);
		break;
	  case 'l':
		comline->locks = atoll(optarg);
		break;
	  case '?':
		print_usage();
		exit(EXIT_FAILURE);
		break;
	  default:
		fprintf(stderr, "%s: unknown option -- %c\n", prog_name, optchar);
		print_usage();
		exit(EXIT_FAILURE);
		break;
	  };


  if (!comline->paramfile)
	{
	  if (!argv[optind])
		{
		  fprintf(stderr, "%s: missing argument\n", prog_name);
		  print_usage();
		  exit(EXIT_FAILURE);
		}

	  comline->device = argv[optind];
	  optind++;
	}


  if (optind < argc) 
	{
	  fprintf(stderr, "%s: Unrecognized option(s): ", prog_name);
	  for( ; optind < argc; optind++)
		fprintf(stderr, "%s ", argv[optind]);
	  fprintf(stderr, "\n");
	  print_usage();
	  exit(EXIT_FAILURE);
	}


  if (comline->debug)
  {
    printf("Command Line Arguments:\n");
    printf("  bsize = %u\n", comline->bsize);
    printf("  debug = %d\n", comline->debug);
    printf("  paramfile = %s\n", comline->paramfile);
    printf("  quiet = %d\n", comline->quiet);
    printf("  rgrps = %u\n", comline->rgrps);
    printf("  zero = %d\n", comline->zero);
    printf("  device = %s\n", comline->device);
    printf("  proto = %s\n", comline->lockproto);
    printf("  table = %s\n", comline->locktable);
    printf("  locks = %"FMT_64"u\n", comline->locks);
  }
}


/******************************************************************************
*******************************************************************************
**
** read_paramfile()
**
** Description:
**   Read in the parameter file
**
** Input(s):
**   comline - command line arguments (Specifies the parameter filename)
**
** Output(s):
**   params - Parameters read in from the file
**
*******************************************************************************
******************************************************************************/

void read_paramfile(commandline_t *comline, param_t *params)
{
  FILE *fp;

  char line[LINE_SIZE], tag[LINE_SIZE], arg1[LINE_SIZE], arg2[LINE_SIZE], arg3[LINE_SIZE];
  int line_count = 0;
  int rgrp;


  if ((fp = fopen(comline->paramfile, "r")) == NULL)
	{
	  fprintf(stderr, "%s: Error opening the parameter file \"%s\": %s.\n", 
			  prog_name, comline->paramfile, strerror(errno));
	  exit(EXIT_FAILURE);
	}


  while (fgets(line, LINE_SIZE - 1, fp) != NULL)
	{
	  line_count++;

	  if (line[0] == '#')	     /*  Skip comments  */
		continue;

	  if (line[0] == '\n')     /*  Skip blank lines  */
		continue;

	  /*  Scan for a keywork and up to three arguments  */

	  tag[0] = arg1[0] = arg2[0] = arg3[0] = 0;

	  sscanf(line, "%s %s %s %s", tag, arg1, arg2, arg3);
    
	  if (!strcmp(tag, "rawdevice"))
		{
		  if (comline->device)
			{
			  fprintf(stderr, "%s: multiple \"rawdevice\" tags in parameter file (line %d)\n",
					  prog_name, line_count);
			  exit(EXIT_FAILURE);
			}
		  type_alloc(comline->device, char, strlen(arg1) + 1);
		  strcpy(comline->device, arg1);
		}
	  else if (!strcmp(tag, "blocksize"))
		comline->bsize = (unsigned int)atoi(arg1);
	  else if (!strcmp(tag, "zerofs"))
		comline->zero = atoi(arg1);
	  else if (!strcmp(tag, "subpools"))
		{
		  if (params->subpools)
			{
			  fprintf(stderr, "%s: multiple \"subpools\" tags in parameter file (line %d)\n",
					  prog_name, line_count);
			  exit(EXIT_FAILURE);
			}
		  params->subpools = (unsigned int)atoi(arg1);
		  type_zalloc(params->rgrps, uint32, params->subpools);
		}
	  else if (!strcmp(tag, "subpool"))
		{
		  if (!params->subpools)
			{
			  fprintf(stderr, "%s: \"subpool\" tag without \"subpools\" tag (line %d)\n",
					  prog_name, line_count);
			  exit(EXIT_FAILURE);
			}
		  rgrp = (unsigned int)atoi(arg1);
		  if (rgrp >= params->subpools)
			{
			  fprintf(stderr, "%s: subpool out of range (line %d)\n", prog_name, line_count);
			  exit(EXIT_FAILURE);
			}
		  params->rgrps[rgrp] = (unsigned int)atoi(arg2);
		  comline->rgrps += params->rgrps[rgrp];
		}
	  else if (!strcmp(tag, "lockproto"))
		{
		  if (strlen(arg1) > GFS_LOCKNAME_LEN)
			{
			  fprintf(stderr, "%s: lock protocol name %s is too long\n", prog_name, arg1);
			  exit(EXIT_FAILURE);
			}
		  strcpy(comline->lockproto, arg1);
		}
	  else if (!strcmp(tag, "locktable"))
		{
		  if (strlen(arg1) > GFS_LOCKNAME_LEN)
			{
			  fprintf(stderr, "%s: lock table name %s is too long\n", prog_name, arg1);
			  exit(EXIT_FAILURE);
			}
		  strcpy(comline->locktable, arg1);
		}
	  else if (!strcmp(tag, "locks"))
		{
		  comline->locks = atoll(arg1);
		}
	  else
		{
		  fprintf(stderr, "%s: invalid tag (line %d)\n", prog_name, line_count);
		  exit(EXIT_FAILURE);
		}
	}

  
  fclose(fp);


  if (comline->debug)
  {
    printf("From Parameter File:\n");
    printf("  device = %s\n", comline->device);
    printf("  bsize = %u\n", comline->bsize);
    printf("  rgrps = %u\n", comline->rgrps);
    printf("  proto = %s\n", comline->lockproto);
    printf("  table = %s\n", comline->locktable);
    printf("  locks = %"FMT_64"u\n", comline->locks);
  }
}


/******************************************************************************
*******************************************************************************
**
** get_pool_geometry()
**
** Description:
**   Talk to the pool and find out the number of subpools, their size, and
**   the number of Glocks.
**
** Input(s):
**   comline - the command line arguments (Specifies the pool name)
**
** Output(s):
**   splist - the structures containing the subpool geometry
**
*******************************************************************************
******************************************************************************/

int get_pool_geometry(commandline_t *comline, pool_splist_t *splist)
{
  int fd;
  int i;
  uint32 bbsize = BTOBB(comline->bsize);
  uint64 offset, len;


  /*  Open pool device  */

  if ((fd = open(comline->device, O_RDONLY)) < 0)
	{
	  fprintf(stderr, "%s: Error opening the device special file \"%s\": %s.\n", 
			  prog_name, comline->device, strerror(errno));
	  exit(EXIT_FAILURE);
	}

  /*  Get subpool information  */

  for (i = 128; i < 65536; i += 128) 
	{
	  splist->spl_count = i;
	  type_zalloc(splist->spl_list, pool_spstatus_t, i);

	  if (ioctl(fd, POOL_SPLIST, splist) != 0) 
		return(-1);
    
	  if (splist->spl_count < i)
		break;
    
	  free(splist->spl_list);
	}
  
  if (i >= 65536)
	{
	  fprintf(stderr, "%s: unable to get pool subpool list: to many subpools.\n",
			  prog_name);
	  exit(EXIT_FAILURE);
	}

  assert(splist->spl_count > 0);
  

  close(fd);


  /*  Make sure all the subpools are aligned  */

  for (i = 0; i < splist->spl_count; i++)
  {
    offset = splist->spl_list[i].sps_boffset;
    len = splist->spl_list[i].sps_bcount;

    if (len < 2 * bbsize)
    {
      fprintf(stderr, "%s: pool to small %"FMT_64"d bytes\n",
	      prog_name, len * GFS_BASIC_BLOCK);
      exit(EXIT_FAILURE);
    }

    if (offset % bbsize)
    {
      len -= bbsize - (offset % bbsize);
      offset += bbsize - (offset % bbsize);
    }

    len -= len % bbsize;

    splist->spl_list[i].sps_boffset = offset;
    splist->spl_list[i].sps_bcount = len;
  }


  if (comline->debug)
  {
    printf("SubPool Geometry:\n");
    for (i = 0; i < splist->spl_count; i++)
      printf("  Subpool #%d:  BStart = %"FMT_64"u, BLength = %"FMT_64"u, DStart = %"FMT_64"u, DLength = %"FMT_32"u\n",
	     i, splist->spl_list[i].sps_boffset, splist->spl_list[i].sps_bcount,
	     splist->spl_list[i].sps_dlstart, splist->spl_list[i].sps_dlocks);
  }


  return(0);
}


/******************************************************************************
*******************************************************************************
**
** fake_pool_geometry()
**
** Description:
**   Talk to the underlying device and figure out how many blocks it has.
**   Stuff the value into a pool_splist_t structure with only on subpool.
**
** Input(s):
**   comline - the command line arguments (Specifies the pool name)
**
** Output(s):
**   splist - the structures containing the subpool geometry
**
*******************************************************************************
******************************************************************************/

void fake_pool_geometry(commandline_t *comline, pool_splist_t *splist)
{
  int fd, error, i;
  long size;

#if defined(__FreeBSD__)
  static struct disklabel lab;	
  register struct partition *pp;
  char *cp;  
  char vflag = 0; /* need to add this later */
#endif

  splist->spl_count = 1;

  type_zalloc(splist->spl_list, pool_spstatus_t, 1);

  
  /*  Open pool device  */

  if ((fd = open(comline->device, O_RDONLY)) < 0)
	{
	  fprintf(stderr, "%s: Error opening the device special file \"%s\": %s.\n", 
			  prog_name, comline->device, strerror(errno));
	  exit(EXIT_FAILURE);
	}
#if defined(__linux__)
  error = ioctl(fd, BLKGETSIZE, &size);
#elif defined(__FreeBSD__)
  error = ioctl(fd, BLKGETSIZE, (char *)&lab);
#endif
  if (error < 0)
	{
	  fprintf(stderr, "%s: can't determine device size\n",
			  prog_name);
	  exit(EXIT_FAILURE);
	}

  close(fd);
#if defined (__FreeBSD__)  
  cp = strchr(comline->device, '\0');
  if (cp == comline->device){
	fprintf(stderr,"null special file name");
	exit(-1);
  }
  cp--;
  if (!vflag && (*cp < 'a' || *cp > 'h') && !isdigit(*cp)){
	fprintf(stderr,"%s: can't figure out file system partition",
		  comline->device);
	exit(-1);
  }
  if (vflag)
	pp = &lab.d_partitions[0];
  else
	pp = &lab.d_partitions[*cp - 'a'];
  if (pp->p_size == 0)
	fprintf(stderr,"%s: `%c' partition is unavailable", comline->device, *cp);
  size = pp->p_size;
  fprintf(stderr,"partition %c size %d\n",*cp,size);
#endif

  size = size - (size % BTOBB(comline->bsize));
  splist->spl_list->sps_bcount = size;


  if (comline->debug)
  {
    printf("SubPool Geometry:\n");
    for (i = 0; i < splist->spl_count; i++)
      printf("  Subpool #%d:  BStart = %"FMT_64"u, BLength = %"FMT_64"u, DStart = %"FMT_64"u, DLength = %"FMT_32"u\n",
	     i, splist->spl_list[i].sps_boffset, splist->spl_list[i].sps_bcount,
	     splist->spl_list[i].sps_dlstart, splist->spl_list[i].sps_dlocks);
  }
}


/******************************************************************************
*******************************************************************************
**
** compute_params()
**
** Description:
**   This is called if there wasn't a parameter file specified.
**   Compute the values ourself
**
** Input(s):
**   comline - the command line arguments (Block size, # of Resource Groups)
**   splist - the subpool geometry
**
** Output(s):
**   params - The generated parameters
**
*******************************************************************************
******************************************************************************/

void compute_params(commandline_t *comline, param_t *params, pool_splist_t *splist)
{
  int i;
  uint32 frgrps;


  params->subpools = splist->spl_count;
  type_alloc(params->rgrps, uint32, params->subpools);

  frgrps = comline->rgrps / params->subpools;

  /*  Distribute the resource groups and glocks among the last n-1 subpools  */

  for (i = 1; i < params->subpools; i++)
    params->rgrps[i] = frgrps;

  /*  Give the first subpool all remaining resource groups and glocks.  */

  params->rgrps[0] = comline->rgrps - frgrps * (params->subpools - 1);
}


/******************************************************************************
*******************************************************************************
**
** verify_params()
**
** Description:
**   This is called when a parameter file is specified.  It verifies the
**   parameters are sane.
**
** Input(s):
**   comline - the command line arguments
**   params - the parameters from the file
**   splist - the subpool geometry from the file
**
*******************************************************************************
******************************************************************************/

void verify_params(commandline_t *comline, param_t *params, pool_splist_t *splist)
{
  unsigned int i;
  int rgrps = 0;

  if (params->subpools != splist->spl_count)
  {
    fprintf(stderr, "%s: bad number of subpools (%u != %u)\n",
	    prog_name, params->subpools, splist->spl_count);
    exit(EXIT_FAILURE);
  }


  /*  If the user specified the -l option, override the number of locks that
      are pool returns and distribute the -l locks evenly across the subpools.  */

  if (comline->locks)
  {
    if (comline->debug)
      printf("Setting number of locks to %"FMT_64"u\n", comline->locks);
      
    splist->spl_list[0].sps_dlocks = comline->locks -
      (comline->locks / params->subpools) * (params->subpools - 1);
    for (i = 1; i < params->subpools; i++)
      splist->spl_list[i].sps_dlocks = comline->locks / params->subpools;
  }


  for (i = 0; i < params->subpools; i++)
  {
    if (params->rgrps[i] < 1)
    {
      fprintf(stderr, "%s: invalid number of resource groups (%u) in subpool #%d \n",
	      prog_name, params->rgrps[i], i);
      exit(EXIT_FAILURE);
    }

    rgrps += params->rgrps[i];

    if (splist->spl_list[i].sps_bcount == 0)
    {
      fprintf(stderr, "%s: invalid number of blocks (%"FMT_64"u) in subpool #%d\n",
	      prog_name, splist->spl_list[i].sps_bcount, i);
      exit(EXIT_FAILURE);
    }

    if (splist->spl_list[i].sps_dlocks == 0)
    {
      fprintf(stderr, "%s: invalid number of glocks (%u) in subpool #%d\n",
	      prog_name, splist->spl_list[i].sps_dlocks, i);
      exit(EXIT_FAILURE);
    }
  }

  if (rgrps != comline->rgrps)
	{
	  fprintf(stderr, "%s:  invalid number of resource groups\n", prog_name);
	  exit(EXIT_FAILURE);
	}
}


/******************************************************************************
*******************************************************************************
**
** compute_geometry()
**
** Description:
**   This routine takes the parameters (from the file/that were computed),
**   the command line arguments, and the subpool geometry and computes
**   the layout of the resource groups.
**
**   One geometry_t structure is allocate per resource group
**
** Input(s):
**   comline - the command line arguments
**   params - the parameters about resource group placement
**   splist - the subpool geometry
**
** Returns:
**   An array of geometry_t structures detailing the layout of the resource
**   groups
**
*******************************************************************************
******************************************************************************/

geometry_t *compute_geometry(commandline_t *comline, param_t *params,
							 pool_splist_t *splist)
{
  geometry_t *geo;
  int rgrp, sp, rg;
  uint64 fbbs;
  uint32 fglocks;
  uint32 bbsize = BTOBB(comline->bsize);



  type_zalloc(geo, geometry_t, comline->rgrps);


  /*  Distribute Glocks and blocks among the resource groups.  */

  for (rgrp = 0, sp = 0; sp < params->subpools; sp++)
	{
	  fbbs = splist->spl_list[sp].sps_bcount / params->rgrps[sp];
	  fglocks = splist->spl_list[sp].sps_dlocks / params->rgrps[sp];
	  if (fglocks < 2)
		{
		  fprintf(stderr, "%s: too many resource groups for this number of glocks\n",
				  prog_name);
		  exit(EXIT_FAILURE);
		}

	  for (rg = 1; rg < params->rgrps[sp]; rg++)
		{
		  geo[rgrp + rg].subpool = sp;
		  geo[rgrp + rg].rg_length = fbbs / bbsize;
		  geo[rgrp + rg].gl_length = fglocks;
		}

	  geo[rgrp].subpool = sp;
	  geo[rgrp].rg_length = (splist->spl_list[sp].sps_bcount - fbbs * (params->rgrps[sp] - 1)) /
		bbsize;
	  geo[rgrp].gl_length = splist->spl_list[sp].sps_dlocks - fglocks * (params->rgrps[sp] - 1);

	  rgrp += params->rgrps[sp];
	}


  /*  Compute offsets  */
  
  for (rgrp = 0, sp = 0; sp < params->subpools; sp++) {
	for (rg = 0; rg < params->rgrps[sp]; rg++) {
	  if (rg == 0) {
		if (sp == 0) {
		  assert(splist->spl_list[0].sps_boffset == 0);
		  geo[0].rg_offset = ((sizeof(gfs_sb_t) + sizeof(gfs_rindex_t) * comline->rgrps) + comline->bsize - 1) /
			comline->bsize;
		  geo[0].rg_length -= geo[0].rg_offset;
		  geo[0].gl_offset = GFS_RGRP0_LOCK;
		  geo[0].gl_length -= geo[0].gl_offset;
		} else {
		  geo[rgrp].rg_offset = splist->spl_list[sp].sps_boffset / bbsize;
		  geo[rgrp].gl_offset = splist->spl_list[sp].sps_dlstart;
		}
	  } else {
		geo[rgrp + rg].rg_offset = geo[rgrp + rg - 1].rg_offset + geo[rgrp + rg - 1].rg_length;
		geo[rgrp + rg].gl_offset = geo[rgrp + rg - 1].gl_offset + geo[rgrp + rg - 1].gl_length;
	  }
	}
	
	rgrp += params->rgrps[sp];
  }
  
  
  if (comline->debug)
  {
    int i;

    printf("Resource Group Geometry:\n");
    for (i = 0; i < comline->rgrps; i++)
      printf("  rgrp #%d:  spool = %u, boffset = %"FMT_64"u, blength = %"FMT_64"u, doffset = %"FMT_64"u, dlength = %"FMT_64"u\n",
	     i, geo[i].subpool, geo[i].rg_offset, geo[i].rg_length,
	     geo[i].gl_offset, geo[i].gl_length);
  }


  return(geo);
}


/******************************************************************************
*******************************************************************************
**
** fill_rgrps()
**
** Description:
**   Take the resource group geometry array and fill in the resource index
**   and resource group header arrays.
**
** Input(s):
**   comline - the command line arguments
**   geo - the resource group geometry
**
** Output(s):
**   fs - the filesystem structures
**
*******************************************************************************
******************************************************************************/

void fill_rgrps(commandline_t *comline, filesystem_t *fs, geometry_t *geo)
{
  int i;


  type_zalloc(fs->ri, gfs_rindex_t, comline->rgrps);
  type_zalloc(fs->rg, gfs_rgrp_t, comline->rgrps);


  for (i = 0; i < comline->rgrps; i++)
	{
	  /*  Fill in the Resource Index  */

	  fs->ri[i].ri_addr = geo[i].rg_offset;
	  fs->ri[i].ri_length = (sizeof(gfs_rgrp_t) + (geo[i].rg_length / GFS_NBBY) + comline->bsize - 1) /
		comline->bsize;
	  fs->ri[i].ri_subpool = geo[i].subpool;
	  fs->ri[i].ri_glock = geo[i].gl_offset;

	  fs->ri[i].ri_data1 = fs->ri[i].ri_addr + fs->ri[i].ri_length;
	  fs->ri[i].ri_data = geo[i].rg_length - fs->ri[i].ri_length;

	  fs->ri[i].ri_glock1 = fs->ri[i].ri_glock + 1;
	  fs->ri[i].ri_glocks = geo[i].gl_length - 1;

	  fs->ri[i].ri_bitbytes = fs->ri[i].ri_data / GFS_NBBY;

	  /*  Fill in the Resource Group header  */

	  fs->rg[i].rg_magic = GFS_MAGIC_RG;
	  fs->rg[i].rg_time = time(NULL);

	  if (i)
		{
		  fs->rg[i].rg_free = fs->ri[i].ri_data;
		  fs->rg[i].rg_dinodes = 0;
		}
	  else  /*  The first RGRP contains the root dinode  */
		{
		  fs->rg[i].rg_free = fs->ri[i].ri_data - 1;
		  fs->rg[i].rg_dinodes = 1;
		}
	}
}


/******************************************************************************
*******************************************************************************
**
** fill_super()
**
** Description:
**   Fill in the superblock structure
**
** Input(s):
**   comline - the command line arguments
**   fs - the filesystem structures
**
*******************************************************************************
******************************************************************************/

void fill_super(commandline_t *comline, filesystem_t *fs)
{
  int i;


  fs->sb.sb_magic = GFS_MAGIC_SB;
  fs->sb.sb_version = GFS_VERSION;

  fs->sb.sb_bsize = comline->bsize;
  fs->sb.sb_length = (sizeof(gfs_sb_t) + sizeof(gfs_rindex_t) * comline->rgrps + comline->bsize - 1) /
    comline->bsize;
  fs->sb.sb_rootdi = fs->ri[0].ri_data1;
  fs->sb.sb_rgrps = comline->rgrps;

  strcpy(fs->sb.sb_name, "GFS");

  fs->sb.sb_fssize = fs->ri[0].ri_addr;
  fs->sb.sb_glocks = GFS_RGRP0_LOCK;
  for (i = 0; i < comline->rgrps; i++)
	{
	  fs->sb.sb_fssize += fs->ri[i].ri_length + fs->ri[i].ri_data;
	  fs->sb.sb_glocks += 1 + fs->ri[i].ri_glocks;
	}

  strcpy(fs->sb.sb_lockproto, comline->lockproto);
  strcpy(fs->sb.sb_locktable, comline->locktable);

  if ((sizeof(gfs_sb_t) % 8) + 
      (sizeof(gfs_rindex_t) % 8) +
      (sizeof(gfs_rgrp_t) % 8) +
      (sizeof(gfs_dinode_t) % 8) +
      (sizeof(gfs_dirent_t) % 8) +
      (sizeof(gfs_leaf_t) % 8))
	{
	  fprintf(stderr, "%s: one of the structures in gfs_ondisk.h is not a multiple of 8\n",
			  prog_name);
	  exit(EXIT_FAILURE);
	}

  fs->sb.sb_sbsize = sizeof(gfs_sb_t);
  fs->sb.sb_risize = sizeof(gfs_rindex_t);
  fs->sb.sb_rgsize = sizeof(gfs_rgrp_t);
  fs->sb.sb_disize = sizeof(gfs_dinode_t);
  fs->sb.sb_desize = sizeof(gfs_dirent_t);
  fs->sb.sb_lfsize = sizeof(gfs_leaf_t);

  fs->sb.sb_mounts = 0;
  /*  sb_mountids are already zeroed  */
  fs->sb.sb_time = time(NULL);
}


/******************************************************************************
*******************************************************************************
**
** fill_root()
**
** Description:
**   Fill in the root dinode and its "." and ".." entries
**
** Input(s):
**   comline - The command line arguments
**   fs - the filesystem structures
**
*******************************************************************************
******************************************************************************/

void fill_root(commandline_t *comline, filesystem_t *fs)
{
  /*  Fill in the root dinode  */

  fs->di.di_magic = GFS_MAGIC_DI;

  fs->di.di_addr = fs->sb.sb_rootdi;

  fs->di.di_mode = MKFS_ROOT_MODE;
  fs->di.di_uid = 0;
  fs->di.di_gid = 0;
  fs->di.di_nlink = 2;
  fs->di.di_bsize = comline->bsize - sizeof(gfs_dinode_t);
  fs->di.di_blocks = 1;
  fs->di.di_atime = fs->di.di_mtime = fs->di.di_ctime = time(NULL);
  fs->di.di_major = 0;
  fs->di.di_minor = 0;

  fs->di.di_rgrp = 0;
  fs->di.di_glock = GFS_ROOT_LOCK;
  fs->di.di_type = FILE_DIR;
  fs->di.di_height = 0;
  fs->di.di_stamp = 0x4B50;

  fs->di.di_nopen = 0;
  memset(fs->di.di_open, 0, GFS_MOUNT_BITS);

  fs->di.di_regime = GFS_DIR_LINEAR;
  fs->di.di_depth = 0;
  fs->di.di_entries = 2;

  /*  Fill in . and .. in the root dinode  */

  fs->de[0].de_ino = fs->sb.sb_rootdi;
  fs->de[0].de_hash = 0x0ED4E242;
  fs->de[0].de_name_len = 1;
  fs->de[0].de_rec_len = GFS_DIR_REC_LEN(2);
  memcpy(fs->de[0].de_name, ".", 1);

  fs->de[1].de_ino = fs->sb.sb_rootdi;
  fs->de[1].de_hash = 0x9608161C;
  fs->de[1].de_name_len = 2;
  fs->de[1].de_rec_len = (comline->bsize - fs->de[0].de_rec_len - sizeof(gfs_dinode_t));  
  memcpy(fs->de[1].de_name, "..", 2);
}


/******************************************************************************
*******************************************************************************
**
** zero_fs()
**
** Description:
**   Zero out all the blocks in the filesystem
**
** Input(s):
**   comline - the command line arguments
**   splist - the subpool geometry 
**
*******************************************************************************
******************************************************************************/

void zero_fs(commandline_t *comline, pool_splist_t *splist)
{
  int fd;
  int sp;
  uint64 x;
  char *block;
  int bbsize = comline->bsize / GFS_BASIC_BLOCK;


  if ((fd = open(comline->device, O_WRONLY)) < 0)
	{
	  fprintf(stderr, "%s: can't open device %s: %s\n", prog_name,
			  comline->device, strerror(errno));
	  exit(EXIT_FAILURE);
	}


  type_zalloc(block, char, comline->bsize);


  printf("Zeroing FS...");

  for (sp = 0; sp < splist->spl_count; sp++)
	{
	  do_lseek(fd, splist->spl_list[sp].sps_boffset * GFS_BASIC_BLOCK);

	  for (x = 0; x < splist->spl_list[sp].sps_bcount; x += bbsize)
		{
		  do_write(fd, block, comline->bsize);
		  if (x % 1000 == 0)
			{
			  printf(".");
			  fflush(stdout);
			}      
		}
	}

  printf("\n");


  free(block);


  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** write_fs()
**
** Description:
**   Write the filesystem structures to disk
**
** Input(s):
**   comline - The command line arguments (The device filename)
**   fs - the filesystem structures
**
*******************************************************************************
******************************************************************************/

void write_fs(commandline_t *comline, filesystem_t *fs)
{
  int fd;
  int x;
  char *block;


  if ((fd = open(comline->device, O_WRONLY)) < 0)
	{
	  fprintf(stderr, "%s: can't open device %s: %s\n", prog_name,
			  comline->device, strerror(errno));
	  exit(EXIT_FAILURE);
	}


  /*  Write SuperBlock / ResourceIndex  */

  type_zalloc(block, char, fs->sb.sb_length * fs->sb.sb_bsize);

  gfs_sb_out(&fs->sb, block);
  for (x = 0; x < comline->rgrps; x++)
    gfs_rindex_out(&fs->ri[x], block + sizeof(gfs_sb_t) + sizeof(gfs_rindex_t) * x);

  do_lseek(fd, GFS_SB_ADDR * GFS_BASIC_BLOCK);
  do_write(fd, block, fs->sb.sb_length * fs->sb.sb_bsize);

  free(block);


  /*  Write ResourceGroupHeaders  */

  for (x = 0; x < comline->rgrps; x++)
	{
	  type_zalloc(block, char, fs->ri[x].ri_length * comline->bsize);

	  gfs_rgrp_out(&fs->rg[x], block);

	  /*  Allocate space for the root dinode  */

	  if (x == 0)
		block[sizeof(gfs_rgrp_t)] = 0x03;

	  do_lseek(fd, fs->ri[x].ri_addr * comline->bsize);
	  do_write(fd, block, fs->ri[x].ri_length * comline->bsize);

	  free(block);
	}


  /*  Write the RootDinode  */

  type_zalloc(block, char, comline->bsize);
  gfs_dinode_out(&fs->di, block);

  gfs_dirent_out(&fs->de[0], block + sizeof(gfs_dinode_t));

  gfs_dirent_out(&fs->de[1], block + sizeof(gfs_dinode_t) + fs->de[0].de_rec_len);

  do_lseek(fd, fs->sb.sb_rootdi * comline->bsize);
  do_write(fd, block, comline->bsize);


  if (close(fd) != 0)
	{
	  fprintf(stderr, "%s: error on close: %s\n", prog_name, strerror(errno));
	  exit(EXIT_FAILURE);
	}
}


/******************************************************************************
*******************************************************************************
**
** print_results()
**
** Description:
**   print results to the screen
**   Maybe this should look prettier
**
** Input(s):
**   comline - the command line arguments
**   fs - the filesystem structures
**
*******************************************************************************
******************************************************************************/

void print_results(commandline_t *comline, filesystem_t *fs)
{
  printf("Device:  %s\n", comline->device);
  printf("Blocksize:  %u\n", fs->sb.sb_bsize);
  printf("Blocks:  %"FMT_64"u\n", fs->sb.sb_fssize);
  printf("Resource Groups:  %u\n", fs->sb.sb_rgrps);
  printf("Glocks:  %"FMT_64"u\n", fs->sb.sb_glocks);
  printf("Locking Protocol:  %s\n", comline->lockproto);
  printf("Lock Table:  %s\n", comline->locktable);
  printf("\nAll Done\n");
}


/******************************************************************************
*******************************************************************************
**
** main()
**
** Description:
**   Do everything
**
*******************************************************************************
******************************************************************************/

int main(int argc, char *argv[])
{
  commandline_t comline;
  param_t params;
  pool_splist_t splist;
  geometry_t *geo;
  filesystem_t fs;
  struct stat st;
  unsigned int x;


  prog_name = argv[0];


  /*  Process the command line arguments  */

  memset(&comline, 0, sizeof(commandline_t));

  decode_arguments(argc, argv, &comline);

  /*  Read in the parameter file  */

  memset(&params, 0, sizeof(param_t));

  if (comline.paramfile)
    read_paramfile(&comline, &params);

  /*  Make sure it's a character special  */

  if (stat(comline.device, &st) < 0)
	{
	  fprintf(stderr, "%s: could not stat %s: %s\n", prog_name, 
			  comline.device, strerror(errno));
	  exit(EXIT_FAILURE);
	}

#ifdef IRIX
  if (!S_ISCHR(st.st_mode))
	{
	  fprintf(stderr, "%s: filesystems can only be built on character device nodes\n",
			  prog_name);
	  fprintf(stderr, "%s: (not %s)\n", prog_name, comline.device);
	  exit(EXIT_FAILURE);
	}
#endif  /*  IRIX  */
#ifdef linux
  if (!S_ISBLK(st.st_mode))
	{
	  fprintf(stderr, "%s: filesystems can only be built on block device nodes\n",
			  prog_name);
	  fprintf(stderr, "%s: (not %s)\n", prog_name, comline.device);
	  exit(EXIT_FAILURE);
	}
#endif /*  linux  */
  
  /*  Check the block size  */

  if (!comline.bsize)
	{
	  comline.bsize = MKFS_DEFAULT_BSIZE;
	  if (comline.debug)
		printf("Blocksize defaulting to %u\n", comline.bsize);
	}

  /*  Block sizes must be at least big enough to hold one dinode and two dentries  */

  for (x = 1024; x; x <<= 1)
    if (x == comline.bsize)
      break;

  if (!x)
	{
	  fprintf(stderr, "%s: invalid block size %u (must be at least 1024)\n",
			  prog_name, comline.bsize);
	  exit(EXIT_FAILURE);
	}

  /*  Get the pool geometry  */

  memset(&splist, 0, sizeof(pool_splist_t));

  if (get_pool_geometry(&comline, &splist) < 0)
    fake_pool_geometry(&comline, &splist);

  if (!comline.rgrps)
	{
	  uint32 x;
	  uint64 glocks = 0;

	  for (x = 0; x < splist.spl_count; x++)
		glocks += splist.spl_list[x].sps_dlocks;

	  comline.rgrps = MKFS_FRACTION_LOCKS * glocks;
	  if (comline.debug)
		printf("Number of Resource Groups defaulting to %u\n", comline.rgrps);
	}

  /*  Set up parameters  */

  if (!comline.paramfile)
    compute_params(&comline, &params, &splist);

  /*  Verify parameters  */

  verify_params(&comline, &params, &splist);

  if (comline.debug)
	{
	  int i;

	  printf("Params:  %d\n", params.subpools);
	  for (i = 0; i < params.subpools; i++)
		printf("  Subpool #%d:  rgrps = %u\n",
			   i, params.rgrps[i]);
	}

  /*  Compute the Resource Group geometry  */

  geo = compute_geometry(&comline, &params, &splist);

  memset(&fs, 0, sizeof(filesystem_t));
  
  /*  Fill in Resource Groups  */

  fill_rgrps(&comline, &fs, geo);

  /*  Fill in Super Block  */

  fill_super(&comline, &fs);

  /*  Fill in the root dinode and dirents  */

  fill_root(&comline, &fs);

  /*  Zero the FS if necessary  */

  if (comline.zero)
    zero_fs(&comline, &splist);

  /*  Write out the data  */

  write_fs(&comline, &fs);

  /*  Print out FileSystem Stats  */

  if (!comline.quiet)
    print_results(&comline, &fs);

  exit(EXIT_SUCCESS);
}
