/******************************************************************************
*******************************************************************************
**
**    Copyright 1998-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: pinfo.c
**
**    Description:
**	Obtain pool information from kernel
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <paths.h>
#include <ctype.h>
#include <assert.h>

#include "global.h"
#include "pool.h"


#define POOL_DIR_PATH	"/dev/pool/"
#define SEARCH_DIR	"/dev/"


static void printUsage(char *progName, int status);
static int getname(unsigned int dev, char *devname);

int	debug, listsub;


int main(int argc, char *argv[])
{
  char c;
  pool_list_t		plist;
  pool_status_t		*pstat;
  pool_splist_t		splist;
  pool_spstatus_t	*spstat;
  int			statsize = 32;	  /* FIXME: start at a lower num */   
  int			spstatsize = 16;  /* and grow if # pools is larger */
  int			commfd, fd;
  int			i,j,k;
  char			*devName;
  int			major, minor;
  char			*name = NULL;


  while ((c = getopt(argc, argv, "hds")) != -1) {

    switch (c) {
    case 'h':
      printUsage(argv[0], EXIT_SUCCESS);
      exit(0);
      break;
    case 'd':
      debug = 1;
      fprintf(stderr, "Debugging mode on\n");
      break;
    case 's':
      listsub = 1;
      break;
    default:
      printUsage(argv[0], EXIT_FAILURE);
      exit(1);
    }
  }



  if ((commfd = open("/dev/poolcomm", O_RDONLY)) < 0)  {
    fprintf(stderr, "Couldn't open comm device: %s.\n", strerror(errno));
    exit(1);
  }


  pstat = (pool_status_t *)malloc(statsize * sizeof(pool_status_t));
  memset(pstat, 0, statsize * sizeof(pool_status_t));
  plist.pl_count = statsize;
  plist.pl_list = pstat;

  if (ioctl(commfd, POOL_LIST, &plist) < 0){
    fprintf(stderr, "pool communications error: %s\n", strerror(errno));
    exit(-1);
  }


  printf("\n--------------------------------------------------------------------\n");
  printf("Pool List  (%d pools)\n", plist.pl_count);
  
  for(i=0; i< plist.pl_count; i++){
    printf("  (%d) Pool name:    %s\n", i+1, plist.pl_list[i].ps_pool_name);
    printf("       Pool device:  %d,%d\n", 
	   plist.pl_list[i].ps_pool_major, plist.pl_list[i].ps_pool_minor);
    printf("       Pool ID:      %"FMT_64"x\n", plist.pl_list[i].ps_pool_id);
    printf("       In use:       %s\n", plist.pl_list[i].ps_in_use ? "Yes": "No");
    printf("       Capacity:     %"FMT_64"d\n", plist.pl_list[i].ps_capacity); 
    printf("       Dlocks:       %d\n", plist.pl_list[i].ps_dlocks);


    if (listsub){
      devName = (char *)malloc(50);
      strcpy(devName, POOL_DIR_PATH);
      strcat(devName, plist.pl_list[i].ps_pool_name);
      major = plist.pl_list[i].ps_pool_major;
      minor = plist.pl_list[i].ps_pool_minor;

      if ((fd = open(devName, O_RDONLY)) < 0){
	fprintf(stderr, "Couldn't open pool device %s\n", devName);
	exit(-1);
      }      
      spstat = (pool_spstatus_t *)malloc(spstatsize * sizeof(pool_spstatus_t));
      memset(spstat, 0, spstatsize*sizeof(pool_spstatus_t));
      splist.spl_count = spstatsize;
      splist.spl_list = spstat;

      if (ioctl(fd, POOL_SPLIST, &splist) < 0){
	fprintf(stderr, "pool communications error: %s\n", strerror(errno));
	exit(-1);
      }

      printf("       Subpool List (%d subpools) for %d,%d (%s)\n", 
	     splist.spl_count, major, minor, devName);

      /* Note: no per subpool dlock info is returned.  maybe add this later */

      for(j=0; j< splist.spl_count; j++){
	printf("\t[%d] offset: %9"FMT_64"d  blocks: %"FMT_64"d  striping: %u\n", j+1, 
	       splist.spl_list[j].sps_boffset, splist.spl_list[j].sps_bcount,
	       splist.spl_list[j].sps_stripe);
	printf("\t    subpool devices: ");
	for(k=0; k<splist.spl_list[j].sps_numdevs; k++) {
	  name = (char *)malloc(50);
	  if (getname(splist.spl_list[j].sps_devs[k], name) >= 0) {
	    printf("%s ", name+5);
	    free(name);
	  }
	  else {
	    printf("%d,%d ", MAJOR(splist.spl_list[j].sps_devs[k]),
		MINOR(splist.spl_list[j].sps_devs[k]));
	    free(name);
	  }
	}
	printf("\n");
      }

      free(devName);
      free(spstat);
    }

  }
  printf("\n--------------------------------------------------------------------\n");

  free(pstat);
  return 0;
}



/******************************************************************************
*******************************************************************************
** NAME
**	getname
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int getname(unsigned int dev, char *devname)
{
  DIR			*dir;
  struct direct 	*file;
  struct stat		stats;
  char			*path = NULL;
  int		        sd;
  int			d_namlen;
  unsigned int		major = MAJOR(dev);
  unsigned int		minor = MINOR(dev);

  dir = opendir(SEARCH_DIR);
  rewinddir(dir);

  while ((file = readdir(dir)) != NULL) {

    sd = strncmp(file->d_name, "sd", 2);

    d_namlen = strlen(file->d_name);

    /*
    if (sd != 0) 
      continue;

    if (!isdigit(file->d_name[d_namlen - 1])) 
      continue;
    */

    path = (char *)malloc(strlen(SEARCH_DIR) + d_namlen + 1);

    if (path == NULL) {
      fprintf(stderr, "unable to allocate memory\n");
      exit(1);
    }

    strcpy(path, SEARCH_DIR);
    strcat(path, file->d_name);
      
    stat(path, &stats);
    if (S_ISDIR(stats.st_mode) || S_ISCHR(stats.st_mode)) {
      free(path);
      continue;
    }    
    if (MAJOR(stats.st_rdev) == major && MINOR(stats.st_rdev) == minor){
      strcpy(devname, path);
      free(path);
      return 0;
    }
    free(path);
    continue;
  }
  return(-1);
}



/******************************************************************************
*******************************************************************************
** NAME
**	printUsage
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

static void printUsage(char *progName, int status)
{
  printf("Usage: %s [-hds]\n", progName);
  
  if (status != EXIT_SUCCESS)
    printf("Try `%s -h' for more information.\n", progName);
  
  if (status != EXIT_FAILURE) {
    printf("  -h         Print this help, then exit.\n");
    printf("  -d         Debugging mode.\n");
    printf("  -s         Print subpool information.\n");
  }
  exit(status);
}










